/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include <memory.h>
#include "scheme.h"
#include "schpriv.h"
#include "../gc/gc.h"

#define TRUE 1
#define FALSE 0

#define DEFINE_CLASS "make-class*"
#define CREATE_OBJECT "make-object"

#define IVAR "ivar"
#define UQ_IVAR "uq-ivar"
#define CVAR "uq-cvar"
#define UQ_CVAR "uq-cvar"

#define MAKE_GENERIC "make-generic"
#define UQ_MAKE_GENERIC "uq-make-generic"

#define IS_CLASS "class?"
#define IS_OBJECT "object?"
#define IS_A "is-a?"
#define SUBCLASS "subclass?"

#define SUPER_INIT "super-init"

#define COLLECT_GARBAGE "collect-garbage"

static Scheme_Object *pub_symbol, *pri_symbol, *static_symbol, *local_symbol;
static Scheme_Object *inh_symbol, *inh_from_symbol;
static Scheme_Object *ren_symbol, *ren_from_symbol;
static Scheme_Object *share_symbol, *share_from_symbol;
static Scheme_Object *init_symbol;
static Scheme_Object *quote_symbol, *uq_ivar_symbol;

static Scheme_Object *this_symbol, *super_symbol;

static Scheme_Object *scheme_undef, *inherited_filler;

#define cons scheme_make_pair
  
enum {
  varEXPRESSION,
  varVALUE,
  varLOCAL,
  varPRIMITIVE_FUNC,
  varINHERIT,
  varSHARE
};

typedef struct Scheme_Ivar_Delay {
  Scheme_Type type;
  Scheme_Closed_Prim *f;
} Scheme_Ivar_Delay;

typedef struct ClassVariable {
  Scheme_Object *name;
  short toeval; /* TRUE if this value is a class variable */
  short vartype;
  union {
    Scheme_Object *value;
    Scheme_Closed_Prim *f;
    struct {
      Scheme_Object *name;
      int which;
    } source;
  } u;
  int pos;
  Scheme_Ivar_Delay *delay;
  struct ClassVariable *next;
} ClassVariable;

typedef struct Scheme_Class {
  Scheme_Type type;

  Scheme_Env *env;

  int priminit;
  Scheme_Object *init;
  Scheme_Closed_Prim *initf;

  Scheme_Object *unbound_handler;

  ClassVariable *pub;
  ClassVariable *pri;

  int num_supers;
  struct Scheme_Class **supers;
  Scheme_Object **supernames;

  int num_public, num_private;
  Scheme_Object **public_names;

  const char *classname;

  int num_all_supers;
  struct Scheme_Class **all_supers; /* breadth-first list of superclasses */
  int *super_starts; /* For each immediate superclass, offset into all_supers */
} Scheme_Class;

typedef struct {
  Scheme_Type type;

  Scheme_Object *init;
  Scheme_Object *unbound;

  ClassVariable *pub;
  ClassVariable *pri;

  int num_supers;
  Scheme_Object **supers;
  Scheme_Object **supernames;
} Class_Data;

typedef struct Scheme_Obj_Frame {
  Scheme_Type type;
  struct Internal_Object *obj;
  Scheme_Class *sclass;
  Scheme_Object *name;

  Scheme_Env *pri_env; /* env extended with private ivars */
  Scheme_Env *init_env; /* pri_env extend with super-init, etc. */

  Scheme_Object *unbound_handler, *initfunc;
  struct Scheme_Obj_Frame **nexts;
  int num_nexts;
  int start_nexts;
  int inited;

  Scheme_Object **public_values; /* can point to overidden */
  Scheme_Object **env_values;
  Scheme_Object **private_values; /* used only until object is created */
} Scheme_Obj_Frame;

typedef struct Internal_Object {
  Scheme_Class_Object o;
  Scheme_Obj_Frame *frames;
  Scheme_Obj_Frame **all_frames;
} Internal_Object;

typedef struct {
  Scheme_Obj_Frame *frame;
  int which;
  int apply;
} SuperInitData;

typedef struct {
  Scheme_Type type;
  Scheme_Class *sclass;
  int superclass;
  int framepos;
} Generic_Data;

static Scheme_Object **scheck_hash = NULL;
long scheck_size = 0, scheck_count = 0, scheck_step = 17;

static Scheme_Class *null_class = NULL;

static Scheme_Object *MakeSuperInitPrim(Scheme_Obj_Frame *, int, int);
static int FindName(Scheme_Class *sclass, Scheme_Object *symbol);

/**********************************************************************/

static void begin_dup_symbol_check(void)
{
  long i;

  for (i = 0; i < scheck_size; i++)
    scheck_hash[i] = 0;
  scheck_count = 0;
}

static void dup_symbol_check(Scheme_Object *symbol)
{
  long pos;

  if (scheck_count >= scheck_size / 2) {
    Scheme_Object **old = scheck_hash;
    long oldsize = scheck_size, i;

    scheck_size = scheck_size ? 2 * scheck_size : 100;
    scheck_hash = 
      (Scheme_Object **)scheme_malloc_atomic(i = scheck_size 
					     * sizeof(Scheme_Object *));
    memset(scheck_hash, 0, i);

    scheck_count = 0;
    for (i = 0; i < oldsize; i++)
      if (old[i])
	dup_symbol_check(old[i]);
  }

  pos = (((long)symbol) >> 2) % scheck_size;
  if (pos < 0)
    pos = -pos;
  while(scheck_hash[pos]) {
    if (SAME_OBJ(scheck_hash[pos], symbol))
      scheme_signal_error("make-class: duplicate ivar name %s",
			  SCHEME_SYM_VAL(symbol));
    pos = (pos + scheck_step) % scheck_size;
  }

  scheck_hash[pos] = symbol;
}

static void end_dup_symbol_check(void)
{
}

/**********************************************************************/

static int CompareObjectPtrs(Scheme_Object **a, Scheme_Object **b)
{
  return ((long)*a) - ((long)*b);
}

Scheme_Object *MapEval(Scheme_Object *list, Scheme_Env *env)
{
  if (SCHEME_NULLP(list))
    return scheme_null;
  else
    return cons(_scheme_eval_compiled(SCHEME_CAR(list), env),
		MapEval(SCHEME_CDR(list), env));
}

/**********************************************************************/

static ClassVariable *ReadItemList(char *what, Scheme_Object *varlists, 
				   Scheme_Env *env, int vartype,
				   Class_Data *data,
				   ClassVariable *start,
				   Scheme_Compile_Info *rec)
{
  Scheme_Object *l, *var, *value, *vars;
  ClassVariable *classvar = start, *oldvar;
  int position, alias, eval;

  if (!varlists)
    return start;

  eval = (vartype == varVALUE);
  alias = (vartype == varINHERIT || vartype == varSHARE);

  while (classvar && classvar->next)
    classvar = classvar->next;

  for (; !SCHEME_NULLP(varlists); varlists = SCHEME_CDR(varlists)) {
    vars = SCHEME_CAR(varlists);

    if (data) {
      Scheme_Object *name;
      int i;

      name = SCHEME_CAR(vars);
      for (i = 0; i < data->num_supers; i++)
	if (SAME_OBJ(name, data->supernames[i]))
	  break;

      if (i >= data->num_supers)
	scheme_signal_error(DEFINE_CLASS
			    ": bad class name \"%s\" for %s clause",
			    SCHEME_SYM_VAL(name), what);

      position = i;

      vars = SCHEME_CDR(vars);
    } else
      position = -1;

    for (; !SCHEME_NULLP(vars); vars = SCHEME_CDR(vars)) {
      l = SCHEME_CAR(vars);
    
      if (SCHEME_LISTP(l)) {
	var = SCHEME_CAR(l);
	
	l = SCHEME_CDR(l);
	if (SCHEME_NULLP(l))
	  value = NULL;
	else
	  value = SCHEME_CAR(l);
      } else {
	var = l;
	value = NULL;
      }

      if (!value) {
	if (alias)
	  value = var;
	else
	  value = scheme_compiled_void(rec->can_optimize_constants);
      }

      scheme_check_identifier(DEFINE_CLASS, var, NULL, env);

      oldvar = classvar;
      classvar = (ClassVariable *)scheme_malloc(sizeof(ClassVariable));
      classvar->name = var;
      classvar->toeval = eval;
      classvar->vartype = vartype;
      classvar->delay = NULL;
    
      if (alias) {
	classvar->u.source.name = value;
	classvar->u.source.which = position;
      } else
	classvar->u.value = scheme_compile_simple(value, env, rec);
      classvar->next = NULL;

      if (oldvar)
	oldvar->next = classvar;
      else
	start = classvar;
    }
  }

  return start;
}

static ClassVariable *EvalClassItems(ClassVariable *var, Scheme_Env *env)
{
  ClassVariable *start = var, *prev = NULL, *naya;

  while (var) {
    if (var->toeval) {
      naya = (ClassVariable *)scheme_malloc(sizeof(ClassVariable));
      memcpy(naya, var, sizeof(ClassVariable));
      naya->u.value = _scheme_eval_compiled(var->u.value, env);
      if (!prev)
	start = naya;
      else
	prev->next = naya;
      prev = naya;
      var = naya->next;
    } else
      break;
  }

  return start;
}

static void GenCheckIvarList(Scheme_Object *clause, 
			     Scheme_Object *naya, 
			     Scheme_Object *old,
			     int flags)
{
  Scheme_Object *l, *p, *n;

  for (l = naya; SCHEME_PAIRP(l); l = SCHEME_CDR(l)) {
    p = SCHEME_CAR(l);
    if (SCHEME_SYMBOLP(p)) {
      if (flags < 0)
	scheme_signal_error(DEFINE_CLASS ": %s clause must contain pairs",
			    SCHEME_SYM_VAL(clause));
    } else if (SCHEME_PAIRP(p)) {
      if (!flags)
	scheme_signal_error(DEFINE_CLASS ": %s clause must contain only symbols",
			    SCHEME_SYM_VAL(clause));

      n = SCHEME_CAR(p);
      p = SCHEME_CDR(p);
      if (!SCHEME_SYMBOLP(n))
	scheme_signal_error(DEFINE_CLASS ": badly formed %s clause (variable"
			    " name not a symbol)", SCHEME_SYM_VAL(clause));

      if (!SCHEME_NULLP(p)) {
	if (!SCHEME_PAIRP(p))
	  scheme_signal_error(DEFINE_CLASS ": badly formed %s clause (variable"
			      " %s declaration is improper)", 
			      SCHEME_SYM_VAL(clause), SCHEME_SYM_VAL(n));	
	if (!SCHEME_NULLP(SCHEME_CDR(p)))
	  scheme_signal_error(DEFINE_CLASS ": badly formed %s clause (variable"
			      " declaration not a pair)", SCHEME_SYM_VAL(clause));
	if (flags < 0 || flags == 2) {
	  Scheme_Object *name = n;
	  n = SCHEME_CAR(p);
	  if (!SCHEME_SYMBOLP(n))
	    scheme_signal_error(DEFINE_CLASS ": variable association for "
				"%s (in %s clause) "
				"must be a symbol",
				SCHEME_SYM_VAL(name),
				SCHEME_SYM_VAL(clause));
	}
      } else if (flags < 0)
	scheme_signal_error(DEFINE_CLASS ": %s clause must contain pairs",
			    SCHEME_SYM_VAL(clause));
    } else
      scheme_signal_error(DEFINE_CLASS ": badly formed %s clause (declaration"
			  " not a symbol or pair)", SCHEME_SYM_VAL(clause));
  }

  if (!SCHEME_NULLP(l))
    scheme_signal_error(DEFINE_CLASS ": badly formed %s clause (improper"
			" variable list)", 
			SCHEME_SYM_VAL(clause));

}

static Scheme_Object *CheckIvarList(Scheme_Object *l, Scheme_Object *old,
				    int flags)
{
  GenCheckIvarList(SCHEME_CAR(l), SCHEME_CDR(l), old, flags);
  return cons(SCHEME_CDR(l), old);
}

static Scheme_Object *CheckFromIvarList(Scheme_Object *l, Scheme_Object *old,
					 int flags)
{
  Scheme_Object *rest, *clause;

  clause = SCHEME_CAR(l);
  rest = SCHEME_CDR(l);

  if (!SCHEME_PAIRP(rest))
    scheme_signal_error(DEFINE_CLASS ": badly formed %s clause",
			SCHEME_SYM_VAL(clause));

  if (!SCHEME_SYMBOLP(SCHEME_CAR(rest)))
    scheme_signal_error(DEFINE_CLASS ": %s clause requires a symbol"
			" for the superclass name",
			SCHEME_SYM_VAL(clause));

  GenCheckIvarList(clause, SCHEME_CDR(rest), old, flags);
  return cons(rest, old);
}


static Scheme_Object *Prefix(Scheme_Object *l, Scheme_Object *all, int withfrom)
{
  Scheme_Object *a, *al;

  for (; !SCHEME_NULLP(all); all = SCHEME_CDR(all)) {
    al = SCHEME_CAR(all);
    if (withfrom)
      al = SCHEME_CDR(al);

    for (; !SCHEME_NULLP(al); al = SCHEME_CDR(al)) {
      a = SCHEME_CAR(al);
      if (SCHEME_PAIRP(a))
	a = SCHEME_CAR(a);
      l = cons(a, l);
    }
  }

  return l;
}

static Scheme_Object *expandall(Scheme_Object *vars, Scheme_Env *env, int depth,
				int expand_expr, Scheme_Env *first_from)
{
  Scheme_Object *first, *last;
  Scheme_Object *name, *var;

  name = SCHEME_CAR(vars);
  vars = SCHEME_CDR(vars);

  first = last = cons(name, scheme_null);

  if (first_from) {
    name = SCHEME_CAR(vars);
    vars = SCHEME_CDR(vars);
    
    if (depth < 0)
      name = scheme_annotate_frame_id(first_from, name);

    SCHEME_CDR(first) = last = cons(name, scheme_null);
  }

  while (!SCHEME_NULLP(vars)) {
    var = SCHEME_CAR(vars);

    if (!SCHEME_PAIRP(var)) {
      if (depth < 0)
	var = scheme_annotate_frame_id(env, var);
    } else {
      Scheme_Object *rest, *name;
    
      name = SCHEME_CAR(var);
      rest = SCHEME_CDR(var);

      if (depth < 0)
	name = scheme_annotate_frame_id(env, name);

      if (expand_expr)
	rest = scheme_expand_expr(rest, env, depth);
      
      var = cons(name, rest);
    }
    
    var = cons(var, scheme_null);

    SCHEME_CDR(last) = var;
    last = var;

    vars = SCHEME_CDR(vars);
  }

  return first;
}

static void CheckInherited(Scheme_Class *sclass, ClassVariable *item)
{
  for (; item; item = item->next) {
    if (item->vartype == varINHERIT || item->vartype == varSHARE) {
      Scheme_Class *from_class;
      int i, found;
      
      found = 0;

      if (item->u.source.which < 0)
	from_class = sclass;
      else {
	from_class = sclass->supers[item->u.source.which];

	if (FindName(from_class, item->u.source.name) >= 0)
	  found = 1;
      }
      
      for (i = 0; i < from_class->num_all_supers; i++)
	if (FindName(from_class->all_supers[i], item->u.source.name) >= 0) {
	  found = 1;
	  break;
	}

      if (!found)
	scheme_signal_error(DEFINE_CLASS
			    ": inherited ivar \"%s\" not found in superclass",
			    SCHEME_SYM_VAL(item->u.source.name));
    }
  }
}

static void InsureClassReady(Scheme_Class *sclass)
{
  int i;
  ClassVariable *cvar;

  if (sclass->num_public && !sclass->public_names) {
    sclass->public_names = 
      (Scheme_Object **)scheme_malloc_atomic(sclass->num_public
					     * sizeof(Scheme_Object *));

    for (i = 0, cvar = sclass->pub; cvar; cvar = cvar->next, i++)
      sclass->public_names[i] = cvar->name;

    qsort((char *)sclass->public_names, sclass->num_public, 
	  sizeof(Scheme_Object *), (int (*)())CompareObjectPtrs); 

    for (cvar = sclass->pub; cvar; cvar = cvar->next)
      cvar->pos = FindName(sclass, cvar->name);
  }
}
  
static Scheme_Object *DefineClass_Execute(Scheme_Object *form, Scheme_Env *env)
{
  Scheme_Class *sclass;
  Class_Data *data;
  ClassVariable *item;
  Scheme_Env *cvar_env;
  Scheme_Object *superobj, *superlist;
  int i, total;

  data = (Class_Data *)form;

  sclass = (Scheme_Class *)scheme_malloc(sizeof(Scheme_Class));

  sclass->env = env;

  sclass->priminit = FALSE;

  sclass->num_supers = data->num_supers;

  if (sclass->num_supers) {
    sclass->supers = (Scheme_Class **)scheme_malloc(sizeof(Scheme_Object *) 
						    * sclass->num_supers);
    superlist = scheme_null;
    for (i = 0; i < data->num_supers; i++) {
      superobj = _scheme_eval_compiled(data->supers[i], env);
      if (SCHEME_NULLP(superobj)) {
	if (!null_class) {
	  null_class = (Scheme_Class *)scheme_malloc(sizeof(Scheme_Class));

	  null_class->type = scheme_class_type;
	  null_class->env = NULL;

	  null_class->priminit = FALSE;
	  null_class->init = NULL;
	  null_class->initf = NULL;
	  null_class->unbound_handler = NULL;

	  null_class->pub = NULL;
	  null_class->pri = NULL;

	  null_class->num_supers = 0;
	  null_class->supers = NULL;
	  null_class->supernames = NULL;

	  null_class->num_public = 0;
	  null_class->num_private = 0;
	  null_class->public_names = NULL;

	  null_class->classname = "null";

	  null_class->num_all_supers = 0;
	  null_class->all_supers= NULL;
	  null_class->super_starts = NULL;
	}
	superobj = (Scheme_Object *)null_class;
      }
      if (!SCHEME_CLASSP(superobj)) {
	scheme_wrong_type(DEFINE_CLASS, "class", -1, 0, &superobj);
	return scheme_void;
      }
	
      sclass->supers[i] = (Scheme_Class *)superobj;

      if (sclass->supers[i]->priminit && !sclass->supers[i]->initf)
	scheme_signal_error("can't derive from the class"
			    " (%s)", sclass->supers[i]->classname);
      
      InsureClassReady(sclass->supers[i]);
      
      superlist = cons(superobj, superlist);
    }
    cvar_env = scheme_add_compilation_frame(superlist, env, 0);
  } else
    cvar_env = env;

  sclass->supernames = data->supernames;

  /* Make all_supers array */
  total = sclass->num_supers;
  for (i = sclass->num_supers; i--; )
    total += sclass->supers[i]->num_all_supers;
  
  sclass->num_all_supers = total;


  if (total == sclass->num_supers) {
    sclass->all_supers = sclass->supers;
    sclass->super_starts = NULL;
  } else {
    sclass->all_supers = (Scheme_Class **)scheme_malloc(sizeof(Scheme_Class *) 
							* total);
    
    sclass->super_starts = (int *)scheme_malloc(sizeof(int) * total);

    total = sclass->num_supers;
    for (i = 0; i < sclass->num_supers; i++) {
      int j;
      sclass->super_starts[i] = total;
      sclass->all_supers[i] = sclass->supers[i];
      for (j = 0; j < sclass->supers[i]->num_all_supers; j++)
	sclass->all_supers[total++] = sclass->supers[i]->all_supers[j];
    }
  }

  /* Setup public/private */
  data->pub = EvalClassItems(data->pub, cvar_env);

  sclass->pub = data->pub;
  sclass->pri = data->pri;

  sclass->unbound_handler = data->unbound;
  sclass->init = data->init;

  CheckInherited(sclass, sclass->pub);
  CheckInherited(sclass, sclass->pri);

  if (sclass->num_supers)
    sclass->classname = sclass->supers[0]->classname;
  else
    sclass->classname = NULL;

  sclass->type = scheme_class_type;

  sclass->num_public = 0;
  for (item = sclass->pub; item; item = item->next)
    sclass->num_public++;
  sclass->num_private = 0;
  for (item = sclass->pri; item; item = item->next)
    sclass->num_private++;

  sclass->public_names = NULL;

  InsureClassReady(sclass);

  return (Scheme_Object *)sclass;
}

static Scheme_Object *Do_DefineClass(Scheme_Object *form, Scheme_Env *env,
				     Scheme_Compile_Info *rec, int depth)
{
#define BAD_MSG ": bad syntax or missing (init ...) in class definition"

  Scheme_Object *l, *superclass, *vars, *tag;
  Scheme_Object **superclasses, **names, *super_list;
  Scheme_Object *init = NULL, *unbound = NULL;
  Scheme_Object *pub = scheme_null, *pri = scheme_null;
  Scheme_Object *classpub = scheme_null, *local = scheme_null;
  Scheme_Object *inh = scheme_null, *inh_from = scheme_null;
  Scheme_Object *ren = scheme_null, *ren_from = scheme_null;
  Scheme_Object *share = scheme_null, *share_from = scheme_null;
  Scheme_Object *ivarlist;
  Scheme_Env *objenv, *cvar_env, *initenv;
  Class_Data *data;
  ClassVariable *cl_vars;
  int i, len, num_supers, fr_flag;
  static char *name_buffer;
  static int name_len = 0;

  l = SCHEME_CDR(form);
  SCHEME_ASSERT(SCHEME_LISTP(l), DEFINE_CLASS ": bad syntax");
  SCHEME_ASSERT(!SCHEME_NULLP(l), DEFINE_CLASS 
		 ": class definition missing superclass");
  superclass = SCHEME_CAR(l);

 {
    Scheme_Object *l;

    num_supers = 0;
    l = superclass;
    while (SCHEME_PAIRP(l)) {
      l = SCHEME_CDR(l);
      num_supers++;
    }
    
    SCHEME_ASSERT(SCHEME_NULLP(l), DEFINE_CLASS ": bad syntax in class list");

    superclasses = (Scheme_Object **)scheme_malloc(sizeof(Scheme_Object *) 
						   * num_supers);
    names = (Scheme_Object **)scheme_malloc(sizeof(Scheme_Object *)
					    * num_supers);


    l = superclass;
    for (i = 0; i < num_supers; i++) {
      Scheme_Object *pair;
      pair = SCHEME_CAR(l);

      SCHEME_ASSERT(SCHEME_PAIRP(pair), DEFINE_CLASS 
		    ": bad syntax in class list");
      
      names[i] = SCHEME_CAR(pair);
      
      SCHEME_ASSERT(SCHEME_SYMBOLP(names[i]), DEFINE_CLASS 
		    ": name must be a symbol in class list");

      pair = SCHEME_CDR(pair);

      SCHEME_ASSERT(SCHEME_PAIRP(pair), DEFINE_CLASS 
		    ": bad syntax in class list - no class expression");
      SCHEME_ASSERT(SCHEME_NULLP(SCHEME_CDR(pair)), DEFINE_CLASS 
		    ": bad syntax in class list "
		    "- too many expressions for class");

      if (rec)
	superclasses[i] = scheme_compile_simple(SCHEME_CAR(pair), env, rec);
      else
	superclasses[i] = scheme_expand_expr(SCHEME_CAR(pair), env, depth);	

      l = SCHEME_CDR(l);
    }
  }

  super_list = scheme_null;
  for (i = 0; i < num_supers; i++)
    super_list = cons(names[i], super_list);

  while (1) {
    l = SCHEME_CDR(l);
    SCHEME_ASSERT(SCHEME_LISTP(l), DEFINE_CLASS ": bad syntax");
    if (SCHEME_NULLP(l))
      break;
    vars = SCHEME_CAR(l);

    SCHEME_ASSERT(SCHEME_LISTP(vars) && !SCHEME_NULLP(vars), DEFINE_CLASS
		  BAD_MSG);

    tag = SCHEME_CAR(vars);

    if (SAME_OBJ(tag,  static_symbol)) {
      classpub = CheckIvarList(vars, classpub, 1);
    } else if (SAME_OBJ(tag, pub_symbol)) {
      pub = CheckIvarList(vars, pub, 1);
    } else if (SAME_OBJ(tag, pri_symbol)) {
      pri = CheckIvarList(vars, pri, 1);
    } else if (SAME_OBJ(tag, local_symbol)) {
      local = CheckIvarList(vars, local, 1);
    } else if (SAME_OBJ(tag, inh_symbol)) {
      inh = CheckIvarList(vars, inh, 0);
    } else if (SAME_OBJ(tag, ren_symbol)) {
      ren = CheckIvarList(vars, ren, -1);
    } else if (SAME_OBJ(tag, share_symbol)) {
      share = CheckIvarList(vars, share, 2);
    } else if (SAME_OBJ(tag, inh_from_symbol)) {
      inh_from = CheckFromIvarList(vars, inh_from, 0);
    } else if (SAME_OBJ(tag, ren_from_symbol)) {
      ren_from = CheckFromIvarList(vars, ren_from, 2);
    } else if (SAME_OBJ(tag, share_from_symbol)) {
      share_from = CheckFromIvarList(vars, share_from, -1);
    } else {
      l = SCHEME_CDR(l);
      if (SCHEME_PAIRP(l)) {
	unbound = SCHEME_CAR(l);
	l = SCHEME_CDR(l);
      }
      SCHEME_ASSERT(SCHEME_NULLP(l), DEFINE_CLASS
		    ": syntax error at the end of class definition");
      init = vars;
      break;
    }
  }

  ivarlist = scheme_null;

  ivarlist = Prefix(ivarlist, classpub, 0);
  ivarlist = Prefix(ivarlist, pub, 0);
  ivarlist = Prefix(ivarlist, local, 0);
  ivarlist = Prefix(ivarlist, inh, 0);
  ivarlist = Prefix(ivarlist, inh_from, 1);
  ivarlist = Prefix(ivarlist, share, 0);
  ivarlist = Prefix(ivarlist, share_from, 1);
  ivarlist = Prefix(ivarlist, pri, 0);
  ivarlist = Prefix(ivarlist, ren, 0);
  ivarlist = Prefix(ivarlist, ren_from, 1);

  ivarlist = cons(this_symbol, ivarlist);

  for (i = 0; i < num_supers; i++)
    ivarlist = cons(names[i], ivarlist);

  fr_flag = ((!rec && depth < 0) ? SCHEME_ANNOTATING : 0);

  objenv = scheme_add_compilation_frame(ivarlist, env, 
					SCHEME_AUTO_UNBOX + fr_flag);

  cvar_env = scheme_add_compilation_frame(super_list, env, fr_flag);

  ivarlist = scheme_null;
  for (i = 0; i < num_supers; i++) {
    len = strlen(SCHEME_SYM_VAL(names[i]));
    if (len + 20 > name_len) {
      name_len = 2 * (len + 20);
      name_buffer = (char *)scheme_malloc_atomic(name_len + 1);
    }

    strcpy(name_buffer, "apply-");
    strcat(name_buffer, SCHEME_SYM_VAL(names[i]));
    strcat(name_buffer, "-init");
    ivarlist = cons(scheme_intern_symbol(name_buffer), ivarlist);

    strcpy(name_buffer, SCHEME_SYM_VAL(names[i]));
    strcat(name_buffer, "-init");
    ivarlist = cons(scheme_intern_symbol(name_buffer), ivarlist);
  }

  initenv = num_supers ? scheme_add_compilation_frame(ivarlist, objenv, fr_flag) 
    : objenv;

  if (!rec) {
    /* Just expanding: */
    Scheme_Object *first, *last;

    l = SCHEME_CDR(SCHEME_CDR(form));
    first = last = NULL;
    while (!SCHEME_NULLP(l)) {
      vars = SCHEME_CAR(l);

      tag = SCHEME_CAR(vars);

      if (SAME_OBJ(tag,  static_symbol))
	vars = expandall(vars, cvar_env, depth, 1, NULL);
      else if (SAME_OBJ(tag, pub_symbol) || SAME_OBJ(tag, pri_symbol)
	       || SAME_OBJ(tag, local_symbol))
	vars = expandall(vars, objenv, depth, 1, NULL);
      else if (SAME_OBJ(tag, inh_symbol) || SAME_OBJ(tag, ren_symbol) 
	       || SAME_OBJ(tag, share_symbol)) {
	if (depth >= 0)
	  vars = vars;
	else
	  vars = expandall(vars, objenv, depth, 0, NULL);
      } else if (SAME_OBJ(tag, inh_from_symbol)|| SAME_OBJ(tag, ren_from_symbol)
		 || SAME_OBJ(tag, share_from_symbol)) {
	if (depth >= 0)
	  vars = vars;
	else
	  vars = expandall(vars, objenv, depth, 0, cvar_env);
      } else
	break;

      vars = cons(vars, scheme_null);

      if (!first)
	first = vars;
      else
	SCHEME_CDR(last) = vars;

      last = vars;
      l = SCHEME_CDR(l);
    }

    l = scheme_null;
    if (unbound)
      l = cons(scheme_expand_expr(unbound, objenv, depth), l);
    if (init)
      l = cons(scheme_expand_expr(init, initenv, depth), l);
    
    if (!first)
      first = l;
    else
      SCHEME_CDR(last) = l;

    
    superclass = scheme_null;
    for (i = num_supers; i--; ) {
      Scheme_Object *name;

      name = names[i];
      if (depth < 0)
	name = scheme_annotate_frame_id(cvar_env, name);
	
      superclass = cons(cons(name, cons(superclasses[i], scheme_null)),
			superclass);
    }

    return cons(SCHEME_CAR(form), cons(superclass, first));
  } else {
    /* Compiling: */
    data = (Class_Data *)scheme_malloc(sizeof(Class_Data));
    
    data->num_supers = num_supers;
    data->supers = superclasses;
    data->supernames = names;
    
    data->pub = ReadItemList("class", classpub, cvar_env, varVALUE, NULL, NULL, rec);
    data->pub = ReadItemList("public", pub, objenv, varEXPRESSION, NULL, data->pub, rec);
    data->pub = ReadItemList("local", local, objenv, varLOCAL, NULL, data->pub, rec);
    data->pub = ReadItemList("inherit", inh, objenv, varINHERIT, NULL, data->pub, rec);
    data->pub = ReadItemList("inherit-from", inh_from, objenv, varINHERIT, data, data->pub, rec);
    data->pub = ReadItemList("share", share, objenv, varSHARE, NULL, data->pub, rec);
    data->pub = ReadItemList("share-from", share_from, objenv, varSHARE, data, data->pub, rec);
    
    data->pri = ReadItemList("private", pri, objenv, varEXPRESSION, NULL, NULL, rec);
    data->pri = ReadItemList("rename", ren, objenv, varINHERIT, NULL, data->pri, rec);
    data->pri = ReadItemList("rename-from", ren_from, objenv, varINHERIT, data, data->pri, rec);

    begin_dup_symbol_check();
    cl_vars = data->pub;
    while (cl_vars) {
      dup_symbol_check(cl_vars->name);
      cl_vars = cl_vars->next;
    }
    cl_vars = data->pri;
    while (cl_vars) {
      dup_symbol_check(cl_vars->name);
      cl_vars = cl_vars->next;
    }
    end_dup_symbol_check();
    
    if (unbound)
      data->unbound = scheme_compile_simple(unbound, objenv, rec);
    else
      data->unbound = NULL;
    
    if (init)
      data->init = scheme_compile_simple(init, initenv, rec);
    else
      data->init = NULL;
    
    data->type = scheme_class_data_type;

    rec->can_return_closure = 1;
    rec->max_let_depth = 0;
    
    return scheme_make_syntax_compilation(DefineClass_Execute, 
					  (Scheme_Object *)data);
  }
}

static Scheme_Object *DefineClass(Scheme_Object *form, Scheme_Env *env,
				  Scheme_Compile_Info *rec)
{
  return Do_DefineClass(form, env, rec, 0);
}

static Scheme_Object *DefineClass_expand(Scheme_Object *form, Scheme_Env *env,
					 int depth)
{
  return Do_DefineClass(form, env, NULL, depth);
}

/**********************************************************************/

static Scheme_Object *CV_Bundle(ClassVariable *cvar)
{
  Scheme_Object *l = scheme_null, *f;

  while (cvar) {
    if (cvar->vartype == varPRIMITIVE_FUNC)
      f = scheme_null;
    else if (cvar->vartype == varINHERIT || cvar->vartype == varSHARE)
      f = cons(cvar->u.source.name, scheme_make_integer(cvar->u.source.which));
    else
      f = cvar->u.value;
    l = cons(cvar->name,
	     cons(cvar->toeval ? scheme_true : scheme_false,
		  cons(scheme_make_integer(cvar->vartype),
		       cons(f, l))));
    cvar = cvar->next;
  }

  return l;
}

#define BAD_CCD "bad compiled class definition"

static ClassVariable *CV_Unbundle(Scheme_Object *l)
{
  ClassVariable *cvar = NULL, *prev;
  Scheme_Object *v;

  while (!SCHEME_NULLP(l)) {
    prev = cvar;
    cvar = (ClassVariable *)scheme_malloc(sizeof(ClassVariable));
    cvar->next = prev;
    cvar->delay = NULL;

    SCHEME_ASSERT(SCHEME_PAIRP(l), BAD_CCD);
    cvar->name = SCHEME_CAR(l);
    SCHEME_ASSERT(SCHEME_SYMBOLP(cvar->name), BAD_CCD);
    l = SCHEME_CDR(l);

    SCHEME_ASSERT(SCHEME_PAIRP(l), BAD_CCD);
    SCHEME_ASSERT(SCHEME_BOOLP(SCHEME_CAR(l)), BAD_CCD);
    cvar->toeval = SAME_OBJ(SCHEME_CAR(l), scheme_true);
    l = SCHEME_CDR(l);

    SCHEME_ASSERT(SCHEME_PAIRP(l), BAD_CCD);
    SCHEME_ASSERT(SCHEME_INTP(SCHEME_CAR(l)), BAD_CCD);
    cvar->vartype = SCHEME_INT_VAL(SCHEME_CAR(l));
    l = SCHEME_CDR(l);

    if (cvar->vartype == varPRIMITIVE_FUNC)
      scheme_signal_error("compiled class deifnition had primitives");

    SCHEME_ASSERT(SCHEME_PAIRP(l), BAD_CCD);
    v = SCHEME_CAR(l);
    if (cvar->vartype == varINHERIT || cvar->vartype == varSHARE) {
     cvar->u.source.name = SCHEME_CAR(v);
     cvar->u.source.which = SCHEME_INT_VAL(SCHEME_CDR(v));
    } else
      cvar->u.value = v;
    l = SCHEME_CDR(l);
  }

  return cvar;
}

static Scheme_Object *write_Class_Data(Scheme_Object *obj)
{
  Class_Data *data;
  Scheme_Object *superlist;
  int i;

  data = (Class_Data *)obj;

  superlist = scheme_null;
  for (i = data->num_supers; i--; )
    superlist = cons(data->supernames[i],
		     cons(data->supers[i], superlist));

  return cons(data->init ? data->init : scheme_null,
	      cons(data->unbound ? data->unbound : scheme_null,
		   cons(CV_Bundle(data->pub),
			cons(CV_Bundle(data->pri),
			     cons(scheme_make_integer(data->num_supers),
				  superlist)))));
}

static Scheme_Object *read_Class_Data(Scheme_Object *obj)
{
  Class_Data *data;
  int i;

  data = (Class_Data *)scheme_malloc(sizeof(Class_Data));
  data->type = scheme_class_data_type;

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CCD);
  data->init = SCHEME_CAR(obj);
  if (SCHEME_NULLP(data->init))
    data->init = NULL;
  obj = SCHEME_CDR(obj);

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CCD);
  data->unbound = SCHEME_CAR(obj);
  if (SCHEME_NULLP(data->unbound))
    data->unbound = NULL;
  obj = SCHEME_CDR(obj);

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CCD);
  data->pub = CV_Unbundle(SCHEME_CAR(obj));
  obj = SCHEME_CDR(obj);

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CCD);
  data->pri = CV_Unbundle(SCHEME_CAR(obj));
  obj = SCHEME_CDR(obj);

  SCHEME_ASSERT(SCHEME_PAIRP(obj), BAD_CCD);
  data->num_supers = SCHEME_INT_VAL(SCHEME_CAR(obj));
  obj = SCHEME_CDR(obj);

  data->supers = (Scheme_Object **)scheme_malloc(data->num_supers 
						 * sizeof(Scheme_Object *));
  data->supernames = (Scheme_Object **)scheme_malloc(data->num_supers 
						      * sizeof(Scheme_Object *));

  for (i = 0; i < data->num_supers; i++) {
    data->supernames[i] = SCHEME_CAR(obj);
    obj = SCHEME_CDR(obj);
    data->supers[i] = SCHEME_CAR(obj);
    obj = SCHEME_CDR(obj);
  }

  return (Scheme_Object *)data;
}

/**********************************************************************/

Scheme_Object *scheme_make_class(const char *name, Scheme_Object *sup, 
				 Scheme_Method_Prim *initf)
{
  Scheme_Class *sclass;

  if (sup && !SCHEME_CLASSP(sup))
    scheme_signal_error("make-prim-class: bad superclass");

  sclass = (Scheme_Class *)scheme_malloc(sizeof(Scheme_Class));

  if (sup) {
    Scheme_Class *super = (Scheme_Class *)sup;
    int i;

    InsureClassReady(super);
    sclass->num_supers = 1;
    sclass->supers = (Scheme_Class **)scheme_malloc(sizeof(Scheme_Class *));
    sclass->supers[0] = super;
    sclass->supernames = (Scheme_Object **)scheme_malloc(sizeof(Scheme_Object *));
    sclass->supernames[0] = super_symbol;
    if (super->num_all_supers) {
      int count = 1 + super->num_all_supers;
      sclass->num_all_supers = count;
      sclass->all_supers = 
	(Scheme_Class **)scheme_malloc(count * sizeof(Scheme_Class *));
      memcpy(sclass->all_supers + 1, super->all_supers,
	     (count - 1) * sizeof(Scheme_Class *));
      sclass->all_supers[0] = super;

      sclass->super_starts = (int *)scheme_malloc(sizeof(int) * count);
      for (i = 1; i < count; i++)
	sclass->super_starts[i] = super->super_starts[i - 1] + 1;
      sclass->super_starts[0] = 1;
    } else {
      sclass->num_all_supers = sclass->num_supers;
      sclass->all_supers = sclass->supers;
      sclass->super_starts = scheme_malloc(sizeof(int));
      *sclass->super_starts = 1;
    }
  } else {
    sclass->num_supers = 0;
    sclass->supers = NULL;
    sclass->supernames = NULL;
    sclass->num_all_supers = 0;
    sclass->all_supers = NULL;
  }
  
  sclass->env = NULL;

  sclass->pub = NULL;
  sclass->pri = NULL;

  sclass->type = scheme_class_type;

  sclass->priminit = TRUE;
  sclass->initf = (Scheme_Closed_Prim *)initf;

  sclass->num_public = 0;
  sclass->public_names = NULL;
  sclass->num_private = 0;

  sclass->classname = name;

  return (Scheme_Object *)sclass;
}

char *scheme_get_class_name(Scheme_Object *c)
{
  if (!SCHEME_CLASSP(c))
    return NULL;

  return (char *)((Scheme_Class *)c)->classname;
}

void scheme_add_method(Scheme_Object *c, const char *name,
		       Scheme_Method_Prim *f)
{
  Scheme_Class *sclass;
  ClassVariable *item, **p;

  if (!SCHEME_CLASSP(c))
    return;
  
  sclass = (Scheme_Class *)c;

  item = (ClassVariable *)scheme_malloc(sizeof(ClassVariable));
  
  item->vartype = varPRIMITIVE_FUNC;
  item->name = scheme_intern_symbol(name);
  item->u.f = (Scheme_Closed_Prim *)f;

  item->delay = (Scheme_Ivar_Delay *)scheme_malloc(sizeof(Scheme_Ivar_Delay));
  item->delay->type = scheme_delayed_ivar_type;
  item->delay->f = (Scheme_Closed_Prim *)f;
  
  p = &sclass->pub;

  while (*p)
    p = &((*p)->next);

  item->next = *p;
  *p = item;

  sclass->num_public++;
}

/**********************************************************************/

static void ForceIvarValue(Scheme_Obj_Frame *frame, int pos, 
			   Scheme_Object *delay, int forcetype)
{
  Scheme_Object *v;
  
  if (forcetype < 0)
    v = scheme_undef;
  else
    v = scheme_make_closed_prim(((Scheme_Ivar_Delay *)delay)->f, 
				frame->obj);

  v = scheme_make_envunbox(v);
  if (SCHEME_DIVARP(frame->public_values[pos]))
    frame->public_values[pos] = v;
  if (SCHEME_DIVARP(frame->env_values[pos]))
    frame->env_values[pos] = v;
}

static int FindName(Scheme_Class *sclass, Scheme_Object *symbol)
{
  int p, w, o;
  Scheme_Object **pn, *n;

  if (!sclass->num_public)
    return -1;

  o = 0;
  p = sclass->num_public / 2;
  w = sclass->num_public;

  pn = sclass->public_names;
  while (1) {
    n = pn[p];
    if (SAME_OBJ(n, symbol))
      return p;
    if (w == 1)
      return -1;
    if (((long)n) < ((long)symbol)) {
      w = o + w - p;
      o = p;
    } else
      w = p - o;
    p = o + w / 2;
  }

  return p;
}

static Scheme_Object *FrameBox(Scheme_Obj_Frame *fr, Scheme_Object *name, 
			       int start_here, Scheme_Object *setbox,
			       Scheme_Class *sclass, int force)
/*
   Severely overloaded function for breadth-first search. 
   fr is starting frame
   name is ivar name; NULL => look for unbound handler
   start_here (when sclass is NULL): if 0, start search with children
   setbox: if not NULL, set ivar value
   sclass: if not NULL, first find class, then search. If this
     class is never found, then it is returned asthe value
*/
{
  int counter;
  Scheme_Obj_Frame **array;
  Scheme_Object *v;

  if (sclass)
    start_here = 0;

  counter = fr->sclass->num_all_supers;
  array = fr->nexts;

  do { 
    if (sclass && SAME_PTR(fr->sclass, sclass)) {
      counter = fr->sclass->num_all_supers;
      array = fr->nexts;
      start_here = 1;
      sclass = NULL;
    }
    
    if (start_here) {
      if (name) {
	int pos;
	
	pos = FindName(fr->sclass, name);
	if (pos >= 0) {
	  if (setbox)
	    fr->env_values[pos] = setbox;
	  else {
	    v = fr->public_values[pos];
	    if (force && SCHEME_DIVARP(v)) {
	      ForceIvarValue(fr, pos, v, force);
	      v = fr->public_values[pos];
	    }
	    return v;
	  }
	}
      } else if (fr->unbound_handler)
	return fr->unbound_handler;
    } else if (!sclass)
      start_here = 1;

    if (counter--)
      fr = *(array++);
    else
      break;
  } while (1);

  return (Scheme_Object *)sclass;
}

static Scheme_Object *LookupInherited(Scheme_Obj_Frame *frame, 
				      ClassVariable *cvar)
{
  int start;

  if (cvar->u.source.which > 0) {
    frame = frame->nexts[cvar->u.source.which];
    start = 1;
  } else
    start = 0;

  return FrameBox(frame, cvar->u.source.name, start, NULL, NULL, -1);
}

static Scheme_Obj_Frame *BuildObjectFrames(Internal_Object *obj,
					   Scheme_Class *sclass,
					   int start_position)
{
  ClassVariable *cvar;
  Scheme_Obj_Frame *frame;
  int i, num_all;
  Scheme_Object *box, **all_values;
  Scheme_Object **env_values, **public_values, **private_values;

  frame = (Scheme_Obj_Frame *)scheme_malloc(sizeof(Scheme_Obj_Frame));

  frame->type = scheme_object_frame_type;
  frame->sclass = sclass;
  frame->obj = obj;
  frame->initfunc = NULL;
  frame->unbound_handler = NULL;
  frame->inited = FALSE;

  frame->start_nexts = start_position;

  if (sclass->num_supers) {
    int pos;
    frame->num_nexts = sclass->num_supers;

    frame->nexts = obj->all_frames + start_position;
    
    pos = start_position + frame->num_nexts;

    for (i = 0; i < frame->num_nexts; i++) {
      frame->nexts[i] = BuildObjectFrames(obj, sclass->supers[i], pos);
      pos += sclass->supers[i]->num_all_supers;
    }
  } else {
    frame->num_nexts = 0;
    frame->nexts = NULL;
  }

  num_all = 2 * sclass->num_public + sclass->num_private;
  all_values = (Scheme_Object **)scheme_malloc(num_all*sizeof(Scheme_Object *));

  env_values = frame->env_values = all_values;
  public_values = frame->public_values = all_values + sclass->num_public;
  private_values = frame->private_values = all_values + 2 * sclass->num_public;

  for (cvar = sclass->pub; cvar; cvar = cvar->next) {
    i = cvar->pos;
    box = NULL;
    
    /* The FrameBox finds could be done at class creation time... */

    if (cvar->vartype == varINHERIT || cvar->vartype == varSHARE)
      box = LookupInherited(frame, cvar);
    else if (cvar->delay)
      box = (Scheme_Object *)cvar->delay;
    else
      box = scheme_make_envunbox(scheme_undef);

    if (frame->nexts)
      if (cvar->vartype != varINHERIT && cvar->vartype != varLOCAL)
	FrameBox(frame, cvar->name, 0, box, NULL, 0);
    
    env_values[i] = public_values[i] = box;
  }

  for (i = 0, cvar = sclass->pri; i < frame->sclass->num_private; 
       i++, cvar = cvar->next) {
    if (cvar->vartype != varINHERIT)
      box = scheme_make_envunbox(scheme_undef);
    else
      box = LookupInherited(frame, cvar);
    private_values[i] = box;
  }

  return frame;
}

static void CreateObjectFrameEnvironments(Scheme_Obj_Frame *frame)
{
  Scheme_Env *f;
  Scheme_Object *box;
  ClassVariable *cvar;
  int i, k, count;

  if (frame->sclass->priminit)
    return;

  for (i = 0; i < frame->num_nexts; i++)
    CreateObjectFrameEnvironments(frame->nexts[i]);

  count = frame->sclass->num_public + frame->sclass->num_private
    + 1 + frame->sclass->num_supers;
  
  f = frame->pri_env = scheme_new_special_frame(count, SCHEME_AUTO_UNBOX);
  scheme_extend_env(f, frame->sclass->env);

  k = 0;
  for (i = frame->num_nexts; i--; k++) {
    box = scheme_make_envunbox((Scheme_Object *)frame->sclass->supers[i]);
    scheme_add_good_binding(k, box, f);
  }

  box = scheme_make_envunbox((Scheme_Object *)frame->obj);
  scheme_add_good_binding(k, box, f);
  k++;

  for (i = frame->sclass->num_private; i--; k++)
    scheme_add_good_binding(k, frame->private_values[i], f);

  cvar = frame->sclass->pub;
  for (i = frame->sclass->num_public; i--; cvar = cvar->next)
    scheme_add_good_binding(k + i, frame->env_values[cvar->pos], f);


  f= frame->init_env = scheme_new_special_frame(2 * frame->num_nexts, 0);
  scheme_extend_env(f, frame->pri_env);
  

  for (k = 0, i = frame->num_nexts; i--; ) {
    box = MakeSuperInitPrim(frame, i, 0);
    scheme_add_good_binding(k, box, f);
    k++;
    box = MakeSuperInitPrim(frame, i, 1);
    scheme_add_good_binding(k, box, f);
    k++;
  }
}

static void CreateObjectFrames(Internal_Object *obj,
			       Scheme_Class *sclass)
{
  Scheme_Obj_Frame *frame;

  InsureClassReady(sclass);

  if (sclass->num_all_supers)
    obj->all_frames = 
      (Scheme_Obj_Frame **)scheme_malloc(sclass->num_all_supers
					 * sizeof(Scheme_Obj_Frame *));
  else
    obj->all_frames = NULL;

  frame = BuildObjectFrames(obj, sclass, 0);
  CreateObjectFrameEnvironments(frame);

  obj->frames = frame;
}

/**************************************************************************/

static void SetIVarValues(Scheme_Obj_Frame *frame, /* Frame */
			  Scheme_Env *close_env,/* Env to eval values */
			  ClassVariable *items,  /* Items to be created */
			  int pub)
{
  Scheme_Object *box, *v;
  int i;

  i = 0;
  while (items) {
    if (items->vartype != varINHERIT && items->vartype != varSHARE) {
      if (pub)
	box = frame->public_values[items->pos];
      else
	box = frame->private_values[i++];

      if (!SCHEME_DIVARP(box)) {
	v = SCHEME_PTR_VAL(box);
	if (SAME_OBJ(v, scheme_undef)) {
	  if (items->vartype == varPRIMITIVE_FUNC)
	    v = scheme_make_closed_prim(items->u.f, frame->obj);
	  else if (items->vartype == varVALUE)
	    v = items->u.value;
	  else
	    v = _scheme_eval_compiled(items->u.value, close_env);
	  
	  SCHEME_PTR_VAL(box) = v;
	}
      }
    }

    items = items->next;
  }
}

void InitObjectFrame(Scheme_Obj_Frame *frame)
{
  if (frame->inited)
    return;

  SetIVarValues(frame, frame->pri_env, frame->sclass->pri, 0);
  SetIVarValues(frame, frame->pri_env, frame->sclass->pub, 1);

  frame->private_values = NULL;

  frame->inited = TRUE;

  if (frame->sclass->priminit) {
    int i;
    for (i = 0; i < frame->num_nexts; i++)
      InitObjectFrame(frame->nexts[i]);
  }
}

static Scheme_Object *find_ivar(Internal_Object *obj, 
				Scheme_Object *name,
				Scheme_Class *sclass,
				int force)
{
  Scheme_Object *v;
  Scheme_Obj_Frame *frame = obj->frames;

  v = FrameBox(frame, name, 1, NULL, sclass, force);
  
  if (sclass && SAME_OBJ(v, ((Scheme_Object *)sclass))) {
    scheme_signal_error(IVAR ": object not in class (for ivar %s)",
			SCHEME_SYM_VAL(name));
    return NULL;
  }

  if (!force && SCHEME_DIVARP(v))
    return NULL;

  if (v)
    v = SCHEME_PTR_VAL(v);

  if (!v) {
    /* Look for handler: */
    Scheme_Object *handler;

    handler = FrameBox(frame, NULL, 1, NULL, sclass, 0);

    if (handler)
      return _scheme_tail_apply(handler, 1, &name);
  }

  return v;
}

Scheme_Object *scheme_find_ivar(Scheme_Object *obj, 
				Scheme_Object *sclass,
				Scheme_Object *sym,
				int force)
{
  if (!SCHEME_OBJP(obj))
    scheme_signal_error("ivar: not an object");
  if (sclass && !SCHEME_CLASSP(sclass))
    scheme_signal_error("ivar: not a class");
  if (!SCHEME_SYMBOLP(sym))
    scheme_signal_error("ivar: not a symbol");

  return find_ivar(((Internal_Object *)obj), sym,
		   (Scheme_Class *)sclass, force);
}

static Scheme_Class_Object *allocate_object(Scheme_Class *sclass)
{
  Internal_Object *obj;
  Scheme_Object *varlist, *varboxlist;
  int i, j, pos;

  if (!sclass)
    scheme_signal_error("bad primitive class");

  obj = (Internal_Object *)scheme_malloc(sizeof(Internal_Object));
  obj->o.sclass = (Scheme_Object *)sclass;

  obj->o.primdata = 0;
  obj->o.inited = 0;

  obj->o.type = scheme_object_type;

  varlist = varboxlist = scheme_null;
  CreateObjectFrames(obj, sclass);

  if (sclass->num_all_supers) {
    obj->all_frames = 
      (Scheme_Obj_Frame **)scheme_malloc(sclass->num_all_supers
					 * sizeof(Scheme_Obj_Frame *));
    
    for (i = 0; i < sclass->num_supers; i++)
      obj->all_frames[i] = obj->frames->nexts[i];

    pos = sclass->num_supers;
    for (i = 0; i < sclass->num_all_supers; i++) {
      Scheme_Obj_Frame *fr = obj->all_frames[i];
      for (j = 0; j < fr->num_nexts; j++)
	obj->all_frames[pos++] = fr->nexts[j];
    }
  } else
    obj->all_frames = NULL;

  return &obj->o;
}

Scheme_Object *scheme_make_uninited_object(Scheme_Object *c)
{
  Internal_Object *obj;
  Scheme_Class *sclass;

  sclass = (Scheme_Class *)c;

  obj = (Internal_Object *)allocate_object(sclass);
  InitObjectFrame(obj->frames);

  obj->o.inited = 1;

  return (Scheme_Object *)&obj->o;
}

static void CallInitFrame(Scheme_Obj_Frame *frame,
			  int argc, Scheme_Object *argv[])
{
  Scheme_Class *sclass;

  if (!frame) {
    if (argc)
      scheme_signal_error(SUPER_INIT ": too many arguments");
    else
      return;
  }

  sclass = frame->sclass;

  InitObjectFrame(frame);

  /* Eval intialization fucntion */
  if ((!sclass->priminit && sclass->init)
      || (sclass->priminit && sclass->initf)) {
    if (!sclass->priminit)
      frame->initfunc = _scheme_eval_compiled(sclass->init, frame->init_env);
    else
      frame->initfunc = scheme_make_closed_prim(sclass->initf, frame->obj);
  } else
    frame->initfunc = NULL;

  /* Eval ivar name handler */
  if (sclass->unbound_handler) {
    frame->unbound_handler = _scheme_eval_compiled(sclass->unbound_handler,
						   frame->pri_env);
    SCHEME_ASSERT(SCHEME_PROCP(frame->unbound_handler),
		  CREATE_OBJECT ": ivar name handler must be a closure");
  } else
    frame->unbound_handler = NULL;

  if (frame->initfunc) {
    int i;

    _scheme_apply(frame->initfunc, argc, argv);
    if (sclass->priminit)
      frame->obj->o.inited = 1;

    for (i = 0; i < frame->num_nexts; i++)
      if (!frame->nexts[i]->inited 
	  && NOT_SAME_OBJ(frame->nexts[i]->sclass, null_class))
	scheme_signal_error(CREATE_OBJECT ": superclass %s never intialized",
			    SCHEME_SYM_VAL(frame->sclass->supernames[i]));
  } else if (frame->num_nexts) {
    int i;

    for (i = 0; i < frame->num_nexts; i++)
      CallInitFrame(frame->nexts[i], argc, argv);
  } else if (argc)
    scheme_signal_error(SUPER_INIT ": too many arguments");
}

static Scheme_Class_Object *make_object(Scheme_Class *sclass,
					int argc,
					Scheme_Object *argv[])
{
  Internal_Object *obj;
  
  obj = (Internal_Object *)allocate_object(sclass);
  
  CallInitFrame(obj->frames, argc, argv);

  return &obj->o;
}

Scheme_Object *scheme_make_object(Scheme_Object *sclass, 
				  int argc, Scheme_Object **argv)
{
  if (!SCHEME_CLASSP(sclass))
    scheme_signal_error("make-object: not a class");

  return (Scheme_Object *)make_object((Scheme_Class *)sclass, argc, argv);
}

static Scheme_Object *CreateObject(int argc, Scheme_Object *argv[])
{
  Scheme_Object *classname;
  Scheme_Class *sclass;
  Internal_Object *obj;

  if (!argc)
    scheme_wrong_count(CREATE_OBJECT, 1, -1, argc, argv);

  classname = argv[0];

  if (!SAME_TYPE(SCHEME_TYPE(classname), scheme_class_type))
    scheme_wrong_type(CREATE_OBJECT, "class", 0, argc, argv);

  sclass = (Scheme_Class *)classname;

  if (sclass->priminit && !sclass->initf)
    scheme_signal_error("can't create objects from the class"
			" (%s)", sclass->classname);

  obj = (Internal_Object *)make_object(sclass, argc  - 1, argv + 1);
  
  return (Scheme_Object *)&obj->o;
}

/************************************************************************/

static Scheme_Object *DoSuperInitPrim(SuperInitData *data,
				      int argc, Scheme_Object **argv)
{
  Scheme_Obj_Frame *frame = data->frame;

  if (!frame->num_nexts)
    scheme_signal_error(SUPER_INIT ": use only when there is a superclass");

  if (frame->nexts[data->which]->inited)
    scheme_signal_error("multiple intializations of super class");

  if (data->apply) {
    int count, i;
    Scheme_Object *l;
    Scheme_Object **old = argv;

    if (!argc)
      scheme_signal_error("apply [superclass init]: must supply a proper list");

    l = argv[argc - 1];
    for (count = 0; SCHEME_PAIRP(l); count++)
      l = SCHEME_CDR(l);
    if (!SCHEME_NULLP(l))
      scheme_signal_error("apply [superclass init]: last arg not a proper list");

    argv = (Scheme_Object **)scheme_malloc((count + argc - 1) 
					   * sizeof(Scheme_Object *));

    for (i = 0; i < argc - 1; i++)
      argv[i] = old[i];
    argc += (count - 1);
    for (l = old[i]; i < argc; i++) {
      argv[i] = SCHEME_CAR(l);
      l = SCHEME_CDR(l);
    }
  }

  CallInitFrame(frame->nexts[data->which], argc, argv);

  return scheme_void;
}

static Scheme_Object *MakeSuperInitPrim(Scheme_Obj_Frame *fr, int which, 
					int apply)
{
  SuperInitData *data;

  data = (SuperInitData *)scheme_malloc(sizeof(SuperInitData));

  data->frame = fr;
  data->which = which;
  data->apply = apply;
  
  return scheme_make_closed_prim((Scheme_Closed_Prim *)DoSuperInitPrim, 
				 (void *)data);
}

/*****************************************************************************/

Scheme_Object *scheme_get_generic_data(Scheme_Object *cl, 
				       Scheme_Object *name)
{
  int i, superclass, framepos;
  Scheme_Class *sclass;
  Generic_Data *data;

  sclass = (Scheme_Class *)cl;

  framepos = FindName(sclass, name);
  if (framepos >= 0) {
    superclass = -1;
  } else {
    int count = sclass->num_all_supers;
    superclass = 0;
    for (i = 0; i < count; i++) {
      framepos = FindName(sclass->all_supers[i], name);
      if (framepos >= 0) {
	superclass = i;
	break;
      }
    }

    if (i >= count)
      return NULL;
  }

  data = (Generic_Data *)scheme_malloc(sizeof(Generic_Data));

  data->type = scheme_generic_data_type;
  data->sclass = sclass;
  data->superclass = superclass;
  data->framepos = framepos;

  return (Scheme_Object *)data;
}

Scheme_Object *scheme_apply_generic_data(Scheme_Object *gdata, 
					 Scheme_Object *sobj,
					 int force)
{
  Scheme_Obj_Frame *frame;
  Internal_Object *obj;
  int i;
  Scheme_Class *sclass;
  Generic_Data *data;
  Scheme_Object *v;

  data = (Generic_Data *)gdata;

  obj = (Internal_Object *)sobj;

  sclass = (Scheme_Class *)obj->o.sclass;
  
  if (NOT_SAME_OBJ(sclass, data->sclass)) {
    for (i = sclass->num_all_supers; i--; )
      if (SAME_OBJ(sclass->all_supers[i], data->sclass))
	break;
    
    if (i < 0) {
      scheme_signal_error("generic" ": object not in the generic's class");
      return scheme_void;
    }
  } else
    i = -1;

  if (data->superclass >= 0) {
    if (i < 0)
      frame = obj->all_frames[data->superclass];
    else
      frame = obj->all_frames[i + sclass->super_starts[data->superclass]];
  } else if (i < 0)
    frame = obj->frames;
  else
    frame = obj->all_frames[i];
  
  v = frame->env_values[data->framepos];
  if (SCHEME_DIVARP(v)) {
    if (!force)
      return NULL;
    ForceIvarValue(frame, data->framepos, v, 1);
    v = frame->env_values[data->framepos];
  }

  return SCHEME_PTR_VAL(v);
}

static Scheme_Object *DoGeneric(Generic_Data *data, 
				int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("generic", 1, 1, argc, argv);
  if (!SCHEME_OBJP(argv[0]))
    scheme_wrong_type("generic", "object", 0, argc, argv);

  return scheme_apply_generic_data((Scheme_Object *)data, argv[0], 1);
}

static Scheme_Object *MakeGeneric(int argc, Scheme_Object *argv[])
{
  Scheme_Object *data;

  if (argc != 2)
    scheme_wrong_count(MAKE_GENERIC, 2, 2, argc, argv);
  if (!SCHEME_CLASSP(argv[0]))
    scheme_wrong_type(MAKE_GENERIC, "class", 0, argc, argv);
  if (!SCHEME_SYMBOLP(argv[1]))
    scheme_wrong_type(MAKE_GENERIC, "symbol", 1, argc, argv);

  data = scheme_get_generic_data(argv[0], argv[1]);

  if (!data)
    scheme_signal_error(MAKE_GENERIC ": can't find instance variable %s",
			SCHEME_SYM_VAL(argv[1]));

  return scheme_make_closed_prim((Scheme_Closed_Prim *)DoGeneric, (void *)data);
}

/*****************************************************************************/

static Scheme_Object *GenIVar_Execute(Scheme_Object *data, Scheme_Env *env)
{
  Scheme_Object *a, *l, *name, *v;
  Internal_Object *obj = NULL;
  Scheme_Class *sclass = NULL;
  Scheme_Object *this_distance;
  int evalName;

  l = SCHEME_CAR(data);
  data = SCHEME_CDR(data);
  name = SCHEME_CAR(data);
  data = SCHEME_CDR(data);
  this_distance = SCHEME_CAR(data);
  evalName = SAME_OBJ(SCHEME_CDR(data), scheme_true);

  while (!SCHEME_NULLP(l)) {
    SCHEME_ASSERT(SCHEME_LISTP(l), IVAR ": syntax error");

    a = SCHEME_CAR(l);
    a = _scheme_eval_compiled(a, env);

    if (SAME_TYPE(SCHEME_TYPE(a), scheme_class_type)) {
      if (sclass)
	scheme_signal_error(IVAR ": two classes specified");
      sclass = (Scheme_Class *)a;
    } else if (SAME_TYPE(SCHEME_TYPE(a), scheme_object_type)) {
      if (obj)
	scheme_signal_error(IVAR ": two objects specified");
      obj = (Internal_Object *)a;
    } else
      scheme_signal_error(IVAR ": parameter not an object, class, or name"
			  " (it was a %s)", 
			  scheme_get_type_name(SCHEME_TYPE(a)));

    l = SCHEME_CDR(l);
  }

  if (evalName) {
    name = _scheme_eval_compiled(name, env);
    SCHEME_ASSERT(SCHEME_SYMBOLP(name), IVAR
		  ": instance variable name must be a symbol");
  }

  if (!obj) {
    v = scheme_lookup_value(this_distance, env);
    if (!v)
      scheme_signal_error(IVAR ": no object bound to \"this\"");
      
    if (NOT_SAME_TYPE(SCHEME_TYPE(v), scheme_object_type))
      scheme_signal_error(IVAR ":\"this\" bound to a non-object");
      
    obj = (Internal_Object *)v;
  }  

  v = find_ivar(obj, name, sclass, 1);

  if (!v)
    scheme_signal_error(IVAR ": instance variable %s not found",
			SCHEME_SYM_VAL(name));

  return v;
}

static Scheme_Object *GenIVar(Scheme_Object *form, Scheme_Env *env, 
			      Scheme_Compile_Info *rec, int evalName)
{
  Scheme_Object *l, *a, *name, *new_l = scheme_null, *eval_name;
  Scheme_Object *this_distance;
  Scheme_Compile_Info recs[10];
  int recpos = 0;

  l = SCHEME_CDR(form);

  SCHEME_ASSERT(SCHEME_LISTP(l), IVAR ": syntax error");
  SCHEME_ASSERT(!SCHEME_NULLP(l), IVAR
		": must specify an instance variable name");

  a = SCHEME_CAR(l);
  l = SCHEME_CDR(l);

  scheme_init_compile_recs(rec, recs, 10);

  while (!SCHEME_NULLP(l)) {
    SCHEME_ASSERT(SCHEME_LISTP(l), IVAR ": syntax error");

    SCHEME_ASSERT(recpos < 10, IVAR ": too many arguments");

    a = scheme_compile_expr(a, env, &recs[recpos++]);

    new_l = cons(a, new_l);

    a = SCHEME_CAR(l);
    l = SCHEME_CDR(l);
  }

  name = a;
  if (evalName) {
    SCHEME_ASSERT(recpos < 10, IVAR ": too many arguments");
    name = scheme_compile_expr(name, env, &recs[recpos++]);
  } else
    SCHEME_ASSERT(SCHEME_SYMBOLP(name), IVAR
		  ": instance variable name must be a symbol");

  scheme_merge_compile_recs(rec, recs, recpos);

  this_distance = scheme_static_distance(this_symbol, env, 0);

  eval_name = evalName ? scheme_true : scheme_false;

  return scheme_make_syntax_compilation(GenIVar_Execute,
					cons(new_l,
					     cons(name,
						  cons(this_distance,
						       eval_name))));
}

static Scheme_Object *IVar(Scheme_Object *form, Scheme_Env *env,
			   Scheme_Compile_Info *rec)
{
  return GenIVar(form, env, rec, FALSE);
}

static Scheme_Object *UQIVar(Scheme_Object *form, Scheme_Env *env,
			     Scheme_Compile_Info *rec)
{
  return GenIVar(form, env, rec, TRUE);
}

static Scheme_Object *IVarExpand_Recur(Scheme_Object *form, Scheme_Env *env,
				 int depth)
/* Expand all but last; quote last item */
{
  if (!SCHEME_PAIRP(form))
      scheme_signal_error(IVAR ": syntax error");

  if (SCHEME_NULLP(SCHEME_CDR(form)))
    return cons(cons(quote_symbol, form), scheme_null);

  return cons(scheme_expand_expr(SCHEME_CAR(form), env, depth),
	      IVarExpand_Recur(SCHEME_CDR(form), env, depth));
}

static Scheme_Object *Var_expand(Scheme_Object *name,
				 Scheme_Object *form, Scheme_Env *env,
				 int depth)
/* Expand all but last */
{
  return cons(name,
	      IVarExpand_Recur(SCHEME_CDR(form), env, depth));
}

static Scheme_Object *IVar_expand(Scheme_Object *form, Scheme_Env *env,
				  int depth)
{
  return Var_expand(uq_ivar_symbol, form, env, depth);
}

static Scheme_Object *UQCVar(int argc, Scheme_Object **argv)
{
  Scheme_Class *sclass;
  Scheme_Object *name, *v;
  ClassVariable *cvar;
  int i;

  if (argc != 2)
    scheme_wrong_count(CVAR, 2, 2, argc, argv);
  if (!SCHEME_CLASSP(argv[0]))
    scheme_wrong_type(CVAR, "class", 0, argc, argv);
  if (!SCHEME_SYMBOLP(argv[1]))
    scheme_wrong_type(CVAR, "symbol", 1, argc, argv);

  sclass = (Scheme_Class *)argv[0];
  name = argv[1];

  v = NULL;

  for (cvar = sclass->pub; cvar; cvar = cvar->next)
    if (SAME_OBJ(name, cvar->name) && cvar->vartype == varVALUE) {
      v = cvar->u.value;
      break;
    }

  if (!v)
    for (i = 0; i < sclass->num_all_supers; i++) {
      Scheme_Class *scl = sclass->all_supers[i];
      for (cvar = scl->pub; cvar; cvar = cvar->next)
	if (SAME_OBJ(name, cvar->name) && cvar->vartype == varVALUE) {
	  v = cvar->u.value;
	  break;
	}
    }

  if (!v)
    scheme_signal_error(CVAR ": class variable %s not found",
			SCHEME_SYM_VAL(name));
  return v;
}

/*****************************************************************************/

static Scheme_Object *IsObject(int c, Scheme_Object *p[])
{
  SCHEME_ASSERT(c==1, IS_OBJECT
		": expected one argument");

  return SAME_TYPE(SCHEME_TYPE(p[0]), scheme_object_type) ? 
    scheme_true : scheme_false;
}

static Scheme_Object *IsClass(int c, Scheme_Object *p[])
{
  if (c != 1)
    scheme_wrong_count(IS_CLASS, 1, 1, c, p);

  return SAME_TYPE(SCHEME_TYPE(p[0]), scheme_class_type) 
    ? scheme_true : scheme_false;
}

int scheme_is_subclass(Scheme_Object *c, Scheme_Object *base)
{
  Scheme_Class *sclass;
  int i;

  if (!SCHEME_CLASSP(c) || !SCHEME_CLASSP(base))
    return 0;

  if (SAME_OBJ(c, base))
    return 1;

  sclass = (Scheme_Class *)c;
  for (i = 0; i < sclass->num_all_supers; i++)
    if (SAME_OBJ((Scheme_Object *)sclass->all_supers[i], base))
      return 1;

  return 0;
}

static Scheme_Object *IsA(int c, Scheme_Object *p[])
{
  if (c != 2)
    scheme_wrong_count(IS_A, 2, 2, c, p);
  if (!SCHEME_CLASSP(p[1]))
    scheme_wrong_type(IS_A, "class", 1, c, p);

  return scheme_is_a(p[0], p[1]) ? scheme_true : scheme_false;
}

int scheme_is_a(Scheme_Object *obj, Scheme_Object *base)
{
  if (!(SAME_TYPE(SCHEME_TYPE(base), scheme_class_type)))
    return 0;

  if (NOT_SAME_TYPE(SCHEME_TYPE(obj),  scheme_object_type))
    return 0;

  return (scheme_is_subclass((Scheme_Object *)
			     ((Internal_Object *)obj)->o.sclass, base));
}

static Scheme_Object *IsSubclass(int c, Scheme_Object *p[])
{
   if (c != 2)
    scheme_wrong_count(SUBCLASS, 2, 2, c, p);
  if (!SCHEME_CLASSP(p[1]))
    scheme_wrong_type(SUBCLASS, "class", 1, c, p);

  if (NOT_SAME_TYPE(SCHEME_TYPE(p[0]), scheme_class_type))
    return scheme_false;

  return scheme_is_subclass(p[0], p[1]) ? scheme_true : scheme_false;
}

static Scheme_Object *CollectGarbage(int c, Scheme_Object *p[])
{
  if (c)
    scheme_wrong_count(COLLECT_GARBAGE, 0, 0, c, p);

  GC_gcollect();

  return scheme_void;
}

void scheme_init_object(Scheme_Env *env)
{
  if (scheme_starting_up) {
    
    scheme_undef = scheme_alloc_eternal_object();
    scheme_undef->type = scheme_undefined_type;

    inherited_filler = scheme_alloc_eternal_object();
    inherited_filler->type = scheme_inherit_filler_type;

    init_symbol = scheme_intern_symbol("init");
    
    pub_symbol = scheme_intern_symbol("public");
    local_symbol = scheme_intern_symbol("local");
    pri_symbol = scheme_intern_symbol("private");
    static_symbol = scheme_intern_symbol("class");
    inh_symbol = scheme_intern_symbol("inherit");
    inh_from_symbol = scheme_intern_symbol("inherit-from");
    ren_symbol = scheme_intern_symbol("rename");
    ren_from_symbol = scheme_intern_symbol("rename-from");
    share_symbol = scheme_intern_symbol("share");
    share_from_symbol = scheme_intern_symbol("share-from");
    
    this_symbol = scheme_intern_symbol("this");
    super_symbol = scheme_intern_symbol("super");

    quote_symbol = scheme_intern_symbol("quote");

    uq_ivar_symbol = scheme_intern_symbol(UQ_IVAR);

    scheme_register_syntax_executer("-dc", DefineClass_Execute);
    scheme_register_syntax_executer("-iv", GenIVar_Execute);

    scheme_install_type_writer(scheme_class_data_type, write_Class_Data);
    scheme_install_type_reader(scheme_class_data_type, read_Class_Data);
  }

  scheme_add_global_keyword(DEFINE_CLASS, 
			    scheme_make_compiled_syntax(DefineClass, 
							DefineClass_expand), 
			    env);
  scheme_add_global_constant(CREATE_OBJECT, 
			     scheme_make_prim_w_arity(CreateObject, 1, -1), 
			     env);
  
  scheme_add_global_constant(IVAR, 
			     scheme_make_compiled_syntax(IVar, IVar_expand), 
			     env);
  scheme_add_global_keyword(UQ_IVAR, 
			    scheme_make_compiled_syntax(UQIVar, 
							scheme_expand_simple),
			    env);

  scheme_add_global_constant(UQ_CVAR, 
			     scheme_make_prim_w_arity(UQCVar, 2, 2), 
			     env);

  scheme_add_global_constant(UQ_MAKE_GENERIC, 
			     scheme_make_prim_w_arity(MakeGeneric, 2, 2), 
			     env);

  scheme_add_global_constant(IS_OBJECT, 
			     scheme_make_prim_w_arity(IsObject, 1, 1), 
			     env);
  scheme_add_global_constant(IS_CLASS, 
			     scheme_make_prim_w_arity(IsClass, 1, 1), 
			     env);
  scheme_add_global_constant(IS_A, 
			     scheme_make_prim_w_arity(IsA, 2, 2), 
			     env);
  scheme_add_global_constant(SUBCLASS, 
			     scheme_make_prim_w_arity(IsSubclass, 2, 2), 
			     env);

  scheme_add_global_constant(COLLECT_GARBAGE, 
			     scheme_make_prim_w_arity(CollectGarbage, 0, 0), 
			     env);
}
