/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"
#include <time.h>
#ifdef FILES_HAVE_FDS
#include <sys/types.h>
#include <sys/time.h>
#endif

Scheme_Process *scheme_current_process = NULL;
Scheme_Process *scheme_first_process = NULL;
Scheme_Process **scheme_current_process_ptr;

void (*scheme_sleep)(int seconds, void *fds);
void (*scheme_notify_multithread)(int on);
void (*scheme_wakeup_on_input)(void *fds);

static int scheme_do_atomic = 0;
static int checking_break = 0;
static int have_activity = 0;
int scheme_active_but_sleeping = 0;
static int process_ended_with_activity;

static Scheme_Trigger_Callback *trigger_callbacks = NULL;
static Scheme_Trigger_From_Port *port_triggers = NULL;

static Scheme_Object *sch_spawn(int argc, Scheme_Object *args[]);
static Scheme_Object *sch_thread(int argc, Scheme_Object *args[]);
static Scheme_Object *sch_sleep(int argc, Scheme_Object *args[]);
static Scheme_Object *sch_atomic(int argc, Scheme_Object *args[]);
static Scheme_Object *sch_current(int argc, Scheme_Object *args[]);
static Scheme_Object *process_read_port(int argc, Scheme_Object *args[]);
static Scheme_Object *process_readerr_port(int argc, Scheme_Object *args[]);
static Scheme_Object *process_write_port(int argc, Scheme_Object *args[]);
static Scheme_Object *process_weight(int argc, Scheme_Object *args[]);
static Scheme_Object *set_process_weight(int argc, Scheme_Object *args[]);

#ifndef PROCESS_STACK_SIZE
#define PROCESS_STACK_SIZE 300000
#endif

#ifdef SPAWN_NEW_STACK
#define SETJMP(p) setjmp(p->jmpup_buf.buf)
#define LONGJMP(p) scheme_longjmp(p->jmpup_buf.buf, 1)
#else /* not SPAWN_NEW_STACK */
#define SETJMP(p) scheme_setjmpup(&p->jmpup_buf, p->stack_start)
#define LONGJMP(p) scheme_longjmpup(&p->jmpup_buf)
#endif

void scheme_init_process(Scheme_Env *env)
{
  scheme_add_global_constant("spawn-eval",
			     scheme_make_prim_w_arity(sch_spawn, 1, 1), 
			     env);
  scheme_add_global_constant("thread", 
			     scheme_make_prim_w_arity(sch_thread, 1, 1), 
			     env);
  
  scheme_add_global_constant("sleep", 
			     scheme_make_prim_w_arity(sch_sleep, 0, 1), 
			     env);
  
  scheme_add_global_constant("atomic",
			     scheme_make_prim_w_arity(sch_atomic, 1, 1), 
			     env);
  
  scheme_add_global_constant("current-process", 
			     scheme_make_prim_w_arity(sch_current, 0, 0), 
			     env);
  
  scheme_add_global_constant("process-read-port", 
			     scheme_make_prim_w_arity(process_read_port, 
						      1, 1), 
			     env);
  scheme_add_global_constant("process-readerr-port", 
			     scheme_make_prim_w_arity(process_readerr_port, 
						      1, 1), 
			     env);
  scheme_add_global_constant("process-write-port", 
			     scheme_make_prim_w_arity(process_write_port, 
						      1, 1), 
			     env);

  scheme_add_global_constant("set-process-weight!",
			     scheme_make_prim_w_arity(set_process_weight, 
						      2, 2), 
			     env);
  scheme_add_global_constant("process-weight",
			     scheme_make_prim_w_arity(process_weight, 1, 1), 
			     env);
}

Scheme_Process *scheme_make_process()
{
  Scheme_Process *process;

  process = (Scheme_Process *)scheme_malloc(sizeof(Scheme_Process));

  process->type = scheme_process_type;

  if (!scheme_current_process) {
    scheme_current_process = process;
    scheme_first_process = process;
    process->next = NULL;

    scheme_current_process_ptr = &scheme_current_process;
  } else {
    process->next = scheme_first_process;
    scheme_first_process = process;
  }

  process->engine_weight = 10000;
  
  process->check_for_break = NULL;
  process->user_break_poll_proc = NULL;
  process->enable_break = 0;

  process->exit_handler = NULL;

  process->evaluator = scheme_standard_evaluator;
  process->eval_handler = NULL;

  scheme_init_jmpup_buf(&process->jmpup_buf);

  process->running = 1;
  process->stack_overflow = 0;

  process->dw = NULL;

  process->block_descriptor = NOT_BLOCKED;

  process->env_stack = scheme_alloc_env_stack(NULL);

  return process;
}



void scheme_swap_process(Scheme_Process *new_process)
{
  if (SETJMP(scheme_current_process)) {
    /* We're back! */
  } else {
    /* We're leaving... */
    scheme_current_process = new_process;
    LONGJMP(scheme_current_process);
  }
}

#ifdef SPAWN_NEW_STACK
static jmp_buf start_escape;
#endif

static void start_child(Scheme_Process *child, 
			Scheme_Process *return_to_process,
			int separate, Scheme_Object *child_eval)
{
  Scheme_Process *p;

  if (SETJMP(child)) {
    if (return_to_process)
      scheme_swap_process(return_to_process);

    if (!setjmp(scheme_error_buf)) {
      if (separate)
	scheme_eval(child_eval, scheme_env);
      else
	scheme_apply(child_eval, 0, NULL);
    }

    scheme_current_process->running = 0;

    scheme_close_input_port(scheme_current_process->in_port);
    scheme_close_output_port(scheme_current_process->out_port);
    scheme_close_output_port(scheme_current_process->err_port);

    if (scheme_current_process == scheme_first_process) {
      scheme_first_process = scheme_current_process->next;
    } else {
      p = scheme_first_process;
      while (p && p->next != scheme_current_process)
	p = p->next;
      if (p)
	p->next = scheme_current_process->next;
    }

    process_ended_with_activity = 1;

    if (scheme_notify_multithread && !scheme_first_process->next
	&& !trigger_callbacks & !port_triggers) {
      scheme_notify_multithread(0);
      have_activity = 0;
    }

    if (scheme_current_process->next)
      scheme_swap_process(scheme_current_process->next);
    else
      scheme_swap_process(scheme_first_process);

    /* Shouldn't get here! */
    scheme_signal_error("bad process switch");
  }
#ifdef SPAWN_NEW_STACK
  longjmp(start_escape, 1);
#endif
}

#ifdef SPAWN_NEW_STACK

static Scheme_Process *sc_child, *sc_rtp;
static int sc_sep;
static Scheme_Object *sc_eval;

static void do_start_child(Scheme_Process *child, Scheme_Process *rtp,
			   int separate, Scheme_Object *child_eval)
{
  if (setjmp(start_escape))
    return;

  sc_child = child;
  sc_rtp = rtp;
  sc_sep = separate;
  sc_eval = child_eval;

#ifdef __BORLANDC__
  _ESP = (long)child->stack_start;
#else
  -> cannot use SPAWN_NEW_STACK
#endif

  start_child(sc_child, sc_rtp, sc_sep, sc_eval);
}
#endif

static Scheme_Object *make_subprocess(Scheme_Object *child_eval, int separate,
				      void *child_start)
{
  Scheme_Process *child, *return_to_process;
  int turn_on_multi;

  turn_on_multi = !scheme_first_process->next && !trigger_callbacks
    && !port_triggers;

  scheme_insure_stack_start(scheme_current_process, child_start);

  child = scheme_make_process();

#ifdef SPAWN_NEW_STACK
  child->stack_start = (void *)scheme_malloc(PROCESS_STACK_SIZE);
  child->stack_end = ((char *)child->stack_start) + PROCESS_STACK_SIZE;
#else
  child->stack_start = child_start;
#endif

  scheme_pipe(&child->in_port, &child->write_port);
  scheme_pipe(&child->read_port, &child->out_port);
  scheme_pipe(&child->readerr_port, &child->err_port);

  if (separate) {
    scheme_top_level_env(child);

    child->stdin_port = child->in_port;
    child->stdout_port = child->out_port;
    child->stderr_port = child->err_port;
  } else {
    child->env = scheme_current_process->env;
    child->globals = scheme_current_process->globals;

    child->stdin_port = scheme_current_process->stdin_port;
    child->stdout_port = scheme_current_process->stdout_port;
    child->stderr_port = scheme_current_process->stderr_port; 

    child->check_for_break = scheme_current_process->check_for_break;
    child->user_break_poll_proc = scheme_current_process->user_break_poll_proc;
    child->enable_break = scheme_current_process->enable_break;

    child->error_handler_proc = scheme_current_process->error_handler_proc;

    child->exit_handler = scheme_current_process->exit_handler;

    child->evaluator = scheme_current_process->evaluator;
    child->eval_handler = scheme_current_process->eval_handler;
  }

  if (scheme_do_atomic)
    return_to_process = scheme_current_process;
  else
    return_to_process = NULL;

#ifdef SPAWN_NEW_STACK
  do_start_child(child, return_to_process, separate, child_eval);
#else
  start_child(child, return_to_process, separate, child_eval);
#endif

  if (scheme_notify_multithread && turn_on_multi) {
    scheme_notify_multithread(1);
    have_activity = 1;
  }

  scheme_swap_process(child);
  
  return (Scheme_Object *)child;
}

Scheme_Object *scheme_spawn(Scheme_Object *expr)
{
  long dummy;

  return make_subprocess(expr, 1, (void *)&dummy);
}

Scheme_Object *scheme_thread(Scheme_Object *thunk)
{
  long dummy;

  return make_subprocess(thunk, 0, (void *)&dummy);
}

void scheme_add_trigger_callback(Scheme_Trigger_Callback *cb)
{
  cb->next = trigger_callbacks;
  trigger_callbacks = cb;
  if (!scheme_first_process->next && !port_triggers) {
    scheme_notify_multithread(1);
    have_activity = 1;
  }
}

void scheme_add_trigger_from_port(Scheme_Trigger_From_Port *pt)
{
  pt->next = port_triggers;
  port_triggers = pt;
  if (!scheme_first_process->next && !trigger_callbacks) {
    scheme_notify_multithread(1);
    have_activity = 1;
  }
}

static void port_trigger_check(void) {

  Scheme_Trigger_From_Port *pt, *pprev;

  pt = port_triggers;
  pprev = NULL;
  while (pt) {
    if (scheme_char_ready(pt->port)) {
      if (pprev)
	pprev->next = pt->next;
      else {
	port_triggers = pt->next;
	if (!port_triggers
	    && !scheme_first_process->next
	    && !trigger_callbacks) {
	  scheme_notify_multithread(0);
	  have_activity = 0;
	}
      }

      pt->trigger->value++;
    } else
      pprev = pt;
    pt = pt->next;
  }
}

int check_sleep(int need_activity, int sleep_now)
{
  Scheme_Process *p, *p2;
  Scheme_Trigger_From_Port *pt;
  int end_with_act;

#ifdef FILES_HAVE_FDS
  fd_set set;
#endif
  void *fds;

  /* Is everything blocked? */
  if (!scheme_do_atomic) {
    p = scheme_first_process;
    while (p->next) {
      if (p->ran_some || p->block_descriptor == NOT_BLOCKED)
	break;
      p = p->next;
    }
  }

  p2 = scheme_first_process;
  while (p2) {
    p2->ran_some = 0;
    p2 = p2->next;
  }

  end_with_act = process_ended_with_activity;
  process_ended_with_activity = 0;

  if (need_activity && !end_with_act && 
      (scheme_do_atomic 
       || (!p->next && SAME_OBJ(p, scheme_current_process)
	   && ((!sleep_now && scheme_wakeup_on_input)
	       || (sleep_now && scheme_sleep))))) {
    /* Poll from top-level process, and all subprocesses are blocked. */
    /* So, everything is blocked pending external input. */
    /* Build a list of file descriptors that we're waiting on */
    /* and turn off polling. */
    if (have_activity)
      scheme_active_but_sleeping = 1;
    if (have_activity && scheme_notify_multithread)
      scheme_notify_multithread(0);

#ifdef FILES_HAVE_FDS
    fds = (void *)&set;
    FD_ZERO(&set);
#else
    fds = NULL;
#endif

    p = scheme_first_process;
    while (p) {
      if (p->block_descriptor == PORT_BLOCKED)
	scheme_need_wakeup(p->blocker, fds);
      p = p->next;
    }

    pt = port_triggers;
    while (pt) {
      scheme_need_wakeup(pt->port, fds);
      pt = pt->next;
    }

    if (sleep_now)
      scheme_sleep(0, fds);
    else
      scheme_wakeup_on_input(fds);
    
    return 1;
  }

  return 0;
}

void scheme_check_threads(void)
{
  Scheme_Trigger_Callback *cb, *prev;
  int save;
  jmp_buf savebuf;

  port_trigger_check();

  cb = trigger_callbacks;
  while (cb) {
    if (cb->start != cb->trigger->value) {
      /* Remove this callback record from the list: */
      if (NOT_SAME_PTR(cb, trigger_callbacks)) {
	prev = trigger_callbacks;
	while (prev) {
	  if (SAME_PTR(prev->next, cb)) {
	    prev->next = cb->next;
	    break;
	  } else
	    prev = prev->next;
	}
      } else {
	trigger_callbacks = cb->next;
	if (!trigger_callbacks 
	    && !scheme_first_process->next
	    && !port_triggers) {
	  scheme_notify_multithread(0);
	  have_activity = 0;
	}
      }

      memcpy(&savebuf, &scheme_error_buf, sizeof(jmp_buf));
      if (!setjmp(scheme_error_buf))
	scheme_apply(cb->callback, 0, NULL);
      memcpy(&scheme_error_buf, &savebuf, sizeof(jmp_buf));
    }
    cb = cb->next;
  }

  save = scheme_current_process->enable_break;
  scheme_current_process->enable_break = 0;
  scheme_process_block(0);
  scheme_current_process->enable_break = save;

  check_sleep(have_activity, 0);
}

void scheme_wake_up(void)
{
  scheme_active_but_sleeping = 0;
  if (have_activity && scheme_notify_multithread)
    scheme_notify_multithread(1);
}

void scheme_process_block(int sleep_time)
{
  time_t start, d;

  if (checking_break)
    return;

  if (scheme_active_but_sleeping)
    scheme_wake_up();

  if (sleep_time)
    start = time(NULL);

  if (scheme_current_process->enable_break) {
    if (scheme_current_process->user_break_poll_proc) {
      int enabled;
      Scheme_Object *proc, *v;
      enabled = scheme_current_process->enable_break;
      proc = scheme_current_process->user_break_poll_proc;
      checking_break = 1;
      v = scheme_apply(proc, 0, NULL);
      checking_break = 0;
      if (NOT_SAME_OBJ(v, scheme_false)) {
	scheme_current_process->enable_break = enabled;
	scheme_signal_error("user break");
      }
      scheme_current_process->enable_break = enabled;
    } else if (scheme_current_process->check_for_break 
	       && scheme_current_process->check_for_break())
      scheme_signal_error("user break");
  }

 swap_or_sleep:
  port_trigger_check();

  if (!scheme_do_atomic && scheme_current_process->next)
    scheme_swap_process(scheme_current_process->next);
  else if (!scheme_do_atomic && scheme_current_process!=scheme_first_process) {
    /* This is the main process */
    scheme_insure_stack_start(scheme_current_process, (void *)&start);

    scheme_swap_process(scheme_first_process);
  } else if (sleep_time && scheme_sleep)
    scheme_sleep(sleep_time, NULL);

  if (!scheme_current_process->next) {
    /* If the main process is blocked, better check the rest: */
    if (scheme_current_process->block_descriptor != NOT_BLOCKED)
      check_sleep(1, 1);
  }

  if (sleep_time) {
    d = time(NULL) - start;
    if (d < 0)
      d = -d;
    if (d < (sleep_time - 1))
      goto swap_or_sleep;
  }
}

static Scheme_Object *sch_spawn(int argc, Scheme_Object *args[])
{
  if (argc != 1)
    scheme_wrong_count("spawn-eval", 1, 1, argc, args);

  return scheme_spawn(args[0]);
}

static Scheme_Object *sch_thread(int argc, Scheme_Object *args[])
{
  if (argc != 1)
    scheme_wrong_count("thread", 1, 1, argc, args);
  if (!SCHEME_PROCP(args[0]))
    scheme_wrong_type("thread", "procedure", 0, argc, args);

  return scheme_thread(args[0]);
}

static Scheme_Object *sch_sleep(int argc, Scheme_Object *args[])
{
  int t;

  if (argc && argc != 1)
    scheme_wrong_count("sleep", 0, 1, argc, args);
  if (argc && !SCHEME_INTP(args[0]))
    scheme_wrong_type("sleep", "integer", 0, argc, args);

  if (argc)
    t = SCHEME_INT_VAL(args[0]);
  else
    t = 0;

  scheme_process_block(t);

  scheme_fuel_counter = scheme_current_process->engine_weight;

  return scheme_void;
}

static void pre_atomic(void *v)
{
  scheme_do_atomic++;
}

static void post_atomic(void *v)
{
  --scheme_do_atomic;
}

static Scheme_Object *do_atomic(void *f)
{ 
  return _scheme_apply((Scheme_Object *)f, 0, NULL);
}

static Scheme_Object *sch_atomic(int argc, Scheme_Object *args[])
{
  if (argc != 1)
    scheme_wrong_count("atomic", 1, 1, argc, args);
  if (!SCHEME_PROCP(args[0]))
    scheme_wrong_type("atomic", "procedure", 0, argc, args);

  return scheme_dynamic_wind(pre_atomic, do_atomic, post_atomic, NULL,args[0]);
}

static Scheme_Object *sch_current(int argc, Scheme_Object *args[])
{
  if (argc)
    scheme_wrong_count("current-process", 0, 0, argc, args);

  return (Scheme_Object *)scheme_current_process;
}

static Scheme_Object *process_read_port(int argc, Scheme_Object *args[])
{
  if (argc != 1)
    scheme_wrong_count("process-read-port", 1, 1, argc, args);
  if (!SAME_TYPE(SCHEME_TYPE(args[0]), scheme_process_type))
    scheme_wrong_type("process-read-port", "process", 0, argc, args);

  if (!((Scheme_Process *)args[0])->read_port)
    scheme_signal_error("process-read-port: top-level process does not"
			" have a read port");

  return ((Scheme_Process *)args[0])->read_port;
}

static Scheme_Object *process_readerr_port(int argc, Scheme_Object *args[])
{
  if (argc != 1)
    scheme_wrong_count("process-readerr-port", 1, 1, argc, args);
  if (!SAME_TYPE(SCHEME_TYPE(args[0]), scheme_process_type))
    scheme_wrong_type("process-readerr-port", "process", 0, argc, args);

  if (!((Scheme_Process *)args[0])->readerr_port)
    scheme_signal_error("process-readerr-port: top-level process does not"
			" have a readerr port");

  return ((Scheme_Process *)args[0])->readerr_port;
}

static Scheme_Object *process_write_port(int argc, Scheme_Object *args[])
{
  if (argc != 1)
    scheme_wrong_count("process-write-port", 1, 1, argc, args);
  if (!SAME_TYPE(SCHEME_TYPE(args[0]), scheme_process_type))
    scheme_wrong_type("process-write-port", "process", 0, argc, args);

  if (!((Scheme_Process *)args[0])->write_port)
    scheme_signal_error("process-write-port: top-level process does not"
			" have a write port");

  return ((Scheme_Process *)args[0])->write_port;
}

static Scheme_Object *process_weight(int argc, Scheme_Object *args[])
{
  if (argc != 1)
    scheme_wrong_count("process-weight", 1, 1, argc, args);
  if (!SAME_TYPE(SCHEME_TYPE(args[0]), scheme_process_type))
    scheme_wrong_type("process-weight", "process", 0, argc, args);

  return scheme_make_integer(((Scheme_Process *)args[0])->engine_weight);
}

static Scheme_Object *set_process_weight(int argc, Scheme_Object *args[])
{
  int v;
  Scheme_Process *p;

  if (argc != 2)
    scheme_wrong_count("set-process-weight!", 2, 2, argc, args);
  if (!SAME_TYPE(SCHEME_TYPE(args[0]), scheme_process_type))
    scheme_wrong_type("set-process-weight!", "process", 0, argc, args);
  if (!SCHEME_INTP(args[1]) && !SCHEME_DBLP(args[1]))
    scheme_wrong_type("set-process-weight!", "number", 1, argc, args);

  v = SCHEME_NUM_VAL(args[1]);
  p = (Scheme_Process *)args[0];

  if (v <= 0)
    scheme_signal_error("set-process-weight!: "
			"process weight must be positive");

  p->engine_weight = v;
  if (p == scheme_current_process && scheme_fuel_counter > v)
    scheme_fuel_counter = v;

  return scheme_void;
}

