/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"
#include <string.h>
#include <ctype.h>

#define HASH_TABLE_SIZE 10313
static Scheme_Hash_Table *symbol_table = NULL;

/* globals */
Scheme_Object *scheme_quote_symbol;
Scheme_Object *scheme_quasiquote_symbol;
Scheme_Object *scheme_unquote_symbol;
Scheme_Object *scheme_unquote_splicing_symbol;

int scheme_case_sensitive;

/* locals */
static Scheme_Object *symbol_p_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *string_to_symbol_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *symbol_to_string_prim (int argc, Scheme_Object *argv[]);
static Scheme_Object *defined(int argc, Scheme_Object *argv[]);
static Scheme_Object *gensym(int argc, Scheme_Object *argv[]);

static int gensym_counter;

void
scheme_init_symbol_table ()
{
  if (scheme_starting_up)
    symbol_table =scheme_hash_table(HASH_TABLE_SIZE, SCHEME_hash_symbol, 0, 1);
}

void
scheme_init_symbol_type (Scheme_Env *env)
{
}

void
scheme_init_symbol (Scheme_Env *env)
{
  if (scheme_starting_up) {
    scheme_quote_symbol = scheme_intern_symbol ("quote");
    scheme_quasiquote_symbol = scheme_intern_symbol ("quasiquote");
    scheme_unquote_symbol = scheme_intern_symbol ("unquote");
    scheme_unquote_splicing_symbol = scheme_intern_symbol ("unquote-splicing");
  }

  scheme_add_global_constant ("symbol?", scheme_make_prim (symbol_p_prim), env);
  scheme_add_global_constant ("string->symbol", scheme_make_prim (string_to_symbol_prim), env);
  scheme_add_global_constant ("symbol->string", scheme_make_prim (symbol_to_string_prim), env);
  
  scheme_add_global_constant ("defined?", scheme_make_prim(defined), env);

  scheme_add_global_constant ("gensym", scheme_make_prim(gensym), env);
}

Scheme_Object *
scheme_make_a_symbol(const char *name, int forever)
{
  Scheme_Symbol *sym;
  int len;

  /* Symbol really only needs one field: */

  len = strlen(name);

  if (forever)
    sym = (Scheme_Symbol *)scheme_malloc_eternal(sizeof(Scheme_Type) +
						 len + 1);
  else
    sym = (Scheme_Symbol *)scheme_malloc_atomic(sizeof(Scheme_Type) + len + 1);

  sym->type = scheme_symbol_type;
  memcpy(sym->s, name, len + 1);
    
  return (Scheme_Object *)(sym);
}

Scheme_Object *
scheme_make_symbol (const char *name)
{
  return scheme_make_a_symbol(name, 0);
}

Scheme_Object *
scheme_intern_exact_symbol(const char *name)
{
  Scheme_Object *sym;

  sym = (Scheme_Object *)scheme_lookup_in_table (symbol_table, name);

  if (sym)
    return (sym);
  else {
    sym = scheme_make_a_symbol(name, 1);
    scheme_add_to_table (symbol_table, SCHEME_SYM_VAL(sym), sym, 0);
    return (sym);
  }
}

#define MAX_SYMBOL_SIZE 256

Scheme_Object *
scheme_intern_symbol(const char *name)
{
  if (!scheme_case_sensitive) {
    long i, len;
    char *naya;
    char on_stack[MAX_SYMBOL_SIZE];
    
    len = strlen(name);
    if (len >= MAX_SYMBOL_SIZE)
      naya = scheme_malloc_atomic(len + 1);
    else
      naya = on_stack;

    for (i = 0; i < len; i++) {
      int c = name[i];

      if (isupper(c))
	c = tolower(c);

      naya[i] = c;
    }

    naya[len] = 0;

    return scheme_intern_exact_symbol(naya);
  }

  return scheme_intern_exact_symbol(name);
}

/* locals */

static Scheme_Object *
symbol_p_prim (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("symbol?", 1, 1, argc, argv);
  return (SCHEME_SYMBOLP(argv[0]) ? scheme_true : scheme_false);
}

static Scheme_Object *
string_to_symbol_prim (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("string->symbol", 1, 1, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("string->symbol", "string", 0, argc, argv);
  return (scheme_intern_exact_symbol (SCHEME_STR_VAL(argv[0])));
}

static Scheme_Object *
symbol_to_string_prim (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("symbol->string", 1, 1, argc, argv);
  if (!SCHEME_SYMBOLP(argv[0]))
    scheme_wrong_type("symbol->string", "symbol", 0, argc, argv);
  
  return (scheme_make_string (SCHEME_SYM_VAL(argv[0])));
}

static Scheme_Object *
defined(int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("defined?", 1, 1, argc, argv);
  if (!SCHEME_SYMBOLP(argv[0]))
    scheme_wrong_type("defined?", "symbol", 0, argc, argv);

  return scheme_lookup_global(argv[0], scheme_env) ? scheme_true:scheme_false;
}

static Scheme_Object *gensym(int argc, Scheme_Object *argv[])
{
  char buffer[100], *str;

  if (argc && argc != 1)
    scheme_wrong_count("gensym", 0, 1, argc, argv);
  
  if (argc && !SCHEME_SYMBOLP(argv[0]) && SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("defined?", "symbol/string", 0, argc, argv);
  
  if (argc) {
    if (SCHEME_STRINGP(argv[0]))
      str = SCHEME_STR_VAL(argv[0]);
    else
      str = SCHEME_SYM_VAL(argv[0]);      
    sprintf(buffer, "%.80s%d", str, gensym_counter++);
  } else
    sprintf(buffer, "g%d", gensym_counter++);

  return scheme_make_symbol(buffer);
}
