; "(constant x)" is simpler than "(constant-symbol 'x)"
(defmacro constant (symbol)
  `(constant-name (quote ,symbol)))

; The following macro is useful for redefining a global in terms of
; it's current value. For example:
;  (define f (lambda () 1))
;  (redefine f (lambda () (+ (old-f) 1)))
; defines and redefines f. The "old-f" variable comes from the macro -
; which generates the name by adding an "old" the the original name -
; and old-f is protected in the sense that no one can change old-f with 
; a define or set!.
(defmacro redefine (global newval)
  (let ((old (string->symbol 
	      (string-append "old-" (symbol->string global)))))
    `(let ((,old ,global))
       (set! ,global ,newval))))
(constant redefine)

; let form where, instead of a list of name-value pairs, just
; provide a list of names and they are bound to a sequence of
; numbers (starting with 1)
(defmacro let-enumerate (name-lists . body)
  (cons 'let
	(cons
	 (let enum-all ((lists name-lists))
	   (if (null? lists)
	       '()
	       (append (let enum ((names (car lists)) (pos 1))
			 (if (null? names)
			     '()
			     (cons (list (car names) pos)
				   (enum (cdr names) (+ pos 1)))))
		       (enum-all (cdr lists)))))
	 body)))
(constant let-enumerate)

(defmacro catch-error (handler . body)
  (let ([orig-error (gensym)]
	[err (gensym)]
	[my-error (gensym)]
	[this-handler (gensym)])
    `(let* ([,orig-error (error-handler)]
	    [,this-handler ,handler])
       (dynamic-wind
	(lambda () #f)
	(lambda ()
	  (catch ,err
		 (let ([,my-error (lambda objs
				    (apply ,this-handler (cons ,err objs))
				    (apply ,orig-error objs))])
		    (error-handler ,my-error)
		    ,@body)))
	(lambda () (error-handler ,orig-error))))))
(constant catch-error)

(defmacro send* (obj . msgs)
  `(begin
     ,@(map (lambda (msg)
	      (cons 'send (cons obj msg)))
	    msgs)))
(constant send*)

; WARNING: This definition of package depends on internal defines being
;  executed in the order in which they are specified
(defmacro package (export-list . body)
  (let ([gensyms
	 (map (lambda (x) (gensym)) export-list)])
  `(begin
     ,@(map (lambda (name)
	      `(define ,name (void)))
	    export-list)
     (apply
      (lambda ,gensyms
	,@(if (null? export-list)
	      '((void))
	      (map (lambda (name gensym)
		     `(set! ,name ,gensym))
		   export-list gensyms)))
      (let ()
	,@body
	(list ,@export-list))))))
(constant package)

(defmacro opt-lambda (args . body)
  (let* ([f (gensym)]
	[required
	 (let loop ([args args])
	   (if (and (pair? args)
		    (symbol? (car args)))
	       (cons (car args) (loop (cdr args)))
	       '()))]
	[not-required-with-defaults
	 (let loop ([args args])
	   (if (and (pair? args)
		    (symbol? (car args)))
	       (loop (cdr args))
	       args))]
	[not-required
	 (let loop ([args not-required-with-defaults])
	   (if (pair? args)
	       (if (pair? (car args))
		   (let ([name (caar args)])
		     (if (symbol? name)
			 (cons name (loop (cdr args)))
			 (error "opt-lambda: all argument names must be symbols")))
		   (error "opt-lambda: all required args must come first"))
	       (if (or (null? args) (symbol? args))
		   args
		   (error "opt-lambda: all argument names must be symbols"))))]
	[defaults
	 (let loop ([args not-required-with-defaults])
	   (if (pair? args)
	       (let ([v (cdar args)])
		 (if (and (pair? v) (null? (cdr v)))
		     (cons (car v) (loop (cdr args)))
		     (error "opt-lambda: only one default value allowed per argument")))
	       ()))])
    `(letrec ([,f
	       (case-lambda
		,@(let loop ([required required]
			     [not-required not-required]
			     [defaults defaults])
		    (if (not (pair? not-required))
			(list `(,(append required not-required) ,@body)
;			      `(else (error 'opt-lambda 
;					    "too many arguments (~s allowed, ~s provided)"
;					    ,(length required) (length else)))
			      )
			(cons `(,required
				,(cons f (append required defaults)))
			      (loop (append required (list (car not-required)))
				    (cdr not-required)
				    (cdr defaults))))))])
       ,f)))
(constant opt-lambda)
