/*
 * File:        wx_keym.cc
 * Purpose:     wxKeymap implementation
 * Author:      Matthew Flatt
 * Created:     1995
 * Copyright:   (c) 1995, Matthew Flatt
 */

#include "wx_main.h"
#include "wx_keym.h"
#include "wx_utils.h"
#include <string.h>
#include <ctype.h>
#include "wx_ptreq.h"

#define DOUBLE_CLICK_THRESHOLD 500 /* Half-second */

class wxKeyFunc : public wxObject
{
 public:
  char *name;
  wxKeyFunction f;
  void *data;

  wxKeyFunc(char *name, wxKeyFunction, void *);
  ~wxKeyFunc();
  Bool Call(wxMediaBuffer *, wxKeyEvent &);
};

class wxMouseFunc : public wxObject
{
 public:
  char *name;
  wxMouseFunction f;
  void *data;

  wxMouseFunc(char *name, wxMouseFunction, void *);
  ~wxMouseFunc();
  Bool Call(wxMediaBuffer *, wxMouseEvent &);
};

class wxKeycode : public wxObject
{
 public:
  long code;
  Bool shift;
  Bool ctrl;
  Bool alt;
  Bool meta;
  char *fname;

  Bool isprefix;
  wxKeycode *seqprefix;

  wxKeycode *next;
};

/***************************************************************/

static int doubleClickThreshold = -1;

wxKeymap::wxKeymap()
{
  keyfunctions = new wxHashTable(wxKEY_STRING);
  mousefunctions = new wxHashTable(wxKEY_STRING);
  keys = new wxHashTable(wxKEY_INTEGER);

  numImpliedShifts = allocedImplies = 0;
  impliesShift = NULL;

  prefix = NULL;
  err = NULL;
  WXGC_IGNORE(errdata);

  active_mouse_function = NULL;

  chainTo = NULL;

  usage = 0;

  lastButton = 0;

  if (doubleClickThreshold < 0) {
#if USE_RESOURCES
    if (!wxGetResource(wxTheApp->wx_class, "doubleClickTime", 
		       &doubleClickThreshold))
#endif
      doubleClickThreshold = DOUBLE_CLICK_THRESHOLD;
  }

  doubleInterval = doubleClickThreshold;
}

wxKeymap::~wxKeymap()
{
  SetNextKeymap(NULL);

  keys->DeleteContents(TRUE);
  delete keys;

  keyfunctions->DeleteContents(TRUE);
  delete keyfunctions;

  mousefunctions->DeleteContents(TRUE);
  delete mousefunctions;

  if (impliesShift)
    delete[] impliesShift;
}

void wxKeymap::BreakSequence(void)
{
  prefix = NULL;
  if (chainTo)
    chainTo->BreakSequence();
}

wxKeycode *wxKeymap::FindKey(long code, Bool shift, Bool ctrl, 
			     Bool alt, Bool meta,
			     wxKeycode *prefix)
{
  wxKeycode *key;
  
  shift = !!shift;
  ctrl = !!ctrl;
  alt = !!alt;
  meta = !!meta;

  key = (wxKeycode *)keys->Get(code);
  while (key) {
    if (key->code == code
	&& key->shift == shift
	&& key->ctrl == ctrl
	&& key->alt == alt
	&& key->meta == meta
	&& key->seqprefix == prefix) {
      return key;
    }
    key = key->next;
  }

  return key;
}

wxKeycode *wxKeymap::MapFunction(long code, Bool shift, Bool ctrl, 
				 Bool alt, Bool meta,
				 char *fname, wxKeycode *prev, int type)
{
  wxKeycode *key, *newkey;

  if (key = FindKey(code, shift, ctrl, alt, meta, prev)) {
    if ((type == wxKEY_PREFIX) != key->isprefix) {
      if (err) {
	char buffer[256];
	
	if (isprint(code))
	  sprintf(buffer, "keymap: key %c ", (char)code);
	else
	  sprintf(buffer, "keymap: key code %ld ", code);

	if (shift)
	  strcat(buffer, "+ shift ");
	if (ctrl)
	  strcat(buffer, "+ control ");
	if (alt)
	  strcat(buffer, "+ alt ");
	if (meta)
	  strcat(buffer, "+ meta ");

	strcat(buffer, "is ");
	if (!key->isprefix)
	  strcat(buffer, "not ");
	strcat(buffer, "a prefix key");

	err(errdata, buffer);
      }

      return NULL;
    }  else {
      if (strcmp(key->fname, fname)) {
	delete[] key->fname;
	key->fname = copystring(fname);
      }
      return key;
    }
  }

  newkey = new wxKeycode;

  newkey->code = code;
  newkey->shift = shift;
  newkey->ctrl = ctrl;
  newkey->alt = alt;
  newkey->meta = meta;
  newkey->fname = copystring(fname);
  newkey->next = NULL;

  newkey->seqprefix = prev;

  newkey->isprefix = (type == wxKEY_PREFIX);

  key = (wxKeycode *)keys->Get(code);
  if (!key)
    keys->Put(code, newkey);
  else {
    while (key->next)
      key = key->next;
    key->next = newkey;
  }

  return newkey;
}

static struct {
  char *str;
  long code;
} keylist[] = { { "leftbutton" , WXK_MOUSE_LEFT },
		{ "rightbutton" , WXK_MOUSE_RIGHT },
		{ "middlebutton" , WXK_MOUSE_MIDDLE },
		{ "leftbuttondouble" , WXK_MOUSE_LEFT_DOUBLE },
		{ "rightbuttondouble" , WXK_MOUSE_RIGHT_DOUBLE },
		{ "middlebuttondouble" , WXK_MOUSE_MIDDLE_DOUBLE },
		{ "leftbuttontriple" , WXK_MOUSE_LEFT_TRIPLE },
		{ "rightbuttontriple" , WXK_MOUSE_RIGHT_TRIPLE },
		{ "middlebuttontriple" , WXK_MOUSE_MIDDLE_TRIPLE },
		{ "esc", WXK_ESCAPE }, 
		{ "delete", WXK_DELETE },
		{ "del", WXK_DELETE },
		{ "insert", WXK_INSERT },
		{ "ins", WXK_INSERT },
		{ "add", WXK_ADD },
		{ "subtract", WXK_SUBTRACT },
		{ "multiply", WXK_MULTIPLY },
		{ "divide", WXK_DIVIDE },
		{ "backspace", WXK_BACK },
		{ "back", WXK_BACK },
		{ "return", WXK_RETURN },
		{ "tab", WXK_TAB },
		{ "space", WXK_SPACE },
		{ "right", WXK_RIGHT },
		{ "left", WXK_RIGHT },
		{ "up", WXK_UP },
		{ "down", WXK_DOWN },
		{ "home", WXK_HOME },
		{ "end", WXK_END },
		{ "pageup", WXK_PRIOR },
		{ "pagedown", WXK_NEXT },
		{ "semicolon", ';' },
		{ "colon", ':' },
		{ "numpad1", WXK_NUMPAD1 },
		{ "numpad2", WXK_NUMPAD2 },
		{ "numpad3", WXK_NUMPAD3 },
		{ "numpad4", WXK_NUMPAD4 },
		{ "numpad5", WXK_NUMPAD5 },
		{ "numpad6", WXK_NUMPAD6 },
		{ "numpad7", WXK_NUMPAD7 },
		{ "numpad8", WXK_NUMPAD8 },
		{ "numpad9", WXK_NUMPAD9 },
		{ "f1", WXK_F1 },
		{ "f2", WXK_F2 },
		{ "f3", WXK_F3 },
		{ "f4", WXK_F4 },
		{ "f5", WXK_F5 },
		{ "f6", WXK_F6 },
		{ "f7", WXK_F7 },
		{ "f8", WXK_F8 },
		{ "f9", WXK_F9 },
		{ "f10", WXK_F10 },
		{ "f11", WXK_F11 },
		{ "f12", WXK_F12 },
		{ "f13", WXK_F13 },
		{ "f14", WXK_F14 },
		{ "f15", WXK_F15 },
		{ "f16", WXK_F16 },
		{ "f17", WXK_F17 },
		{ "f18", WXK_F18 },
		{ "f19", WXK_F19 },
		{ "f20", WXK_F20 },
		{ "f21", WXK_F21 },
		{ "f22", WXK_F22 },
		{ "f23", WXK_F23 },
		{ "f24", WXK_F24 },
		{ NULL, 0 }};

static long GetCode(char **keyseqp)
{
  char *keyseq = *keyseqp;
  long i, code;
#define MAX_BUF 256
  char buffer[MAX_BUF];

  buffer[0] = *(keyseq++);
  for (i = 1; *keyseq && (*keyseq != ';'); i++) {
    if (i >= MAX_BUF - 1)
      return 0;
    buffer[i] = tolower(*keyseq);
    keyseq++;
  }
  buffer[i] = 0;
  code = 0;
  if (buffer[1]) {
    buffer[0] = tolower(buffer[0]);
    for (i = 0; keylist[i].str; i++)
      if (!strcmp(buffer, keylist[i].str)) {
	code = keylist[i].code;
	break;
      }
  } else
    code = buffer[0];

  *keyseqp = keyseq;

  return code;
}

void wxKeymap::ImpliesShift(char *str)
{
  long code;
  char buffer[256];

  code = GetCode(&str);
  if (!code) {
    if (err) {
      sprintf(buffer, "implies-shift: bad key: \"%.100s\"", str);
      err(errdata, buffer);
    }
    return;
  }

  if (numImpliedShifts == allocedImplies) {
    long *old;

    old = impliesShift;
    allocedImplies = (!allocedImplies ? 40 : 2 * allocedImplies);
    impliesShift = new long[allocedImplies];
    if (old) {
      memcpy(impliesShift, old, allocedImplies * sizeof(long));
      delete[] old;
    }
  }

  impliesShift[numImpliedShifts++] = code;
}

void wxKeymap::MapFunction(char *keys, char *fname)
{
  char *keyseq = keys;
  wxKeycode *key = NULL;
  Bool shift, ctrl, alt, meta;
  int part = 1, i;
  long code;
  char *errstr;
  char buffer[256];

  while (*keyseq) {
    shift = ctrl = alt = meta = 0;
    code = 0;
    
    while (*keyseq && (*keyseq != ';')) {
      if (isspace(*keyseq)) {
      } else if (keyseq[1] == ':') {
	switch (tolower(*keyseq)) {
	case 's':
	  shift = TRUE;
	  break;
	case 'c':
	  ctrl = TRUE;
	  break;
	case 'm':
	  meta = TRUE;
	  break;
	case 'a':
	  alt = TRUE;
	  break;
	default:
	  errstr = "bad modifier";
	  goto key_error;
	}
	keyseq += 2;
      } else {
	code = GetCode(&keyseq);
	if (!code) {
	  errstr = "bad keyname";
	  goto key_error;
	}
      }
    }

    if (code) {
      if (shift && isalpha(code)) {
	code = toupper(code);
	shift = FALSE;
      } 

      if (!shift) {
	for (i = 0; i < numImpliedShifts; i++)
	  if (code == impliesShift[i]) {
	    shift = TRUE;
	    break;
	  }
      }

      key = MapFunction(code, shift, ctrl, alt, meta, fname, key, 
			*keyseq ? wxKEY_PREFIX : wxKEY_FINAL);
      part++;
      if (*keyseq)
	keyseq++;
    } else {
      errstr = "no non-modifier key";
      goto key_error;
    }

    if (!key)
      return;
  }

  return;

 key_error:
  if (err) {
    sprintf(buffer, "keymap: %s in keystring: \"%.100s\", part %d", 
	    errstr, keys, part);
    err(errdata, buffer);
  }
}

Bool wxKeymap::HandleEvent(long code, Bool shift, Bool ctrl, 
			   Bool alt, Bool meta,
			   char **fname)
{
  wxKeycode *key;

  key = FindKey(code, shift, ctrl, alt, meta, prefix);
  
  prefix = NULL;

  if (key) {
    if (key->isprefix) {
      prefix = key;
      *fname = NULL;
      return TRUE;
    }
    *fname = key->fname;
    return TRUE;
  }

  return FALSE;
}

Bool wxKeymap::HandleKeyEvent(wxMediaBuffer *media, wxKeyEvent &event)
{
  char *fname;

  lastTime = event.timeStamp;
  lastButton = 0;

  if (event.keyCode == WXK_SHIFT
      || event.keyCode == WXK_CONTROL
      || !event.keyCode)
    return FALSE;

  if (HandleEvent(event.keyCode,
		  event.shiftDown,
		  event.controlDown,
		  event.altDown,
		  event.metaDown,
		  &fname)) {
    if (fname) {
      if (chainTo)
	chainTo->BreakSequence();
      return CallFunction(fname, media, event);
    } else {
      if (chainTo)
	if (chainTo->HandleKeyEvent(media, event) > 0) {
	  BreakSequence();
	  return TRUE;
	}

      return -1;
    }
  } else if (chainTo)
    return chainTo->HandleKeyEvent(media, event);
  else
    return FALSE;
}

static inline long Abs(long x)
{
  if (x < 0)
    return -x;
  else
    return x;
}

Bool wxKeymap::HandleMouseEvent(wxMediaBuffer *media, wxMouseEvent &event)
{
  long code, origCode, lastCode;
  char *fname;

  if (event.timeStamp == lastTime)
    return FALSE; /* Same event */

  if (!event.ButtonDown()) {
    Bool v;

    if (!active_mouse_function)
      return FALSE;
    v = CallFunction(active_mouse_function, media, event);
    if (event.ButtonUp())
	active_mouse_function = NULL;
    return v;
  }

  if (event.rightDown)
    code = WXK_MOUSE_RIGHT;
  else if (event.leftDown)
    code = WXK_MOUSE_LEFT;
  else if (event.middleDown)
    code = WXK_MOUSE_MIDDLE;

  origCode = code;

  if (code == lastButton && event.x == lastX && event.y == lastY) {
    if (Abs(event.timeStamp - lastTime) < doubleInterval) {
      code += WXK_CLICK_ADDER * clickCount;
      clickCount++;
    } else
      clickCount = 1;
  } else {
    lastButton = code;
    clickCount = 1;
  }
  lastTime = event.timeStamp;
  lastX = event.x;
  lastY = event.y;

  do {
    if (HandleEvent(code,
		    event.shiftDown,
		    event.controlDown,
		    event.altDown,
		    event.metaDown,
		    &fname)) {
      if (fname) {
	if (chainTo)
	  chainTo->BreakSequence();
	return CallFunction(active_mouse_function = fname, media, event);
      } else {
	if (chainTo)
	  if (chainTo->HandleMouseEvent(media, event) > 0) {
	    BreakSequence();
	    return TRUE;
	  }

	return -1;
      }
    }
    lastCode = code;
    code = origCode;
  } while (lastCode != origCode);

  if (chainTo)
    return chainTo->HandleMouseEvent(media, event);
  else
    return FALSE;
}

void wxKeymap::AddKeyFunction(char *name, wxKeyFunction func, void *data)
{
  wxKeyFunc *f;

  f = new wxKeyFunc(name, func, data);
  if (keyfunctions->Get(f->name))
    keyfunctions->Delete(f->name);
  keyfunctions->Put(f->name, f);
}
  
void wxKeymap::AddMouseFunction(char *name, wxMouseFunction func, 
				void *data)
{
  wxMouseFunc *f;

  f = new wxMouseFunc(name, func, data);
  if (mousefunctions->Get(f->name))
    mousefunctions->Delete(f->name);
  mousefunctions->Put(f->name, f);
}
  
Bool wxKeymap::CallFunction(char *name, wxMediaBuffer *media, wxKeyEvent &event)
{
  wxKeyFunc *f;

  if (f = (wxKeyFunc *)keyfunctions->Get(name))
    return f->Call(media, event);
  else if (err) {
    char buffer[256];
    
    sprintf(buffer, "keymap: no key function \"%s\"", name);
    err(errdata, buffer);
  }

  return 0;
}

Bool wxKeymap::CallFunction(char *name, wxMediaBuffer *media,wxMouseEvent &event)
{
  wxMouseFunc *f;

  if (f = (wxMouseFunc *)mousefunctions->Get(name))
    return f->Call(media, event);
  else if (err) {
    char buffer[256];
    
    sprintf(buffer, "keymap: no mouse function \"%s\"", name);
    err(errdata, buffer);
  }

  return 0;
}

void wxKeymap::SetErrorCallback(wxKeyErrorFunction f, void *d)
{
  err = f;
  errdata = d;
}

void wxKeymap::AdjustUsage(Bool newUser)
{
  if (newUser)
    usage++;
  else
    --usage;
}

Bool wxKeymap::IsUsed(void)
{
  return (usage != 0);
}

long wxKeymap::GetDoubleClickInterval()
{
  return doubleInterval;
}

void wxKeymap::SetDoubleClickInterval(long d)
{
  doubleInterval = d;
}

void wxKeymap::SetNextKeymap(wxKeymap *km)
{
  wxKeymap *cycleCheck;

  cycleCheck = km;
  while (cycleCheck) {
    if (cycleCheck == this) {
      km = NULL;
      break;
    }
    cycleCheck = cycleCheck->chainTo;
  }

  if (PTREQ(km, chainTo))
    return;

  if (chainTo) {
    chainTo->AdjustUsage(FALSE);
#if !WXGARBAGE_COLLECTION_ON
    if (!chainTo->IsUsed())
      delete chainTo;
#endif
  }
  chainTo = km;
  if (chainTo)
    chainTo->AdjustUsage(TRUE);
}

/***************************************************************/

wxKeyFunc::wxKeyFunc(char *fname, wxKeyFunction func, void *d)
{
  name = copystring(fname);
  f = func;
  data = d;
  WXGC_IGNORE(data);
}

wxKeyFunc::~wxKeyFunc()
{
  delete[] name;
}

Bool wxKeyFunc::Call(wxMediaBuffer *media, wxKeyEvent &event)
{
  return f(media, event, data);
}

/***************************************************************/

wxMouseFunc::wxMouseFunc(char *fname, wxMouseFunction func, void *d)
{
  name = copystring(fname);
  f = func;
  data = d;
  WXGC_IGNORE(data);
}

wxMouseFunc::~wxMouseFunc()
{
  delete[] name;
}

Bool wxMouseFunc::Call(wxMediaBuffer *media, wxMouseEvent &event)
{
  return f(media, event, data);
}
