/*
 * File:        wx_mbuf.cc
 * Purpose:     wxMediaBuffer implementation
 * Author:      Matthew Flatt
 * Created:     1995
 * Copyright:   (c) 1995, Matthew Flatt
 */

#include "wx.h"
#include "wx_media.h"
#ifndef OLD_WXWINDOWS
#include "wx_clipb.h"
#else
#include "wx_gclip.h"
#endif
#include "wx_ptreq.h"

#define NUM_MAX_UNDOS 500

static void MediaStyleNotify(wxStyle *which, wxMediaBuffer *media);

class wxMediaClipboardClient : public wxClipboardClient
{
 public:
  wxMediaClipboardClient();
  char *GetData(char *format, long *size);
  void BeingReplaced(void);
};
static wxMediaClipboardClient TheMediaClipboardClient;

typedef wxChangeRecord *wxChangeRecordPtr;

/******************************************************************/

wxMediaBuffer::wxMediaBuffer()
{  
  map = new wxKeymap();
  map->AdjustUsage(TRUE);
  AddBufferFunctions(map);

  styleList = new wxStyleList;
  styleList->AdjustUsage(TRUE);
  styleList->NewNamedStyle(STD_STYLE, NULL);
  notifyId = styleList->NotifyOnChange((wxStyleNotifyFunc)MediaStyleNotify, 
				       this);

  filename = NULL;

  undomode = redomode = interceptmode = FALSE;
  maxUndos = NUM_MAX_UNDOS;
  changes = new wxChangeRecordPtr[maxUndos];
  changes_start = changes_end = 0;
  redochanges =  new wxChangeRecordPtr[maxUndos];
  redochanges_start = redochanges_end = 0;

  ownCaret = FALSE;
  caretSnip = NULL;  

  if (!wxmb_commonCopyBuffer) {
    wxmb_commonCopyBuffer = new wxList();
    wxmb_commonCopyBuffer2 = new wxList();
  }

  admin = NULL;
  WXGC_IGNORE(admin);
}

wxMediaBuffer::~wxMediaBuffer()
{
  if (map)
    SetKeymap(NULL);

  styleList->AdjustUsage(FALSE);
  styleList->ForgetNotification(notifyId);
#if !WXGARBAGE_COLLECTION_ON
  if (!styleList->IsUsed())
    delete styleList;
#endif

  if (filename)
    delete[] filename;

  ClearUndos();

  delete[] changes;
  delete[] redochanges;
}

/******************************************************************/

void wxMediaBuffer::OnLocalEvent(wxMouseEvent &event)
{
  if (map) {
    if (map->HandleMouseEvent(this, event))
      return;
    else
      map->BreakSequence();
  }

  OnDefaultEvent(event);
}

void wxMediaBuffer::OnLocalChar(wxKeyEvent &event)
{
  if (map) {
    if (map->HandleKeyEvent(this, event))
      return;
    else
      map->BreakSequence();
  }

  OnDefaultChar(event);
}

/******************************************************************/

void wxMediaBuffer::SetAdmin(wxMediaAdmin *administrator)
{
  SettingAdmin(administrator);

  admin = administrator;
  if (!admin)
    ownCaret = FALSE;
}

void wxMediaBuffer::SettingAdmin(wxMediaAdmin *)
{
}

wxMediaAdmin *wxMediaBuffer::GetAdmin(void)
{
  return admin;
}

/******************************************************************/

Bool wxMediaBuffer::DoOwnCaret(Bool ownit)
{
  Bool refresh;

  refresh = (!caretSnip && PTRNE(ownCaret, ownit));

  ownCaret = ownit;
  if (caretSnip)
    caretSnip->OwnCaret(ownit);
  if (map && !ownit && refresh)
    map->BreakSequence();
  return refresh;
}

wxDC *wxMediaBuffer::GetDC()
{
  /* This can be called by snips to get a DC appropriate for
     sizing text, etc., outside of draws. It isn't the destination 
     for draws, though. */
  if (admin)
    return admin->GetDC(NULL, NULL);
  else
    return NULL;
}

void wxMediaBuffer::GetViewSize(float *w, float *h)
{
  if (admin)
    admin->GetView(NULL, NULL, w, h);
  else {
    if (w)
      *w = 0;
    if (h)
      *h = 0;
  }
}

Bool wxMediaBuffer::DoSetCaretOwner(wxSnip *snip)
{
  Bool hadCaret;
  wxSnip *oldCaret;
  Bool refresh;

  if (PTREQ(snip, caretSnip))
    return FALSE;

  refresh = FALSE;

  if (!snip || !(snip->flags & wxSNIP_HANDLES_EVENTS)) {
    if (ownCaret && caretSnip) {
      caretSnip->OwnCaret(FALSE);
      refresh = TRUE;
    }
    caretSnip = NULL;
    return refresh;
  }

  if (!GetSnipLocation(snip, NULL, NULL))
    return refresh;

  if (!ownCaret) {
    hadCaret = FALSE;
  } else
    hadCaret = !caretSnip;

  oldCaret = caretSnip;
  caretSnip = snip;

  if (ownCaret) {
    BeginEditSequence();
    if (oldCaret)
      oldCaret->OwnCaret(FALSE);
    if (hadCaret)
      refresh = TRUE;
    snip->OwnCaret(TRUE);
    EndEditSequence();
  } else if (admin)
    admin->GrabCaret();

  return refresh;
}

void wxMediaBuffer::GlobalToLocal(float *x, float *y)
{
  float lx, ly;

  if (admin) {
    admin->GetDC(&lx, &ly);
    if (x)
      *x += lx;
    if (y)
      *y += ly;
  }
}

void wxMediaBuffer::LocalToGlobal(float *x, float *y)
{
  float lx, ly;

  if (admin) {
    admin->GetDC(&lx, &ly);
    if (x)
      *x -= lx;
    if (y)
      *y -= ly;
  }
}

/******************************************************************/

void wxMediaBuffer::SetKeymap(wxKeymap *keymap)
{
  if (map) {
    map->AdjustUsage(FALSE);
#if !WXGARBAGE_COLLECTION_ON
    if (!map->IsUsed())
      delete map;
#endif
  }
  map = keymap;
  if (keymap)
    keymap->AdjustUsage(TRUE);
}

wxKeymap *wxMediaBuffer::GetKeymap(void)
{
  return map;
}

wxStyleList *wxMediaBuffer::GetStyleList(void)
{
  return styleList;
}

void wxMediaBuffer::SetStyleList(wxStyleList *newList)
{
  styleList->AdjustUsage(FALSE);
  styleList->ForgetNotification(notifyId);
  if (!styleList->IsUsed())
    delete styleList;
  newList->AdjustUsage(TRUE);
  notifyId = newList->NotifyOnChange((wxStyleNotifyFunc)MediaStyleNotify, 
				     this);
  styleList = newList;

  if (!styleList->FindNamedStyle(STD_STYLE))
    styleList->NewNamedStyle(STD_STYLE, NULL);
}

static void MediaStyleNotify(wxStyle *which, wxMediaBuffer *media)
{
  media->StyleHasChanged(which);
}

/******************************************************************/

#define NUM_IMAGE_TYPES 4

static long image_type[NUM_IMAGE_TYPES]
  = {wxBITMAP_TYPE_BMP,
     wxBITMAP_TYPE_GIF,
     wxBITMAP_TYPE_XBM,
     wxBITMAP_TYPE_XPM};
static char *image_type_name[NUM_IMAGE_TYPES]
  = {"Windows Bitmap", "GIF", "X Bitmap", "XPM"};
static char *image_extension[NUM_IMAGE_TYPES]
  = {"bmp", "gif", "xbm", "xpm"};

static int image_hit_ok;

void wxmbSetCancel(wxObject& obj, wxEvent& evt)
{
  image_hit_ok = 0;
  ((wxButton *)&obj)->GetParent()->Show(FALSE);
}

void wxmbSetOK(wxObject& obj, wxEvent& evt)
{
  wxmbSetCancel(obj, evt);
  image_hit_ok = 1;
}

int wxMediaBuffer::AppendEditItems(wxMenu *edit, int idOffset)
{
  edit->Append(idOffset + wxEDIT_UNDO, "Undo");
  edit->Append(idOffset + wxEDIT_REDO, "Redo");
  edit->AppendSeparator();
  edit->Append(idOffset + wxEDIT_CLEAR, "Clear");
  edit->Append(idOffset + wxEDIT_COPY, "Copy");
  edit->Append(idOffset + wxEDIT_CUT, "Cut");
  edit->Append(idOffset + wxEDIT_PASTE, "Paste");
  edit->AppendSeparator();
  edit->Append(idOffset + wxEDIT_INSERT_TEXT_BOX, "Insert Text Box");
  edit->Append(idOffset + wxEDIT_INSERT_GRAPHIC_BOX, "Insert Graphic Box");
  edit->Append(idOffset + wxEDIT_INSERT_IMAGE, "Insert Image...");

  return idOffset + _wx_EDIT_counter + 1;
}

void wxMediaBuffer::DoEdit(int op, Bool recursive, long time)
{  
  if (recursive && caretSnip) {
    caretSnip->DoEdit(op, TRUE, time);
    return;
  }

  switch(op) {
  case wxEDIT_UNDO:
    Undo();
    break;
  case wxEDIT_REDO:
    Redo();
    break;
  case wxEDIT_CLEAR:
    Clear();
    break;
  case wxEDIT_CUT:
    Cut(FALSE, time);
    break;
  case wxEDIT_COPY:
    Copy(FALSE, time);
    break;
  case wxEDIT_PASTE:
    Paste(time);
    break;
  case wxEDIT_KILL:
    Kill(time);
    break;
  case wxEDIT_INSERT_TEXT_BOX:
    InsertBox(wxEDIT_BUFFER);
    break;
  case wxEDIT_INSERT_GRAPHIC_BOX:
    InsertBox(wxPASTEBOARD_BUFFER);
    break;
  case wxEDIT_INSERT_IMAGE:
    InsertImage();
    break;
  }
}

#define NUM_FAMILIES 7
#define NUM_DELTA_SIZES 10
#define NUM_SIZES 8
#define NUM_STYLES 3
#define NUM_WEIGHTS 3
#define NUM_UNDERLINES 3
#define NUM_ALIGNS 3
#define NUM_COLOURS 11

int m_families[NUM_FAMILIES] = {wxDEFAULT,wxDECORATIVE,wxROMAN,
			      wxSCRIPT,wxSWISS,wxMODERN, -1};
int m_pl_deltasizes[NUM_DELTA_SIZES] = {1,2,4,8,16,32,0,0,0,0};
int m_ti_deltasizes[NUM_DELTA_SIZES] = {1,1,1,1,01,01,2,3,4,5};
int m_sizes[NUM_SIZES] = {9, 10, 12, 14, 16, 24, 32, 48};
int m_styles[NUM_STYLES] = {wxNORMAL, wxSLANT, wxITALIC};
int m_weights[NUM_WEIGHTS] = {wxNORMAL, wxLIGHT, wxBOLD};
int m_alignments[NUM_ALIGNS] = {wxALIGN_TOP, wxALIGN_CENTER, wxALIGN_BOTTOM};
const char *m_colour_names[NUM_COLOURS] = {"Black", "White",
				      "Red", "Orange", "Yellow",
				      "Green", "Blue", "Purple",
				      "Cyan", "Magenta", "Grey"};
const char *m_colours[NUM_COLOURS] = {"BLACK", "WHITE",
				      "RED", "ORANGE", "YELLOW",
				      "GREEN", "BLUE", "PURPLE",
				      "CYAN", "MAGENTA", "GREY"};
				      

int wxMediaBuffer::AppendFontItems(wxMenu *font, int idOffset)
{
  wxMenu *family, *size, *style, *weight, *underline;
  wxMenu *colour, *background, *alignment;
  wxMenu *bigger, *smaller;
  int i;
  
  family = new wxMenu(NULL, (wxFunction)NULL);
  bigger = new wxMenu(NULL, (wxFunction)NULL);
  smaller = new wxMenu(NULL, (wxFunction)NULL);
  size = new wxMenu(NULL, (wxFunction)NULL);
  style = new wxMenu(NULL, (wxFunction)NULL);
  weight = new wxMenu(NULL, (wxFunction)NULL);
  underline = new wxMenu(NULL, (wxFunction)NULL);
  colour = new wxMenu(NULL, (wxFunction)NULL);
  background = new wxMenu(NULL, (wxFunction)NULL);
  alignment = new wxMenu(NULL, (wxFunction)NULL);
  
  font->Append(0, "Face", family);
  font->Append(1, "Size", size);
  font->Append(2, "Style", style);
  font->Append(3, "Weight", weight);
  font->Append(4, "Underline", underline);
  font->Append(5, "Alignment", alignment);
  font->Append(6, "Color", colour);
  font->Append(7, "Background", background);

  family->Append(idOffset++, "Standard");
  family->Append(idOffset++, "Decorative");
  family->Append(idOffset++, "Roman");
  family->Append(idOffset++, "Script");
  family->Append(idOffset++, "Swiss");
  family->Append(idOffset++, "Fixed");
  family->AppendSeparator();
  family->Append(idOffset++, "Other...");

  bigger->Append(idOffset++, "+1");
  bigger->Append(idOffset++, "+2");
  bigger->Append(idOffset++, "+4");
  bigger->Append(idOffset++, "+8");
  bigger->Append(idOffset++, "+16");
  bigger->Append(idOffset++, "+32");
  bigger->AppendSeparator();
  bigger->Append(idOffset++, "x2");
  bigger->Append(idOffset++, "x3");
  bigger->Append(idOffset++, "x4");
  bigger->Append(idOffset++, "x5");
  
  smaller->Append(idOffset++, "-1");
  smaller->Append(idOffset++, "-2");
  smaller->Append(idOffset++, "-4");
  smaller->Append(idOffset++, "-8");
  smaller->Append(idOffset++, "-16");
  smaller->Append(idOffset++, "-32");
  smaller->AppendSeparator();
  smaller->Append(idOffset++, "1/2");
  smaller->Append(idOffset++, "1/3");
  smaller->Append(idOffset++, "1/4");
  smaller->Append(idOffset++, "1/5");

  size->Append(0, "Bigger", bigger);
  size->Append(1, "Smaller", smaller);
  size->Append(idOffset++, "9");
  size->Append(idOffset++, "10");
  size->Append(idOffset++, "12");
  size->Append(idOffset++, "14");
  size->Append(idOffset++, "16");
  size->Append(idOffset++, "24");
  size->Append(idOffset++, "32");
  size->Append(idOffset++, "48");

  style->Append(idOffset++, "Normal");
  style->Append(idOffset++, "Slant");
  style->Append(idOffset++, "Italic");

  weight->Append(idOffset++, "Normal");
  weight->Append(idOffset++, "Light");
  weight->Append(idOffset++, "Bold");

  underline->Append(idOffset++, "No Underline");
  underline->Append(idOffset++, "Underlined");
  underline->Append(idOffset++, "Toggle");

  alignment->Append(idOffset++, "Top");
  alignment->Append(idOffset++, "Center");
  alignment->Append(idOffset++, "Bottom");

  for (i = 0; i < NUM_COLOURS; i++)
    colour->Append(idOffset++, (char *)m_colour_names[i]);

  for (i = 0; i < NUM_COLOURS; i++)
    background->Append(idOffset++, (char *)m_colour_names[i]);

  return idOffset;
}

void wxMediaBuffer::DoFont(int op, Bool recursive)
{
  if (recursive && caretSnip) {
    caretSnip->DoFont(op, TRUE);
    return;
  }

  if (op < NUM_FAMILIES) {
    int family = m_families[op];
    if (family < 0) {
      char *s = wxGetTextFromUser("Face name:", "Select Face");
      if (!s)
	return;
      wxStyleDelta delta;

      delta.SetDeltaFace(s);

      ChangeStyle(&delta);
    } else {
      wxStyleDelta delta(wxCHANGE_FAMILY, family);
      ChangeStyle(&delta);
    }
    return;
  }
  op -= NUM_FAMILIES;
  if (op < NUM_DELTA_SIZES) {
    wxStyleDelta delta(wxCHANGE_BIGGER, m_pl_deltasizes[op]);
    delta.sizeMult = m_ti_deltasizes[op];
    ChangeStyle(&delta);
    return;
  }
  op -= NUM_DELTA_SIZES;
  if (op < NUM_DELTA_SIZES) {
    wxStyleDelta delta(wxCHANGE_SMALLER, m_pl_deltasizes[op]);
    delta.sizeMult = 1.0/m_ti_deltasizes[op];
    ChangeStyle(&delta);
    return;
  }
  op -= NUM_DELTA_SIZES;
  if (op < NUM_SIZES) {
    wxStyleDelta delta(wxCHANGE_SIZE, m_sizes[op]);
    ChangeStyle(&delta);
    return;
  }
  op -= NUM_SIZES;
  if (op < NUM_STYLES) {
    wxStyleDelta delta(wxCHANGE_STYLE, m_styles[op]);
    ChangeStyle(&delta);
    return;
  }
  op -= NUM_STYLES;
  if (op < NUM_WEIGHTS) {
    wxStyleDelta delta(wxCHANGE_WEIGHT, m_weights[op]);
    ChangeStyle(&delta);
    return;
  }
  op -= NUM_WEIGHTS;
  if (op < NUM_UNDERLINES) {
    wxStyleDelta delta;

    if (op == 2)
      delta.SetDelta(wxCHANGE_TOGGLE_UNDERLINE);
    else
      delta.SetDelta(wxCHANGE_UNDERLINE, op);

    ChangeStyle(&delta);
    return;
  }
  op -= NUM_UNDERLINES;
  if (op < NUM_ALIGNS) {
    wxStyleDelta delta(wxCHANGE_ALIGNMENT, m_alignments[op]);
    ChangeStyle(&delta);
    return;
  }
  op -= NUM_ALIGNS;
  if (op < NUM_COLOURS) {
    wxStyleDelta delta;
    delta.SetDeltaForeground((char *)m_colours[op]);
    ChangeStyle(&delta);
    return;
  }
  op -= NUM_COLOURS;
  if (op < NUM_COLOURS) {
    wxStyleDelta delta;
    delta.SetDeltaBackground((char *)m_colours[op]);
    ChangeStyle(&delta);
    return;
  }
  op -= NUM_COLOURS;
}

void wxMediaBuffer::InsertBox(int type)
{
  wxMediaSnip *snip;
  wxMediaBuffer *media;

  BeginEditSequence();
  if (type == wxEDIT_BUFFER)
    media = new wxMediaEdit();
  else
    media = new wxMediaPasteboard();
  snip = new wxMediaSnip(media);
  snip->style = styleList->FindNamedStyle(STD_STYLE);
  Insert(snip);
  media->SetKeymap(map);
  media->SetStyleList(styleList);
  SetCaretOwner(snip);
  EndEditSequence();
}

void wxMediaBuffer::InsertImage(char *filename, long type, Bool relative)
{
  wxDialogBox *dial;
  wxRadioBox *boxes;
  int i, len;
  char extension[10];
    
  if (!filename)
    filename = wxFileSelector("Image", NULL, NULL, NULL, NULL, wxOPEN);

  if (!filename)
    return;

  for (i = 0; i < NUM_IMAGE_TYPES; i++)
    if (type == image_type[i])
      break;

  if (i >=  NUM_IMAGE_TYPES) {
    dial = new wxDialogBox(NULL, "Image Type", TRUE);
    
    len = strlen(filename);
    for (i = len; --i; )
      if (filename[i] == '.')
	break;
    
    if (i + 10 > len) {
      strcpy(extension, filename + i + 1);
      for (i = 0; extension[i]; i++)
	extension[i] = tolower(extension[i]);
    } else
      extension[0] = 0;
    
    dial->SetLabelPosition(wxVERTICAL);
    boxes = new wxRadioBox(dial, NULL,
			   "Image format:", -1, -1, -1, -1,
			   NUM_IMAGE_TYPES, image_type_name);
    
    for (i = 0; i < NUM_IMAGE_TYPES; i++)
      if (!strcmp(image_extension[i], extension))
	boxes->SetSelection(i);
    
    dial->NewLine();
    
    (void)(new wxButton(dial, wxmbSetCancel, "Cancel"));
    (void)(new wxButton(dial, wxmbSetOK, "OK"));
    
    dial->NewLine();
    
    dial->Fit();
    dial->Show(TRUE);
    
    if (!image_hit_ok)
      return;

    type = image_type[boxes->GetSelection()];
  }

  wxImageSnip *snip;
      
  snip = new wxImageSnip(filename, type, relative);
  Insert(snip);
}

/**********************************************************************/

wxBufferData *wxMediaBuffer::GetSnipData(wxSnip *)
{
  return NULL;
}

void wxMediaBuffer::SetSnipData(wxSnip *, wxBufferData *)
{
}

/**********************************************************************/

Bool wxmbReadGlobalHeader(wxStyleList *styleList, wxMediaStreamIn &f)
{
  wxTheSnipClassList.ResetHeaderFlags();
  if (!wxTheSnipClassList.Read(f))
    return FALSE;

  wxmbSetupStyleReadsWrites();

  return wxTheBufferDataClassList.Read(f);
}

Bool wxmbReadGlobalFooter(wxMediaStreamIn &f)
{
  wxmbDoneStyleReadsWrites();
  wxTheSnipClassList.ResetHeaderFlags(wxRESET_DONE_READ);
  return TRUE;
}

Bool wxmbWriteGlobalHeader(wxStyleList *styleList, wxMediaStreamOut &f)
{
  wxTheSnipClassList.ResetHeaderFlags();
  if (!wxTheSnipClassList.Write(f))
    return FALSE;

  wxmbSetupStyleReadsWrites();

  return wxTheBufferDataClassList.Write(f);
}

Bool wxmbWriteGlobalFooter(wxMediaStreamOut &f)
{
  wxmbDoneStyleReadsWrites();
  wxTheSnipClassList.ResetHeaderFlags(wxRESET_DONE_WRITE);
  return TRUE;
}

/**********************************************************************/

char wxme_current_read_format[MRED_FORMAT_STR_LEN + 1];
char wxme_current_read_version[MRED_VERSION_STR_LEN + 1];

int wxmeCheckFormatAndVersion(void)
{
  if (strcmp(wxme_current_read_format, MRED_FORMAT_STR)) {
    wxMessageBox("Unknown format number.", "Error");
    return 0;
  }
  if (strcmp(wxme_current_read_version, MRED_VERSION_STR)
      && strcmp(wxme_current_read_version, "01")) {
    wxMessageBox("Unknown version number.", "Error");
    return 0;
  }

  return 1;
}

Bool wxMediaBuffer::ReadHeaderFromFile(wxMediaStreamIn &, char *headerName)
{
  char buffer[256];

  sprintf(buffer, "Unknown header data: \"%.100s\"."
	  " The file will be loaded anyway.", headerName);

  wxMessageBox(buffer, "Warning");

  return TRUE;
}

Bool wxMediaBuffer::ReadFooterFromFile(wxMediaStreamIn &, char *headerName)
{
  char buffer[256];

  sprintf(buffer, "Unknown header data: \"%.100s\"."
	  " The file will be loaded anyway.", headerName);

  wxMessageBox(buffer, "Warning");

  return TRUE;
}

Bool wxMediaBuffer::WriteHeadersToFile(wxMediaStreamOut &f)
{
  return TRUE;
}

Bool wxMediaBuffer::WriteFootersToFile(wxMediaStreamOut &f)
{
  return TRUE;
}

Bool wxMediaBuffer::BeginWriteHeaderFooterToFile(wxMediaStreamOut &f, 
						 char *headerName,
						 long *dataBuffer)
{
  *dataBuffer = f.Tell();
  f.PutFixed(0);
  f << headerName;

  return TRUE;
}

Bool wxMediaBuffer::EndWriteHeaderFooterToFile(wxMediaStreamOut &f, 
					       long dataBuffer)
{
  long end, pos;

  end = f.Tell();

  f.JumpTo(dataBuffer);
  f.PutFixed(0);
  pos = f.Tell();
  
  f.JumpTo(dataBuffer);
  f.PutFixed(end - pos);

  f.JumpTo(end);

  numExtraHeaders++;
  
  return TRUE;
}

Bool wxMediaBuffer::ReadHeadersFooters(wxMediaStreamIn &f, Bool headers)
{
  char headerName[256];
  long len, hlen, i, pos, numHeaders;

  f.GetFixed(numHeaders);
  if (numHeaders > 1000) {
    if (wxMessageBox("File contains suspiciously large value for special"
		     " headers/footers. Should I give up?", "Warning",
		     wxYES_NO | wxCENTRE) == wxYES)
      return FALSE;
  }

  for (i = 0; i < numHeaders; i++) {
    f.GetFixed(len);
    if (!f.Ok())
      return FALSE;
    if (len) {
      pos = f.Tell();
      f.SetBoundary(len);
      hlen = 256;
      f.Get((long *)&hlen, (char *)headerName);
      if (headers) {
	if (!ReadHeaderFromFile(f, headerName))
	  return FALSE;
      } else {
	if (!ReadFooterFromFile(f, headerName))
	  return FALSE;
      }
      if (!f.Ok())
	return FALSE;
      f.RemoveBoundary();
      if (len -= (f.Tell() - pos))
	f.Skip(len);
      if (!f.Ok())
	return FALSE;
    }
  }

  return TRUE;
}

Bool wxMediaBuffer::DoWriteHeadersFooters(wxMediaStreamOut &f, Bool headers)
{
  long allStart, allEnd;

  allStart = f.Tell();
  f.PutFixed(0);
  numExtraHeaders = 0;

  if (headers) {
    if (!WriteHeadersToFile(f))
      return FALSE;
  } else {
    if (!WriteFootersToFile(f))
      return FALSE;
  }

  if (numExtraHeaders) {
    allEnd = f.Tell();
    f.JumpTo(allStart);
    f.PutFixed(numExtraHeaders);
    f.JumpTo(allEnd);
  }

  return TRUE;
}

/**********************************************************************/

static wxBufferData *ReadBufferData(wxMediaStreamIn &f)
{
  wxBufferData *data, *newdata;
  wxBufferDataClass *dclass;
  int extraDataIndex;
  long datalen;

  data = NULL;
  do {
    f >> extraDataIndex;
    if (extraDataIndex) {
      dclass = wxTheBufferDataClassList.FindByMapPosition(extraDataIndex);
      
      if (!dclass || !dclass->required)
	f >> datalen;
      else
	datalen = -1;
      
      if (dclass) {
	if (datalen >= 0)
	  f.SetBoundary(datalen);
	if (!(newdata = dclass->Read(f)))
	  return FALSE;
	newdata->next = data;
	data = newdata;
	if (datalen >= 0)
	  f.RemoveBoundary();
      } else {
	/* Unknown extra data */
	f.Skip(datalen);
      }
      if (!f.Ok())
	return FALSE;
    }
  } while (extraDataIndex);

  return data;
}

Bool wxMediaBuffer::ReadSnipsFromFile(wxMediaStreamIn &f)
{
  long len, numHeaders, numSnips, i;
  int styleIndex;
  short n;
  wxStyleList *newList;
  wxSnipClass *sclass;
  wxBufferData *data;
  wxSnip *snip;

  if (!ReadHeadersFooters(f, TRUE))
    return FALSE;

  if (!(newList = wxmbReadStylesFromFile(styleList, f)))
    return FALSE;

  if (PTRNE(newList, styleList))
    SetStyleList(newList);
  
  f.GetFixed(numHeaders);
  
  if (numHeaders > 100) {
    if (wxMessageBox("File contains suspiciously large value for class"
		     " headers. Should I give up?", "Warning",
		     wxYES_NO | wxCENTRE) == wxYES)
      return FALSE;
  }

  for (i = 0; i < numHeaders; i++) {
    f >> n;
    f.GetFixed(len);
    if (!f.Ok())
      return FALSE;
    if (len) {
      sclass = wxTheSnipClassList.FindByMapPosition(n);
      if (sclass) {
	f.SetBoundary(len);
	if (!sclass->ReadHeader(f))
	  return FALSE;
	if (!f.Ok())
	  return FALSE;
	sclass->headerFlag = 1;
	f.RemoveBoundary();
      } else {
	f.Skip(len);
      }
      if (!f.Ok())
	return FALSE;
    }
  }

  f >> numSnips;

  for (i = 0; i < numSnips; i++) {
    f >> n;
    if (n >= 0)
      sclass = wxTheSnipClassList.FindByMapPosition(n);
    else
      sclass = NULL; /* -1 => unknown */
    if (!sclass || !sclass->required)
      f.GetFixed(len);
    else
      len = -1;
    if (!f.Ok())
      return FALSE;
    if (len) {
      if (sclass) {
	if (len >= 0)
	  f.SetBoundary(len);
	f >> styleIndex;
	if (styleIndex < 0 || styleIndex >= styleList->numMappedStyles) {
	  wxMessageBox("Bad style index for snip.", "Error");
	  return FALSE;
	}
	if (snip = sclass->Read(f)) {
	  if (snip->flags & wxSNIP_OWNED)
	    snip->flags -= wxSNIP_OWNED;
	  snip->style = styleList->MapIndexToStyle(styleIndex);
	  if (!ReadInsert(snip))
	    return FALSE;
	} else
	  return FALSE;

	data = ReadBufferData(f);
	if (!f.Ok())
	  return FALSE;

	if (data)
	  SetSnipData(snip, data);

	if (len >= 0)
	  f.RemoveBoundary();
      } else {
	f.Skip(len);
      }
      if (!f.Ok())
	return FALSE;
    }
  }

  if (!ReadHeadersFooters(f, FALSE))
    return FALSE;

  return TRUE;
}

Bool wxmbWriteBufferData(wxMediaStreamOut &f, wxBufferData *data)
{
  long dataPos, dataStart, dataEnd;
  
  while (data) {
    f << data->dataclass->mapPosition;
    
    if (!data->dataclass->required) {
      dataStart = f.Tell();
      f.PutFixed(0);
      dataPos = f.Tell();
    }

    if (!data->Write(f))
      return FALSE;
    
    if (!data->dataclass->required) {
      dataEnd = f.Tell();
      f.JumpTo(dataStart);
      f.PutFixed(dataEnd - dataPos);
      f.JumpTo(dataEnd);
    }
    data = data->next;
  }

  f << 0;

  return TRUE;
}

Bool wxmbWriteSnipsToFile(wxMediaStreamOut &f, 
			  wxStyleList *styleList,
			  wxList *snipList, 
			  wxSnip *startSnip, wxSnip *endSnip,
			  wxList *extraData,
			  wxMediaBuffer *buffer)
{
  long allStart, allEnd, headerPos, headerStart, headerEnd;
  long snipCount, snipPos, snipStart, snipEnd;
  long numHeaders;
  int styleIndex;
  wxNode *node, *node2;
  wxSnip *snip;
  wxSnipClass *sclass;
  wxBufferData *data;

  if (!wxmbWriteStylesToFile(styleList, f))
    return FALSE;
  
  allStart = f.Tell();
  f.PutFixed(0);

  if (snipList) {
    node = snipList->First();
    if (!node)
      return FALSE;
    startSnip = (wxSnip *)node->Data();
    endSnip = NULL;
  }

  numHeaders = 0;
  snipCount = 0;
  for (snip = startSnip; PTRNE(snip, endSnip); snipCount++) {
    sclass = snip->snipclass;
    if (!sclass) {
      wxMessageBox("There's a snip without a class."
		   " Data will be lost.", "Warning");
    } else if (!sclass->headerFlag) {
      f << (short)sclass->mapPosition;
      headerStart = f.Tell();
      f.PutFixed(0);
      headerPos = f.Tell();
      if (!sclass->WriteHeader(f))
	return FALSE;
      sclass->headerFlag = 1;
      headerEnd = f.Tell();
      f.JumpTo(headerStart);
      f.PutFixed(headerEnd - headerPos);
      f.JumpTo(headerEnd);
      numHeaders++;
      if (!f.Ok())
	return FALSE;
    }

    if (snipList) {
      node = node->Next();
      if (node)
	snip = (wxSnip *)node->Data();
      else
	snip = NULL;
    } else
      snip = snip->next;
  }

  allEnd = f.Tell();
  f.JumpTo(allStart);
  f.PutFixed(numHeaders);
  f.JumpTo(allEnd);

  f << snipCount;

  if (snipList)
    node = snipList->First();
  if (extraData)
    node2 = extraData->First();
  else
    node2 = NULL;

  for (snip = startSnip; PTRNE(snip, endSnip); ) {
    sclass = snip->snipclass;

    if (sclass)
      f << (short)sclass->mapPosition;
    else
      f << (short)(-1);

    if (!snip->snipclass || !snip->snipclass->required) {
      snipStart = f.Tell();
      f.PutFixed(0);
      snipPos = f.Tell();
    }

    styleIndex = styleList->StyleToIndex(snip->style);
    if (styleIndex < 0) {
      wxMessageBox("Bad style discovered.", "Warning");
      styleIndex = 0;
    }
    f << styleIndex;

    snip->Write(f);
    if (node2)
      data = (wxBufferData *)node2->Data();
    else
      data = buffer->GetSnipData(snip);

    if (!wxmbWriteBufferData(f, data))
      return FALSE;

    if (!snip->snipclass || !snip->snipclass->required) {
      snipEnd = f.Tell();
      f.JumpTo(snipStart);
      f.PutFixed(snipEnd - snipPos);
      f.JumpTo(snipEnd);
    }

    if (!f.Ok())
      return FALSE;

    if (snipList) {
      node = node->Next();
      if (node)
	snip = (wxSnip *)node->Data();
      else
	snip = NULL;
    } else
      snip = snip->next;
    if (extraData)
      node2 = node2->Next();
  }

  return TRUE;
}

char *wxMediaBuffer::GetFilename(Bool *temp)
{
  if (temp)
    *temp = tempFilename;
  return filename;
}

void wxMediaBuffer::Print(char *filename, Bool interactive)
{
  wxDC *dc;
  Bool removeit, delname;

  dc = new wxPostScriptDC(filename, interactive);

#ifdef wx_x
#define FILTER "*"
#else
#define FILTER "*.*"
#endif

  if (!filename) {
    if (wxGetPrinterMode() == PS_FILE) {
      filename = wxFileSelector("PostScript file name", NULL, NULL, "ps",
				FILTER, wxSAVE);
      if (!filename) {
	delete dc;
	return;
      }

      filename = copystring(filename);

      delname = TRUE;
      removeit = FALSE;
    } else {
      filename = wxGetTempFileName("wxmeps");
      removeit = delname = TRUE;
    }
  } else
    removeit = delname = FALSE;

  if (dc->Ok()) { 
    dc->StartDoc("Printing buffer");
    PrintToDC(dc);
    dc->EndDoc();
  }

  delete dc;

  if (removeit)
    wxRemoveFile(filename);

  if (delname)
    delete filename;
}

/****************************************************************/

void wxMediaBuffer::Undo(void)
{
  undomode = TRUE;

  PerformUndos(changes, changes_start, changes_end);

  undomode = FALSE;
}

void wxMediaBuffer::Redo(void)
{ 
  redomode = TRUE;

  PerformUndos(redochanges, redochanges_start, redochanges_end);

  redomode = FALSE;
}

static void wxmeClearUndos(wxChangeRecord **changes, int& start, int& end,
			   int maxUndos)
{
  int i;

  for (i = start; i != end; i = (i + 1) % maxUndos)
    delete changes[i];

  start = end = 0;  
}

void wxMediaBuffer::AddUndo(wxChangeRecord *rec)
{
  if (interceptmode)
    intercepted->Append((wxObject *)rec);
  else if (undomode)
    AppendUndo(rec, redochanges, redochanges_start, redochanges_end);
  else if (!noundomode) {
    if (!redomode)
      wxmeClearUndos(redochanges, redochanges_start,
		     redochanges_end, maxUndos);
    AppendUndo(rec, changes, changes_start, changes_end);    
  } else
    delete rec;
}

void wxMediaBuffer::AppendUndo(wxChangeRecord *rec, wxChangeRecord **changes, 
			       int& start, int& end)
{
  if (maxUndos) {
    changes[end] = rec;
    end = (end + 1) % maxUndos;
    if (end == start) {
      delete changes[start];
      start = (start + 1) % maxUndos;
    }
  } else
    delete rec;
}

void wxMediaBuffer::PerformUndos(wxChangeRecord **changes, 
				 int& start, int& end)
{
  wxChangeRecord *rec;
  Bool cont;
  
  BeginEditSequence();
  while (start != end) {
    end = (end - 1 + maxUndos) % maxUndos;
    rec = changes[end];
    cont = rec->Undo(this);
    delete rec;
    if (!cont)
      break;
  }
  EndEditSequence();
}

void wxMediaBuffer::PerformUndoList(wxList *changes)
{
  wxNode *node;
  wxChangeRecord *rec;
  Bool cont;
  
  BeginEditSequence();
  do {
    node = changes->Last();
    if (node) {
      rec = (wxChangeRecord *)node->Data();
      cont = rec->Undo(this);
      delete rec;
      changes->DeleteNode(node);
    }
  } while (node && cont);
  EndEditSequence();
}

void wxMediaBuffer::ClearUndos()
{
  wxmeClearUndos(changes, changes_start, changes_end, maxUndos);
  wxmeClearUndos(redochanges, redochanges_start, redochanges_end, maxUndos);
}

void wxMediaBuffer::SetMaxUndoHistory(int v)
{
  if (undomode || redomode || (v == maxUndos))
    return;

  wxChangeRecord **naya;
  int i, j;

  naya = new wxChangeRecordPtr[v];
  for (j = 0, i = changes_start; 
       (i != changes_end) && (j < v); 
       j++, i = (i + 1) % maxUndos)
    naya[j] = changes[i];
  changes_start = 0;
  changes_end = j % v;

  naya = new wxChangeRecordPtr[v];
  for (j = 0, i = redochanges_start; 
       (i != redochanges_end) && (j < v); 
       j++, i = (i + 1) % maxUndos)
    naya[j] = redochanges[i];
  redochanges_start = 0;
  redochanges_end = j % v;

  maxUndos = v;
}

int wxMediaBuffer::GetMaxUndoHistory()
{
  return maxUndos;
}

/****************************************************************/

wxList *wxmb_commonCopyBuffer = NULL;
wxList *wxmb_commonCopyBuffer2 = NULL;
wxStyleList *wxmb_copyStyleList = NULL;
wxBufferData *wxmb_commonCopyRegionData = NULL;

void wxMediaBuffer::FreeOldCopies(void)
{
  wxNode *node, *next;

  for (node = wxmb_commonCopyBuffer->First(); node; node = next) {
    next = node->Next();
    delete (wxSnip *)node->Data();
    wxmb_commonCopyBuffer->DeleteNode(node);
  }  
  for (node = wxmb_commonCopyBuffer2->First(); node; node = next) {
    next = node->Next();
    if (node->Data())
      delete node->Data();
    wxmb_commonCopyBuffer2->DeleteNode(node);
  }
  if (wxmb_commonCopyRegionData)
    delete wxmb_commonCopyRegionData;
}

void wxMediaBuffer::InstallCopyBuffer(long time)
{
  if (wxmb_copyStyleList) {
    wxmb_copyStyleList->AdjustUsage(FALSE);
    if (!wxmb_copyStyleList->IsUsed())
      delete wxmb_copyStyleList;
  }
  wxmb_copyStyleList = styleList;
  wxmb_copyStyleList->AdjustUsage(TRUE);
  wxTheClipboard->SetClipboardClient(&TheMediaClipboardClient, time);
}

void wxMediaBuffer::DoPaste(long time)
{
  wxClipboardClient *owner;
  wxNode *node, *node2;
  wxSnip *snip;
  
  owner = wxTheClipboard->GetClipboardClient();
  if (PTREQ(owner, &TheMediaClipboardClient)) {
    for (node = wxmb_commonCopyBuffer->First(),
	 node2 = wxmb_commonCopyBuffer2->First(); 
	 node; 
	 node = node->Next(),
	 node2 = node2->Next())
      InsertPasteSnip(snip = ((wxSnip *)node->Data())->Copy(),
		      (wxBufferData *)node2->Data());
    if (wxmb_commonCopyRegionData && bufferType == wxEDIT_BUFFER)
      ((wxMediaEdit *)this)->PasteRegionData(wxmb_commonCopyRegionData);
  } else {
    char *str;
    long len;

    if (str = wxTheClipboard->GetClipboardData("WXME", &len, time)) {

      strcpy(wxme_current_read_format, MRED_FORMAT_STR);
      strcpy(wxme_current_read_version, MRED_VERSION_STR);

      wxMediaStreamInStringBase b(str, len);
      wxMediaStreamIn mf(b);

      if (wxmbReadGlobalHeader(styleList, mf))
	if (mf.Ok())
	  if (ReadFromFile(mf)) {
	    wxBufferData *data;
	    data = ReadBufferData(mf);
	    if (data && bufferType == wxEDIT_BUFFER)
	      ((wxMediaEdit *)this)->PasteRegionData(data);
	  }
      wxmbReadGlobalFooter(mf);
    } else {
      str = wxTheClipboard->GetClipboardString(time);
      InsertPasteString(str);
      delete[] str;
    }
  }
}

wxMediaClipboardClient::wxMediaClipboardClient()
{
  formats.Add("TEXT");
  formats.Add("WXME");
}

char *wxMediaClipboardClient::GetData(char *format, long *size)
{
  wxNode *node;
  wxSnip *snip;
  long l, length = 0;
  char *str, *total = NULL, *old;

  if (!strcmp(format, "TEXT")) {
    for (node = wxmb_commonCopyBuffer->First(); node; node = node->Next()) {
      snip = (wxSnip *)node->Data();
      str = snip->GetText(0, snip->count, TRUE);
      l = strlen(str);
      if (total) {
	old = total;
	total = new char[length + l];
	memcpy(total, old, length);
	memcpy(total + length, str, l);
	delete[] old;
      } else
	total = str;
      length += l;
    }
    
    if (!total) {
      total = new char[1];
    }
    
    *size = length;

    return total;
  } else if (!strcmp(format, "WXME")) {
    wxMediaStreamOutStringBase b;
    wxMediaStreamOut mf(b);

    wxmbWriteGlobalHeader(wxmb_copyStyleList, mf);
    if (mf.Ok()) {
      mf.PutFixed(0);
      if (!wxmbWriteSnipsToFile(mf, wxmb_copyStyleList, wxmb_commonCopyBuffer, 
				NULL, NULL, wxmb_commonCopyBuffer2, NULL))
	return FALSE;
      mf.PutFixed(0);
      wxmbWriteBufferData(mf, wxmb_commonCopyRegionData);
    }
    wxmbWriteGlobalFooter(mf);

    return b.GetString(size);
  } else {
    *size = 0;
    return "";
  }
}  

void wxMediaClipboardClient::BeingReplaced(void)
{
}

/****************************************************************/

void wxMediaBuffer::Lock(Bool lock)
{
  userLocked = lock;
}

Bool wxMediaBuffer::Modified(void)
{
  return modified;
}

void wxMediaBuffer::SetModified(Bool mod)
{
  modified = mod;
}

void wxMediaBuffer::OnPaint(Bool pre,
			    wxDC *dc, float l, float t, float r, float b, 
			    float dx, float dy,
			    Bool show_caret)
{
  /* Do nothing */
}

/****************************************************************/

#define edf(name, action) \
     static Bool ed_##name(wxMediaBuffer *b, wxKeyEvent &event, void *) \
     { b->action; return TRUE; } \

edf(copy, Copy(FALSE, event.timeStamp))
edf(copyappend, Copy(TRUE, event.timeStamp))
edf(paste, Paste(event.timeStamp))
edf(cut, Cut(FALSE, event.timeStamp))
edf(kill, Kill(event.timeStamp))
edf(cutappend, Cut(TRUE, event.timeStamp))
edf(undo, Undo())
edf(redo, Redo())
edf(delete, Clear())

void wxMediaBuffer::AddBufferFunctions(wxKeymap *tab)
{
#define setf(name, func) \
  tab->AddKeyFunction(name, ed_##func, NULL)

  setf("copy-clipboard", copy);
  setf("copy-append-clipboard", copyappend);
  setf("cut-clipboard", cut);
  setf("cut-append-clipboard", cutappend);
  setf("paste-clipboard", paste);
  setf("delete-selection", delete);
  setf("clear-selection", delete);
  setf("delete-to-end-of-line", kill);

  setf("undo", undo);
  setf("redo", redo);
}

/****************************************************************/

wxStandardSnipAdmin::wxStandardSnipAdmin(wxMediaBuffer *m)
{
  media = m;
  WXGC_IGNORE(media);
}

wxMediaBuffer *wxStandardSnipAdmin::GetMedia(void)
{
  return media;
}

wxDC *wxStandardSnipAdmin::GetDC()
{
  return media->GetDC();
}

void wxStandardSnipAdmin::GetViewSize(float *h, float *w)
{
  media->GetViewSize(h, w);
}

Bool wxStandardSnipAdmin::ScrollTo(wxSnip *s, float localx, float localy, 
				   float w, float h, Bool refresh, int bias)
{
  return media->ScrollTo(s, localx, localy, w, h, refresh, bias);
}

void wxStandardSnipAdmin::SetCaretOwner(wxSnip *s)
{
  media->SetCaretOwner(s);
}
 
void wxStandardSnipAdmin::Resized(wxSnip *s, Bool redraw_now)
{
  media->Resized(s, redraw_now);
}

void wxStandardSnipAdmin::NeedsUpdate(wxSnip *s, float localx, float localy, 
				      float w, float h)
{
  media->NeedsUpdate(s, localx, localy, w, h);
}

Bool wxStandardSnipAdmin::ReleaseSnip(wxSnip *snip)
{
  return media->ReleaseSnip(snip);
}

void wxStandardSnipAdmin::SetCursor(wxCursor *cursor)
{
  wxMediaAdmin *admin;

  admin = media->GetAdmin();

  if (admin)
    admin->SetCursor(cursor);
}
