/*
 * File:        wx_medad.cc
 * Purpose:     wxMediaCanvas & wxDrawableMediaAdmin implementation
 * Author:      Matthew Flatt
 * Created:     1995
 * Copyright:   (c) 1995, Matthew Flatt
 */


#include "wx.h"
#include "wx_media.h"
#include "wx_ptreq.h"

#include <string.h>
#include <stdlib.h>
#include <fstream.h>
#include <ctype.h>

#define XMARGIN 5
#define YMARGIN 5

#ifdef wx_motif
#define MEDIA_CANVAS_COMBAT_CLIP_BUG
#endif

/* Correction when not using standard graphics: */
#ifdef wx_x
#if !WX_STANDARD_GRAPHICS
#define DELTA (-1)
#endif
#endif
#ifndef DELTA
#define DELTA 0
#endif

class wxDrawableMediaAdmin : public wxMediaAdmin
{
  friend class wxMediaCanvas;
  wxMediaCanvas *canvas;
  wxDrawableMediaAdmin *nextadmin, *prevadmin;
  Bool resetFlag;
  Bool updateBlock, resizedBlock;

 public:
  wxDrawableMediaAdmin(wxMediaCanvas *);
  ~wxDrawableMediaAdmin();

  /* Usually called by wxMediaBuffer objects: */
  wxDC *GetDC(float *x = NULL, float *y = NULL);
  void GetView(float *x, float *y, float *h, float *w, Bool full = FALSE);
  Bool ScrollTo(float localx, float localy, float, float, 
		Bool refresh = TRUE, int bias = 0);

  void GrabCaret(void);

  void Resized(Bool update);
  void NeedsUpdate(float localx, float localy, float w, float h);

  void SetCursor(wxCursor *cursor);

  inline void AdjustStdFlag(void);
};

class SimpleScroll
{
  Bool horizontal;

  int count, pageStep, value;

#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS
  wxCanvas *canvas;

  int trackState;

  int h, w;
  int arrow_w, bar_w, thumb_w, thumb_x;

  int wherex, wherey;
#endif
 public:
  SimpleScroll(wxCanvas *, long style = 0,
	      int length = 0, int stepsPerPage = 4,
	      int position = 0);
  ~SimpleScroll();

#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS
  void DoCallback(void);

  void Event(int x, int y, int w, int h, wxMouseEvent &);
  void Paint(int x, int y, int w, int h);

  void IntDrawLine(int x1, int y1, int x2, int y2);
  void DrawRectangle(float x, float y, float w, float h);
#endif
  void SetValue(int);
  int GetValue(void);
  void SetScroll(int len = -1, int page = -1, int position = -1);  
};

#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS
#define SB_WIDTH 14
#endif

/************************************************************************/

#ifndef wxOVERRIDE_KEY_TRANSLATIONS
#define wxOVERRIDE_KEY_TRANSLATIONS 0
#endif

wxMediaCanvas::wxMediaCanvas(wxFrame *parent, 
			     int x, int y,
			     int width, int height, 
			     char *name,
			     long style,
			     int scrollsPP,
			     wxMediaBuffer *m) 
: wxCanvas(parent, x, y, width, height, 
	   wxRETAINED + wxBORDER + wxOVERRIDE_KEY_TRANSLATIONS, name)
{
  static int type_added = FALSE;

  __type = wxTYPE_MEDIA_CANVAS;
  if (!type_added) {
    wxAllTypes.AddType(wxTYPE_MEDIA_CANVAS, wxTYPE_CANVAS, "media-canvas");
    type_added = TRUE;
  }

  givenHScrollsPerPage = scrollsPP;
  vscrollsPerPage = hscrollsPerPage = 0;

  hpixelsPerScroll = 0;

  allowXScroll = !(style  & wxMCANVAS_NO_H_SCROLL);
  allowYScroll = !(style  & wxMCANVAS_NO_V_SCROLL);
  fakeXScroll = !allowXScroll || (style & wxMCANVAS_HIDE_H_SCROLL);
  fakeYScroll = !allowYScroll || (style & wxMCANVAS_HIDE_V_SCROLL);

#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS
  hscroll = new SimpleScroll(this, wxHORIZONTAL, 0, 1, 0);
  vscroll = new SimpleScroll(this, wxVERTICAL, 0, 1, 0);
#else
  EnableScrolling(!fakeXScroll, !fakeYScroll);
  noloop = TRUE;
  SetScrollbars(fakeXScroll ? -1 : 1, fakeYScroll ? -1 : 1, 
		1, 1, 1, 1, 0, 0, FALSE);
  hscroll = fakeXScroll ? new SimpleScroll(this, wxHORIZONTAL, 0, 1, 0) : NULL;
  vscroll = fakeYScroll ? new SimpleScroll(this, wxVERTICAL, 0, 1, 0) : NULL;
  noloop = FALSE;
#endif

  media = m;

  SetBackground(wxTheBrushList->FindOrCreateBrush("white", wxSOLID));

  admin = new wxDrawableMediaAdmin(this);
  admin->standard = 1;

  SetMedia(media);
  focuson = FALSE;
  focusforcedon = FALSE;

  GetDC()->SetOptimization(TRUE);
}

wxMediaCanvas::~wxMediaCanvas()
{
  if (media) {
    if (admin->nextadmin || admin->prevadmin)
      SetMedia(NULL);
    else
      delete media;
  }

  delete admin;
}

void wxMediaCanvas::OnSize(int w, int h)
{
  if (noloop)
    return;

  ResetVisual(FALSE);

#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS
  /* Erase */
  wxBrush *brush, *oldbrush;

  brush = wxTheBrushList->FindOrCreateBrush("black", wxSOLID);
  oldbrush = GetDC()->GetBrush();
  SetBrush(brush);
  SetLogicalFunction(wxCLEAR);
  DrawRectangle(0, 0, w, h);
  SetBrush(oldbrush);
#endif

  OnPaint();
}

void wxMediaCanvas::OnFocus(Bool focus)
{
  if (focuson == focus)
    return;

  focuson = focus;
  if (media) {
    wxDrawableMediaAdmin *oldadmin;

    if (PTRNE((oldadmin = (wxDrawableMediaAdmin *)media->GetAdmin()), admin))
      media->SetAdmin(admin);

    media->OwnCaret(focus);

    if (PTRNE(oldadmin, admin))
      media->SetAdmin(oldadmin);
  }
}

void wxMediaCanvas::OnSetFocus()
{
  OnFocus(TRUE);
}

void wxMediaCanvas::OnKillFocus()
{
  OnFocus(FALSE);
}

Bool wxMediaCanvas::IsFocusOn()
{
  return focuson;
}

void wxMediaCanvas::ForceDisplayFocus(Bool on)
{
  int old = focusforcedon;
  
  focusforcedon = on;

  admin->AdjustStdFlag();

  if ((focuson || focusforcedon) != ((focuson || old)))
    OnPaint();
}

void wxMediaCanvas::OnEvent(wxMouseEvent &event)
{
#ifdef wx_msw
  if (!focuson && event.ButtonDown())
    SetFocus();
#endif

  if (media) {
#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS
    {
      int h, w;
      GetClientSize(&w, &h);
      if (allowXScroll) {
	if ((event.x > w - SB_WIDTH) && (event.y < h - SB_WIDTH)) {
	  vscroll->Event(w - SB_WIDTH, 0, SB_WIDTH, h - SB_WIDTH, event);
	  return;
	}
      }
      if (allowYScroll) {
	if ((event.y > h - SB_WIDTH) && (event.x < w - SB_WIDTH)) {
	  hscroll->Event(0, h - SB_WIDTH, w - SB_WIDTH, SB_WIDTH, event);
	  return;
	}
      }
      if ((event.y > h - SB_WIDTH) || (event.x > w - SB_WIDTH))
	return;
    }
#endif
    wxDrawableMediaAdmin *oldadmin;

    if (PTRNE((oldadmin = (wxDrawableMediaAdmin *)media->GetAdmin()), admin))
      media->SetAdmin(admin);

    event.eventObject = this;

    media->OnEvent(event);

    if (PTRNE(oldadmin, admin))
      media->SetAdmin(oldadmin);
  }
}

void wxMediaCanvas::OnChar(wxKeyEvent &event)
{
  if (media) {
    wxDrawableMediaAdmin *oldadmin;

    if (PTRNE((oldadmin = (wxDrawableMediaAdmin *)media->GetAdmin()), admin))
      media->SetAdmin(admin);

    event.eventObject = this;

    media->OnChar(event);

    if (PTRNE(oldadmin, admin))
      media->SetAdmin(oldadmin);
  }
}

void wxMediaCanvas::OnPaint(void)
{
  if (media) {
    float w, h, x, y;
    GetView(&x, &y, &w, &h);
    Redraw(x, y, w, h);
#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS
    PaintScrolls();
#endif
  } else
    Clear();
}

#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS
void wxMediaCanvas::PaintScrolls(void)
{
  int h, w;

  GetClientSize(&w, &h);
  if (!fakeXScroll)
    hscroll->Paint(0, h - SB_WIDTH, w - SB_WIDTH, SB_WIDTH);
  if (!fakeYScroll)
    vscroll->Paint(w - SB_WIDTH, 0, SB_WIDTH, h - SB_WIDTH);
}
#endif

wxDC *wxMediaCanvas::GetDCAndOffset(float *fx, float *fy)
{
  int x, y;

  if (fx || fy) {
    GetScroll(&x, &y);
    if (fx)
      *fx = x * hpixelsPerScroll - XMARGIN;
    if (fy) {
      if (media && y) {
	int h, w;
	GetClientSize(&h, &w);
	h -= 2 * YMARGIN;
#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS
	if (!fakeYScroll)
	  h -= SB_WIDTH;
#endif
	if (h < 0)
	  h = 0;
	*fy = media->ScrollLineLocation(y, h) - YMARGIN;
      } else
	*fy = -YMARGIN;
    }
  }

  return GetDC();
}

void wxMediaCanvas::GetView(float *fx, float *fy, float *fh, float *fw, 
			    Bool full)
{
  int h, w;

  GetClientSize(&h, &w);
  GetDCAndOffset(fx, fy);
  if (!full) {
    if (fx)
      *fx += XMARGIN;
    if (fy)
      *fy += YMARGIN;
  }
#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS
  if (!fakeYScroll)
    h -= SB_WIDTH;
  if (!fakeXScroll)
    w -= SB_WIDTH;
  if (h < 0)
    h = 0;
  if (w < 0)
    w = 0;
#endif
  if (full) {
    if (fh)
      *fh = h;
    if (fw)
      *fw = w;
  } else {
    if (fh) {
      if (h > 2 * YMARGIN)
	*fh = h - 2 * YMARGIN;
      else
	*fh = 0;
    }
    if (fw) {
      if (w > 2 * XMARGIN)
	*fw = w - 2 * XMARGIN;
      else
	*fw = 0;
    }
  }
}

void wxMediaCanvas::Redraw(float localx, float localy, float fw, float fh)
{
  float x, y, w, h, right, bottom;

  if (!media)
    return;

#ifdef MEDIA_CANVAS_COMBAT_CLIP_BUG
  int cliph, clipw;
  GetClientSize(&clipw, &cliph);
  SetClippingRegion(0, 0, clipw, cliph);
#endif

  GetView(&x, &y, &w, &h);

  right = x + w;
  bottom = y + h;

  if (localx > x)
    x = localx;
  if (localy > y)
    y = localy;
  
  if (right > localx + fw)
    right = localx + fw;
  if (bottom > localy + fh)
    bottom = localy + fh;

  w = right - x;
  h = bottom - y;

  if (w < 0)
    w = 0;
  if (h < 0)
    h = 0;

  if (w && h) {
    wxDrawableMediaAdmin *oldadmin;

    if (PTRNE((oldadmin = (wxDrawableMediaAdmin *)media->GetAdmin()), admin))
      media->SetAdmin(admin);
    
    media->Refresh(x, y, w, h, focuson || focusforcedon);

    if (PTRNE(oldadmin, admin))
      media->SetAdmin(oldadmin);
  }
}

Bool wxMediaCanvas::ScrollTo(float localx, float localy, float fw, float fh,
			     Bool refresh, int bias)
{
  int cy, cx, sy, sx;
  float iw, ih;
  float x, y;

  if (!media || (!allowXScroll && !allowYScroll))
    return FALSE;

  GetView(&x, &y, &iw, &ih);

  if (!iw || !ih)
    return FALSE;

  GetScroll(&cx, &cy);

  if (allowYScroll) {
    if ((fh < ih || localy + ih < y || bias == -1) && localy < y)
      sy = media->FindScrollLine(localy, ih);
    else if ((fh < ih || localy > y + fh || bias == 1) && localy + fh > y + ih)
      sy = media->FindScrollLine(localy + fh - ih, ih) + 1;
    else
      sy = cy;
  } else
    sy = cy;

  if (allowXScroll) {
    if (hpixelsPerScroll) {
      if ((fw < iw || localx + iw < x || bias == -1) && localx < x)
	sx = (int)(localx / hpixelsPerScroll);
      else if ((fw<iw || localx > x + fw || bias == 1) && localx + fw > x + iw)
	sx = (int)((localx + fw - iw) / hpixelsPerScroll) + 1;
      else
	sx = cx;
    } else
      sx = 0;
  } else
    sx = cx;

  if (sy != cy || sx != cx) {
    if (hscroll)
      hscroll->SetValue(sx);
    if (vscroll)
      vscroll->SetValue(sy);
#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS
    if (refresh)
      OnPaint();
    else
      PaintScrolls();      
#else
    Scroll(sx, sy, refresh);
#endif
    return TRUE;
  } else
    return FALSE;
}

Bool wxMediaCanvas::ResetVisual(Bool reset_scroll)
{
  int x, y, sx, sy;
  float w, h;
  int hnumScrolls, vnumScrolls, hspp, vspp;
  long tw, top;
  float totalHeight, totalWidth;

  GetScroll(&sx, &sy);

  if (media && (allowXScroll || allowYScroll)) {
    if (reset_scroll)
      x = y = 0;
    else {
      x = sx;
      y = sy;
    }
  
    GetView(NULL, NULL, &w, &h);
    media->GetExtent(&totalWidth, &totalHeight);

    if (!h || h >= totalHeight)
      vnumScrolls = 0;
    else {
      top = (long)(totalHeight - h);
      vnumScrolls = media->FindScrollLine(top, h) + 1;
    }

    if (vnumScrolls) {
      int numLines = media->NumScrollLines(h) - 1;
      vspp = (long)(((float)h * numLines) / totalHeight) - 1;
      if (vspp < 1)
	vspp = 1;
    } else
      vspp = 1;

    if (totalWidth >= w) {
      tw = (long)(totalWidth - w);

      hpixelsPerScroll = (long)(w / givenHScrollsPerPage);
      if (!hpixelsPerScroll)
	hpixelsPerScroll = 2;

      if (tw % hpixelsPerScroll)
	tw += (hpixelsPerScroll - (tw % hpixelsPerScroll));
	
      hnumScrolls = tw / hpixelsPerScroll;
      hspp = givenHScrollsPerPage;
    } else {
      hnumScrolls = 0;
      hspp = 1;
    }
  } else {
    x = y = 0;
    hnumScrolls = vnumScrolls = 0;
    vspp = hspp = 1;
    Clear();
  }    

  if (scrollWidth != hnumScrolls || scrollHeight != vnumScrolls
      || vspp != vscrollsPerPage
      || hspp != hscrollsPerPage
      || x != sx || y != sy) {
    
    if (hscroll)
      hscroll->SetScroll(hnumScrolls, hspp, x);
    if (vscroll)
      vscroll->SetScroll(vnumScrolls, vspp, y);

#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS
    PaintScrolls();
#else
    if (!fakeXScroll) {
      if (x > hnumScrolls)
	x = hnumScrolls;
      if (hspp < hscrollsPerPage)
	SetScrollPage(wxHORIZONTAL, hspp);
      if (x < sx)
	SetScrollPos(wxHORIZONTAL, x);
      if (scrollWidth != hnumScrolls)
	SetScrollRange(wxHORIZONTAL, hnumScrolls);
      if (x > sx)
	SetScrollPos(wxHORIZONTAL, x);
      if (hspp > hscrollsPerPage)
	SetScrollPage(wxHORIZONTAL, hspp);
    }

    if (!fakeYScroll) {
      if (y > vnumScrolls)
	y = vnumScrolls;
      if (vspp < vscrollsPerPage)
	SetScrollPage(wxVERTICAL, vspp);
      if (y < sy)
	SetScrollPos(wxVERTICAL, y);
      if (scrollHeight != vnumScrolls)
	SetScrollRange(wxVERTICAL, vnumScrolls);
      if (y > sy)
	SetScrollPos(wxVERTICAL, y);
      if (vspp > vscrollsPerPage)
	SetScrollPage(wxVERTICAL, vspp);
    }
#endif
    hscrollsPerPage = hspp;
    vscrollsPerPage = vspp;
    scrollWidth = hnumScrolls;
    scrollHeight = vnumScrolls;

    return TRUE;
  } else
    return FALSE;
}

void wxMediaCanvas::Scroll(int x, int y, Bool refresh)
{
  if (x > -1) {
    if (scrollWidth) {
      if (x > scrollWidth)
	x = scrollWidth;
      SetScrollPos(wxHORIZONTAL, x);
    }
  }

  if (y > -1) {
    if (scrollHeight) {
      if (y > scrollHeight)
	y = scrollHeight;
      SetScrollPos(wxVERTICAL, y);
    }
  }

  if (refresh)
    OnPaint();
}

void wxMediaCanvas::Scroll(int x, int y)
{
  Scroll(x, y, TRUE);
}

void wxMediaCanvas::GetScroll(int *x, int *y)
{
  /* Get fake scroll values if available */
  if (hscroll)
    *x = hscroll->GetValue();
  if (vscroll)
    *y = vscroll->GetValue();

#ifndef MEDIA_CANVAS_INTERNAL_SCROLLS
  if (!hscroll)
    *x = GetScrollPos(wxHORIZONTAL);
  if (!vscroll)
    *y = GetScrollPos(wxVERTICAL);
#endif
}

void wxMediaCanvas::OnScroll(wxCommandEvent &event)
{
  OnPaint();
}

wxMediaBuffer *wxMediaCanvas::GetMedia(void)
{
  return media;
}

void wxMediaCanvas::SetMedia(wxMediaBuffer *m, Bool update)
{
  if (media) {
    if (PTREQ((wxDrawableMediaAdmin *)media->GetAdmin(), admin)) {
      if (admin->nextadmin)
	media->SetAdmin(admin->nextadmin);
      else if (admin->prevadmin)
	media->SetAdmin(admin->prevadmin);
      else
	media->SetAdmin(NULL);
    }

    if (admin->nextadmin) {
      admin->nextadmin->prevadmin = admin->prevadmin;
      admin->nextadmin->AdjustStdFlag();
    }
    if (admin->prevadmin) {
      admin->prevadmin->nextadmin = admin->nextadmin;
      admin->prevadmin->AdjustStdFlag();
    }
  }
  media = m;
  if (media) {
    wxMediaAdmin *oldadmin;
    if (oldadmin = media->GetAdmin()) {
      if (!oldadmin->standard) {
	media = NULL;
	return;
      }
      admin->nextadmin = (wxDrawableMediaAdmin *)oldadmin;
      admin->prevadmin = admin->nextadmin->prevadmin;
      admin->nextadmin->prevadmin = admin;
      admin->nextadmin->AdjustStdFlag();
      if (admin->prevadmin) {
	admin->prevadmin->nextadmin = admin;
	admin->prevadmin->AdjustStdFlag();
      }
    } else {
      admin->nextadmin = admin->prevadmin = NULL;
      media->SetAdmin(admin);
      media->OwnCaret(focuson);
    }
  }

  admin->AdjustStdFlag();

  ResetVisual(TRUE);
  if (update)
    OnPaint();
}

/************************************************************************/

wxDrawableMediaAdmin::wxDrawableMediaAdmin(wxMediaCanvas *c)
{
  canvas = c;
  resetFlag = FALSE;
  nextadmin = prevadmin = NULL;
  WXGC_IGNORE(canvas);
  WXGC_IGNORE(nextadmin);
  WXGC_IGNORE(prevadmin);

  updateBlock = resizedBlock = FALSE;
}

wxDrawableMediaAdmin::~wxDrawableMediaAdmin()
{
}

wxDC *wxDrawableMediaAdmin::GetDC(float *fx, float *fy)
{
  return canvas->GetDCAndOffset(fx, fy);
}

void wxDrawableMediaAdmin::GetView(float *fx, float *fy, float *fh, float *fw, 
				   Bool full)
{
  canvas->GetView(fx, fy, fh, fw, full);
}

Bool wxDrawableMediaAdmin::ScrollTo(float localx, float localy,
				    float w, float h, Bool refresh, int bias)
{
  return canvas->ScrollTo(localx, localy, w, h, refresh, bias);
}

void wxDrawableMediaAdmin::GrabCaret(void)
{
  canvas->SetFocus();
}

void wxDrawableMediaAdmin::NeedsUpdate(float localx, float localy, 
				       float w, float h)
{
  if (updateBlock)
    return;

  updateBlock = TRUE;

  if (resetFlag) {
    canvas->OnPaint();
    resetFlag = FALSE;
  } else
    canvas->Redraw(localx, localy, w, h);

  if (nextadmin)
    nextadmin->NeedsUpdate(localx, localy, w, h);
  if (prevadmin)
    prevadmin->NeedsUpdate(localx, localy, w, h);

  updateBlock = FALSE;
}

void wxDrawableMediaAdmin::Resized(Bool update)
{
  if (resizedBlock)
    return;

  resizedBlock = TRUE;

  resetFlag = canvas->ResetVisual(FALSE) || resetFlag;
  if (update) {
    canvas->OnPaint();
    resetFlag = FALSE;
  }
  if (nextadmin)
    nextadmin->Resized(update);
  if (prevadmin)
    prevadmin->Resized(update);

  resizedBlock = FALSE;
}

void wxDrawableMediaAdmin::SetCursor(wxCursor *cursor)
{
  canvas->SetCursor(cursor);
}

void wxDrawableMediaAdmin::AdjustStdFlag(void)
{ 
  standard = (nextadmin 
	      || prevadmin 
	      || (canvas && canvas->focusforcedon)) 
    ? -1 : 1; 
}

/*************************************************************/

#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS
static wxBrush *scrollBrush = NULL;
#endif

SimpleScroll::SimpleScroll(wxCanvas *inside,
			   long style,
			   int length, int stepsPerPage,
			   int position)
{
#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS
  canvas = inside;

  if (!scrollBrush)
    scrollBrush = new wxBrush("BLACK", wxSOLID);
#endif

  horizontal = !!(style & wxHORIZONTAL);
  SetScroll(length, stepsPerPage, position);
}

SimpleScroll::~SimpleScroll()
{
}

void SimpleScroll::SetValue(int position)
{
  if (position < 0)
    position = 0;
  if (position >= count)
    position = count;

  value = position;
}

void SimpleScroll::SetScroll(int length, int stepsPerPage, int position)
{
  if (length > -1)
    count = length;
  if (stepsPerPage > 0)
    pageStep = stepsPerPage;
  if (position > -1)
    value = position;

  if (value < 0)
    value = 0;
  if (value > count)
    value = count;
}

int SimpleScroll::GetValue(void)
{
  return value;
}

#ifdef MEDIA_CANVAS_INTERNAL_SCROLLS

void SimpleScroll::IntDrawLine(int x1, int y1, int x2, int y2)
{
  if (horizontal)
    canvas->IntDrawLine(wherex + x1, wherey + y1, wherex + x2, wherey + y2);
  else
    canvas->IntDrawLine(wherex + y1, wherey + x1, wherex + y2, wherey + x2);
}

void SimpleScroll::DrawRectangle(float x, float y, float w, float h)
{
  if (horizontal)
    canvas->DrawRectangle(wherex + x, wherey + y, w, h);
  else
    canvas->DrawRectangle(wherex + y, wherey + x, h, w);
}

void SimpleScroll::Paint(int x, int y, int w, int h)
{
  int oldvalue;
  wxBrush *save;
  int seg, half;

  if (!horizontal) {
    oldvalue = w;
    w = h;
    h = oldvalue;
  }

  h -= 2;

  wherex = x + (horizontal ? 0 : 1);
  wherey = y + (horizontal ? 1 : 0);

  if (w < 2 * h)
    arrow_w = w / 2;
  else
    arrow_w = h;

  bar_w = w - 2 * arrow_w;
  if (!count) {
    thumb_w = bar_w;
    thumb_x = arrow_w;
  } else {
    thumb_w = (long)(((float)pageStep * bar_w) / (count + pageStep));
    thumb_x = arrow_w + (long)(((float)value * bar_w) / (count + pageStep));
  }
  
  save = canvas->GetDC()->GetBrush();
  canvas->SetBrush(scrollBrush);

  IntDrawLine(arrow_w, 0, arrow_w, h + 1 + DELTA);
  IntDrawLine(w - arrow_w, 0, w - arrow_w, h + 1 + DELTA);
  IntDrawLine(w + 1, 0, w + 1, h + 1 + DELTA);
  IntDrawLine(0, 0, w + 2 + DELTA, 0);
  IntDrawLine(0, h, w + 2 + DELTA, h);

  half = h / 2;
  seg = arrow_w / 4;

  IntDrawLine(3 * seg, 0, seg, half);
  IntDrawLine(3 * seg, h - 1, seg, half - 1);
  IntDrawLine(w - 3 * seg, 0, w - seg, half);
  IntDrawLine(w - 3 * seg, h - 1, w - seg, half - 1);

  if (thumb_w < bar_w) {
    if (thumb_x > arrow_w)
      DrawRectangle(arrow_w + 1, 1, thumb_x - arrow_w + DELTA, h - 1 + DELTA);
    if (value < count)
      DrawRectangle(thumb_x + thumb_w, 1, 
		    w - arrow_w - (thumb_x + thumb_w) + DELTA, h - 1 + DELTA);
  } else
    --thumb_w;

  canvas->SetLogicalFunction(wxCLEAR);
  DrawRectangle(thumb_x + 1, 1, thumb_w + DELTA, h - 1 + DELTA);
  canvas->SetLogicalFunction(wxCOPY);

  canvas->SetBrush(save);
}

void SimpleScroll::Event(int x, int y, int w, int h, wxMouseEvent &event)
{
  int oldvalue;

  enum {
    CLICK = 1,
    PAGE_DOWN,
    PAGE_UP,
    THUMB
  };

  x = (long)event.x - x;
  y = (long)event.y - y;

  if (!horizontal) {
    oldvalue = x;
    x = y;
    y = oldvalue;
    oldvalue = w;
    w = h;
    h = oldvalue;
  }

  if (event.ButtonDown()) {
    if (x < arrow_w || x > w - arrow_w)
      trackState = CLICK;
    else if (x < thumb_x)
      trackState = PAGE_UP;
    else if (x > thumb_x + thumb_w)
      trackState = PAGE_DOWN;
    else
      trackState = THUMB;
  } else if (event.Dragging()) {
  } else if (event.ButtonUp()) {
    trackState = 0;
  } else
    return;

  oldvalue = value;

  if (trackState == CLICK) {
    if (x < arrow_w)
      SetValue(value - 1);
    else if (x > w - arrow_w)
      SetValue(value + 1);
  } else if (trackState == PAGE_DOWN) {
    if (x > thumb_x + thumb_w)
      SetValue(value + pageStep);
  } else if (trackState == PAGE_UP) {
    if (x < thumb_x)
      SetValue(value - pageStep);
  } else if (trackState == THUMB) {
    x = (x - arrow_w - thumb_w / 2);
    w = bar_w - thumb_w;
    if (x < 0)
      x = 0;
    if (x > w)
      x = w;
    SetValue((int)((float)(x * count) / w));
  }

  if (value != oldvalue)
    canvas->OnPaint();
}

#endif
