#include "wx_panel.h"
#include "wx_canvs.h"
#include "wx_dcmem.h"
#include "wx_keym.h"
#include "wx_medio.h"
#include "wx_style.h"
#include "wx_mtype.h"

class wxMediaEdit;
class wxClickback;

#include "wx_snip.h"
#include "wx_cgrec.h"
#include "wx_medad.h"

class wxSnipLocation;
class wxMediaFlashTimer;

/* Edit commands */
enum {
  wxEDIT_UNDO = 1,
  wxEDIT_REDO,
  wxEDIT_CLEAR,
  wxEDIT_CUT,
  wxEDIT_COPY,
  wxEDIT_PASTE,
  wxEDIT_KILL,
  wxEDIT_INSERT_TEXT_BOX,
  wxEDIT_INSERT_GRAPHIC_BOX,
  wxEDIT_INSERT_IMAGE,
  _wx_EDIT_counter
};

/* Movement kinds */
enum {
  wxMOVE_SIMPLE = 1,
  wxMOVE_LINE,
  wxMOVE_PAGE,
  wxMOVE_WORD
};

/* File formats */
enum {
  wxMEDIA_FF_GUESS,
  wxMEDIA_FF_STD,
  wxMEDIA_FF_TEXT,
  wxMEDIA_FF_TEXT_FORCE_CR,
  wxMEDIA_FF_SAME
};

/* For Finding Wordbreaks: */
enum {
  wxBREAK_FOR_CARET = 1,
  wxBREAK_FOR_LINE = 2,
  wxBREAK_FOR_SELECTION = 4
};

/* For FindSnip: */
#define wxSNIP_BEFORE_OR_NULL (-2)
#define wxSNIP_BEFORE (-1)
#define wxSNIP_AFTER 1
#define wxSNIP_AFTER_OR_NULL 2

#define wxTAB_WIDTH 20

typedef void (*wxClickbackFunc)(wxMediaEdit *, long start, long end, void *);
typedef void (*wxWordbreakFunc)(wxMediaEdit *, long *start, long *end, 
				int reason, void *);

class wxMediaAdmin;


class wxMediaWordbreakMap : public wxObject
{
  int usage;
 public:
  char map[256];

  wxMediaWordbreakMap();

  void SetMap(int ch, int mask);
  int GetMap(int ch);

  void AdjustUsage(Bool newUser);
  Bool IsUsed(void);
};

extern wxMediaWordbreakMap wxTheMediaWordbreakMap;

class wxMediaEdit : public wxMediaBuffer
{
  friend class wxMediaLine;
 public:
  wxMediaEdit(float lineSpacing = 1.0, 
	      float *tabstops = NULL, int numtabs = 0);
  ~wxMediaEdit();

  wxMediaBuffer *CopySelf(void);

  /* Usually called by wxMediaAdmin */
  virtual void OnEvent(wxMouseEvent &event);
  virtual void OnChar(wxKeyEvent &event);
  virtual void Refresh(float localx, float localy, float w, float h, 
		       Bool show_caret);
  virtual void OwnCaret(Bool ownit);
  virtual void SizeCacheInvalid(void);

  virtual void OnDefaultEvent(wxMouseEvent &event);
  virtual void OnDefaultChar(wxKeyEvent &event);

  /* Callbacks for the wxSnipAdmin: */
  virtual Bool ScrollTo(wxSnip *, float localx, float localy, 
			float w, float h, Bool refresh, int bias = 0);
  virtual void Resized(wxSnip *, Bool redraw_now);
  virtual void NeedsUpdate(wxSnip *, float localx, float localy, 
			   float w, float h);
  virtual void SetCaretOwner(wxSnip *);
  virtual Bool ReleaseSnip(wxSnip *);

  /* Methods for you to override: */
  virtual void OnChange(void);
  virtual Bool OnInsert(long start, long len);
  virtual void AfterInsert(long start, long len);
  virtual Bool OnDelete(long start, long len);
  virtual void AfterDelete(long start, long len);
  virtual void OnEditSequence(void);
  virtual void AfterEditSequence(void);

  /* Set the caret position: */
  void GetPosition(long *start, long *end = NULL);
  long GetStartPosition(void);
  long GetEndPosition(void);
  void SetPosition(long start, long end = -1, 
		   Bool eol = FALSE, Bool scroll = TRUE);
  void MovePosition(long code, Bool extend = FALSE,
		    int kind = wxMOVE_SIMPLE);
  void SetPositionBiasScroll(int bias, long start, long end = -1, 
			     Bool eol = FALSE, Bool scroll = TRUE);

  void SetAnchor(Bool);
  Bool GetAnchor(void);

  Bool ScrollToPosition(long start, Bool ateol = FALSE, 
			long end = -1, int bias = 01);
  void GetVisiblePositionRange(long *start, long *end);
  void GetVisibleLineRange(long *start, long *end);

  /* Hilite a region without changing the selection position: */
  void FlashOn(long start, long end, Bool ateol = FALSE, 
	       Bool scroll = TRUE, long timeout = 500);
  void FlashOff();

  /* Change the text */
  void Insert(char *str, long start, long end = -1);
  void Insert(char *str);
  void Insert(long len, char *str);
  void Insert(wxSnip *snip, long start, long end = -1);
  void Insert(wxSnip *snip);
  void Insert(unsigned char ascii);
  void Insert(unsigned char ascii, long start, long end = -1);
  void Delete(long start, long end = -1);
  void Delete();
  void Erase();

  void Cut(Bool extend, long time, long start, long end = -1);
  void Cut(Bool extend = FALSE, long time = 0);
  void Copy(Bool extend, long time, long start, long end = -1);
  void Copy(Bool extend = FALSE, long time = 0);
  void Paste(long time, long start, long end = -1);
  void Paste(long time = 0);
  void Kill(long time = 0);
  void Kill(long time, long start, long end);
  void Clear(void);

  /* For making a lot of changes to be displayed at once: */
  void BeginEditSequence(Bool undoable = TRUE);
  void EndEditSequence(void);

  void ChangeStyle(wxStyleDelta *);
  void ChangeStyle(wxStyle *, long start = -1, long end = -1);
  void ChangeStyle(wxStyleDelta *, long start, long end = -1);
  /* Called automatically when a style is changed; no need to call this */
  void StyleHasChanged(wxStyle *style);

  void SetStyleList(wxStyleList *styles);

  /* Convert canvas co-ordinates to caret position */
  long FindPosition(float x, float y, Bool 
		    *ateol = NULL, Bool *onit = NULL);
  long FindLine(float y, Bool *onit = NULL);
  long FindPositionInLine(long i, float x, 
			  Bool *ateol = NULL, Bool *onit =NULL);

  /* Find the line or canvas co-ordinates for a caret position */
  long PositionLine(long start, Bool eol=FALSE);
  void PositionLocation(long start, 
			float *x = NULL, float *y = NULL, 
			Bool front = TRUE, Bool eol = FALSE,
			Bool wholeLine = FALSE);
  float LineLocation(long line, Bool top = TRUE);

  /* Get first/last caret position in a line: */
  long LineStartPosition(long i, Bool visibleOnly = TRUE);
  long LineEndPosition(long i, Bool visibleOnly = TRUE);
  long LineLength(long i);

  long LastPosition(void);
  long LastLine(void);
  
  void GetExtent(float *w, float *h);
  float GetDescent(void);
  float GetSpace(void);

  float ScrollLineLocation(long line, float displayHeight);
  long NumScrollLines(float displayHeight);
  long FindScrollLine(float y, float displayHeight);

  /* Searching */
  long FindString(char *str, int direction = 1, long start =-1, long end =-1,
		  Bool bos = TRUE, Bool caseSens = TRUE);
  long *FindStringAll(char *str, long *cnt, int direction = 1,
		      long start =-1, long end =-1, Bool bos = TRUE,
		      Bool caseSens = TRUE);
  long FindNewline(int direction = 1, long start =-1, long end =-1);

  /* Create clickable ranges: */
  void SetClickback(long start, long end, wxClickbackFunc, void *,
		    wxStyleDelta *hiliteDelta = NULL);
  void RemoveClickback(long start, long end);

  void SetClickback(wxClickback *); /* Used by undo record only */

  wxSnip *FindSnip(long p, int direction, long *sPos = NULL);
  Bool GetSnipPositionAndLocation(wxSnip *thesnip, long *pos, 
				  float *x = NULL, float *y = NULL);
  Bool GetSnipLocation(wxSnip *thesnip, float *x = NULL, float *y = NULL);
  long GetSnipPosition(wxSnip *thesnip);

  char *GetFlattenedText();
  char *GetText(long start = -1, long end = -1, 
		Bool flattened = FALSE, Bool forceCR = FALSE);
  unsigned char GetCharacter(long start);

  Bool LoadFile(char *filename = NULL, int format = wxMEDIA_FF_GUESS);
  Bool SaveFile(char *filename = NULL, int format = wxMEDIA_FF_SAME);

  Bool ReadFromFile(wxMediaStreamIn &, long start);
  Bool ReadFromFile(wxMediaStreamIn &);
  Bool WriteToFile(wxMediaStreamOut &, long start, long end = -1);
  Bool WriteToFile(wxMediaStreamOut &);

  void SetFilename(char *, Bool temp = FALSE);
  int GetFileFormat(void);
  void SetFileFormat(int);

  float *GetTabs(int *length = NULL, 
		 float *tabInc = NULL, Bool *inUnits = NULL);
  void SetTabs(float *tabs, int count, 
	       float tabIncrement = wxTAB_WIDTH, Bool inUnits = TRUE);

  void AddEditorFunctions(wxKeymap *keymap);

  void SetWordbreakFunc(wxWordbreakFunc f, void *data);
  void FindWordbreak(long *start, long *end, int reason);

  wxMediaWordbreakMap *GetWordbreakMap(void);
  void SetWordbreakMap(wxMediaWordbreakMap *map);

  void PrintToDC(wxDC *dc);

  void SetMaxWidth(float w);
  void SetMinWidth(float w);
  float GetMaxWidth();
  float GetMinWidth();
  void SetMinHeight(float h);
  void SetMaxHeight(float w);
  float GetMinHeight();
  float GetMaxHeight();

  virtual void InvalidateBitmapCache(float x=0.0, float y=0.0,
				     float w=-1.0, float h=-1.0);

  /* You might need to call this if you're doing text changes within
     BeginEditSequence() and EndEditSequence(). Hopefully, it will
     get called automatically. */
  void Recalculate(void);

  void SettingAdmin(wxMediaAdmin *);

  virtual wxBufferData *GetRegionData(long start, long end);
  virtual void SetRegionData(long start, long end, wxBufferData *);

  void PasteRegionData(wxBufferData *);

 private:
  Bool locked; /* Drawing or something - don't allow changed */

  float lineSpacing;
  float maxWidth, minWidth, minHeight, maxHeight;

  Bool hiliteOn;

  int delayRefresh;
  Bool changed; /* Set if OnChange() needs to be called */

  long len; /* Total length in "characters" == number of positions - 1 */

  long startpos, endpos;
  Bool posateol; /* display the caret at the end of a line? */
  float startloc, endloc; /* for vcursorStreak */

  Bool flash, flashautoreset;
  wxMediaFlashTimer *flashTimer;
  long flashstartpos, flashendpos;
  Bool flashposateol;
  Bool flashscroll; /* Scroll back after unflashing? */

  wxSnip *snips, *lastSnip; /* The contents of this edit session */
  long snipCount;

  wxStandardSnipAdmin *snipAdmin;
  
  class wxMediaLine *lineRoot, *firstLine, *lastLine; /* Line information */
  long numValidLines;
  Bool graphicsInvalid, flowInvalid, snipCacheInvalid, graphicMaybeInvalid;

  Bool extraLine; /* Empty line at end of file with no representative */
  float extraLineH;

  float totalHeight, totalWidth; /* Total height/width in canvas units */
  float finalDescent; /* Descent of last line */
  float initialSpace; /* Space from first line */

  wxStyle *caretStyle;

  Bool typingStreak, deletionStreak, delayedStreak, vcursorStreak;
  Bool killStreak, anchorStreak;

  Bool dragging;
  long dragstart;

  Bool tracking;
  wxClickback *trackClickback;

  Bool ibeamCursor;

  Bool refreshUnset, refreshBoxUnset, refreshAll;
  long refreshStart, refreshEnd;
  float refreshL, refreshT, refreshR, refreshB;

  Bool drawCachedInBitmap;
  float lastDrawL, lastDrawT, lastDrawR, lastDrawB;
  Bool lastDrawCaret;

  Bool caretOn;
  float caretLocationX, caretLocationT, caretLocationB;

  long delayedscroll, delayedscrollend;
  Bool delayedscrollateol;
  Bool delayedscrollbox;
  int delayedscrollbias;
  wxSnip *delayedscrollsnip;
  float delayedscrollX, delayedscrollY, delayedscrollW, delayedscrollH;

  wxBitmap *bitmap;
  wxMemoryDC *offscreen;
  long bmHeight, bmWidth;

  wxList *clickbacks;

  int fileFormat;

  float *tabs;
  int tabcount;
  float tabSpace;
  Bool tabSpaceInUnits;
  
  long readInsert, readInsertStart;

  wxWordbreakFunc wordBreak;
  void *wordBreakData;

  wxMediaWordbreakMap *wordBreakMap;

  void _Insert(wxSnip *snip, long len, char *str, long start, long end = -1);
  void _Delete(long start, long end, Bool undo);

  void _SetPosition(Bool setflash, int bias, long start, long end, 
		    Bool ateol, Bool scroll);

  void _ChangeStyle(long start, long end, wxStyle *, wxStyleDelta *);

  void MakeOnlySnip(void);
  void SpliceSnip(wxSnip *snip, wxSnip *prev, wxSnip *next);
  void InsertSnip(wxSnip *before, wxSnip *snip);
  void AppendSnip(wxSnip *snip);
  void DeleteSnip(wxSnip *snip);
  
  void MakeSnipset(long start, long end);
  wxTextSnip *InsertTextSnip(long start, wxStyle *style = NULL);
  void CheckMergeSnips(long start);

  void AdjustClickbacks(long start, long end, long d, wxDeleteRecord *rec);
  wxClickback *FindClickback(long start);
  void SetClickbackHilited(wxClickback *, Bool);

  Bool ScrollToPosition(long start, Bool ateol, Bool refresh, long end,
			int bias = 0);

  long FindFirstVisiblePosition(wxMediaLine *line, wxSnip *snip = NULL);
  void FindLastVisiblePosition(wxMediaLine *line, long *p,
			       wxSnip **snipP = NULL);

  long _FindPositionLine(long pos);
  void _CalcValidPositionLine(void);

  long _FindPositionInLine(Bool internal, long i, float x, 
			   Bool *ateol = NULL, Bool *onit =NULL);
  long _FindPositionInSnip(wxDC *dc, float X, float Y,
			   wxSnip *snip, float x);

  void LinesInvalidAfter(long);
  void OneLineInvalid(long);
  void SnipChangedAt(long);

  long _FindStringAll(char *str, int direction,
		      long start, long end, long **positions, 
		      Bool, Bool, Bool);
  
  void RecalcLines(wxDC *dc, Bool calcGraphic = TRUE);
  Bool CheckFlow(float maxw, wxDC *dc, float Y, long startp, wxSnip *start);
  Bool CheckRecalc(Bool need_graphic = TRUE);
  void Redraw(wxDC *, float, float, float, float, float, float, Bool);
  void Redraw();

  void NeedRefresh(long start, long end = -1);
  void NeedRefresh(float, float, float, float);
  void RefreshByLineDemand(void);
  void RefreshBox(float x, float y, float w, float h);

  void NeedCaretRefresh(void);
  void CaretOn(void);
  Bool CaretOff(void);
  void CalcCaretLocation(void);
  
  void IBeam(Bool);
  void EndStreaks(int exception = 0);
 protected:
  Bool ReadInsert(wxSnip *snip);
  void InsertPasteSnip(wxSnip *snip, wxBufferData *);
  void InsertPasteString(char *str);
};

#include "wx_medpb.h"

void wxInitMedia(void);
