/*
 * File:        wx_msnip.cc
 * Purpose:     wxMediaSnip implementation
 * Author:      Matthew Flatt
 * Created:     1995
 * Copyright:   (c) 1995, Matthew Flatt
 */

#include "wx.h"
#include "wx_media.h"

#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "wx_timer.h"
#include "wx_mpriv.h"

typedef struct {
  Bool drawing;
  float x, y;
  wxDC *dc;
} DrawState;

class wxSnipMediaAdmin : public wxMediaAdmin
{
  friend class wxMediaSnip;
  
  DrawState state;

  wxMediaSnip *snip;

  void RestoreState(DrawState *saved);
  void SaveState(DrawState *save, wxDC *dc, float x, float y);

 public:
  wxSnipMediaAdmin(wxMediaSnip *);

  wxDC *GetDC(float *x = NULL, float *y = NULL);
  void GetView(float *x, float *y, float *h, float *w, Bool full = FALSE);
  Bool ScrollTo(float localx, float localy, float w, float h,
			Bool refresh = TRUE, int bias = 0);
  void GrabCaret(void);

  void Resized(Bool redraw_now);
  void NeedsUpdate(float localx, float localy, float w, float h);

  void SetCursor(wxCursor *cursor);
};

wxMediaSnip::wxMediaSnip(wxMediaBuffer *useme,
			 Bool border,
			 int lm, int tm, int rm, int bm,
			 int li, int ti, int ri, int bi,
			 float w, float W, float h, float H)
: wxSnip()
{
  Bool istemp;

  flags |= wxSNIP_HANDLES_EVENTS;

  snipclass = wxTheSnipClassList.Find("wxmedia");

  withBorder = border;
  leftMargin = lm;
  topMargin = tm;
  rightMargin = rm;
  bottomMargin = bm;
  leftInset = li;
  topInset = ti;
  rightInset = ri;
  bottomInset = bi;

  minWidth = w;
  maxWidth = W;
  minHeight = h;
  maxHeight = H;

  if (useme)
    me = useme;
  else
    me = new wxMediaEdit();
  myAdmin = new wxSnipMediaAdmin(this);

  if (!me->GetFilename(&istemp) || istemp)
    /* Turn on flag to mirror filename: */
    flags |= wxSNIP_USES_BUFFER_PATH;

  me->OwnCaret(FALSE);
}

wxMediaSnip::~wxMediaSnip()
{
  delete me;
  delete myAdmin;
}

void wxMediaSnip::SetAdmin(wxSnipAdmin *a)
{
  if (PTRNE(a, admin)) {
    wxSnip::SetAdmin(a);
    if (a)
      me->SetAdmin(myAdmin);
    else
      me->SetAdmin(NULL);
  } 

  if (admin && (flags & wxSNIP_USES_BUFFER_PATH)) {
    /* Propogate a filename change: */
    Bool istemp;
    if (!me->GetFilename(&istemp) || istemp) {
      wxMediaBuffer *b;
      b = admin->GetMedia();
      if (b) {
	char *filename;
	filename = b->GetFilename();
	me->SetFilename(filename, TRUE);
      }
    } else
      flags -= wxSNIP_USES_BUFFER_PATH; /* Turn off the flag; not needed */
  }
}

wxMediaBuffer *wxMediaSnip::GetThisMedia(void)
{
  return me;
}

void wxMediaSnip::OnEvent(wxDC *dc, float x, float y, 
			  float, float, wxMouseEvent &event)
{
  DrawState save;

  myAdmin->SaveState(&save, dc, x, y);
  me->OnEvent(event);
  myAdmin->RestoreState(&save);
}

void wxMediaSnip::OnChar(wxDC *dc, float x, float y, 
			  float, float, wxKeyEvent &event)
{
  DrawState save;

  myAdmin->SaveState(&save, dc, x, y);
  me->OnChar(event);
  myAdmin->RestoreState(&save);
}

void wxMediaSnip::OwnCaret(Bool ownit)
{
  me->OwnCaret(ownit);
}

void wxMediaSnip::DoEdit(int op, Bool recursive, long time)
{
  me->DoEdit(op, recursive, time);
}

void wxMediaSnip::DoFont(int op, Bool recursive)
{
  me->DoFont(op, recursive);
}

Bool wxMediaSnip::Match(wxSnip *other)
{
  return FALSE;
}

void wxMediaSnip::SizeCacheInvalid(void)
{
  me->SizeCacheInvalid();
}

char *wxMediaSnip::GetText(long offset, long num, Bool flat)
{
  if (!flat) {
    char *s = new char[2];
    s[0] = '.';
    s[1] = 0;
    return s;
  } else
    return me->GetFlattenedText();
}


void wxMediaSnip::GetExtent(wxDC *dc, 
			    float x, float y,
			    float *w, float *h,
			    float *descent, float *space,
			    float *lspace, float *rspace)
{
  DrawState save;

  myAdmin->SaveState(&save, dc, x, y);
  
  me->GetExtent(w, h);

  if (w) {
    if (*w < minWidth)
      *w = minWidth;
    else if (maxWidth > 0 && *w > maxWidth)
      *w = maxWidth;

    *w += rightMargin + leftMargin;
  }
  if (h) {
    if (*h < minHeight)
      *h = minHeight;
    else if (maxHeight > 0 && *h > maxHeight)
      *h = maxHeight;

    *h += topMargin + bottomMargin;
  }

  if (descent)
    *descent = me->GetDescent() + bottomMargin;
  if (space)
    *space = me->GetSpace() + topMargin;
  if (lspace)
    *lspace = leftMargin;
  if (rspace)
    *rspace = rightMargin;

  myAdmin->RestoreState(&save);
}
	       
void wxMediaSnip::Draw(wxDC *dc, float x, float y, 
		       float left, float top, float right, float bottom,
		       float dx, float dy, Bool show_caret)
{
  float w, h, r, b;

  DrawState save;

  myAdmin->SaveState(&save, dc, x, y);
  
  me->GetExtent(&w, &h);

  if (w < minWidth)
    w = minWidth;
  else if (maxWidth > 0 && w > maxWidth)
    w = maxWidth;

  if (h < minHeight)
    h = minHeight;
  else if (maxHeight > 0 && h > maxHeight)
    h = maxHeight;

  if (withBorder) {
    wxBrush *oldbk;
    static wxBrush *bk = NULL;
    if (!bk)
      bk = wxTheBrushList->FindOrCreateBrush("WHITE", wxTRANSPARENT);
    oldbk = dc->GetBrush();
    dc->SetBrush(bk);
    dc->DrawRectangle(x + leftInset,
		      y + topInset,
		      w + leftMargin + rightMargin -(leftInset+rightInset)-1, 
		      h + topMargin + bottomMargin -(topInset+bottomInset)-1);
    dc->SetBrush(oldbk);
  }

  x += leftMargin;
  y += topMargin;
  r = x + w;
  b = y + h;

  if (x > left)
    left = x;
  if (y > top)
    top = y;
  if (r < right)
    right = r;
  if (b < bottom)
    bottom = b;

  me->Refresh(left - x, top - y, right - left, bottom - top, show_caret);

  myAdmin->RestoreState(&save);
}

wxSnip *wxMediaSnip::Copy(void)
{
  return new wxMediaSnip(me->CopySelf(), 
			 withBorder,
			 leftMargin, topMargin,
			 rightMargin, bottomMargin,
			 leftInset, topInset,
			 rightInset, bottomInset,
			 minWidth, maxWidth,
			 minHeight, maxHeight);
}

void wxMediaSnip::Write(wxMediaStreamOut &f)
{
  f << me->bufferType
    << withBorder
    << leftMargin << topMargin << rightMargin << bottomMargin
    << leftInset << topInset << rightInset << bottomInset
    << minWidth << maxWidth
    << minHeight << maxHeight;

  me->WriteToFile(f);
}

void wxMediaSnip::SetMaxWidth(float w)
{
  maxWidth = w;
  admin->Resized(this, TRUE);
}

void wxMediaSnip::SetMinWidth(float w)
{
  minWidth = w;
  admin->Resized(this, TRUE);
}

void wxMediaSnip::SetMaxHeight(float h)
{
  maxHeight = h;
  admin->Resized(this, TRUE);
}

void wxMediaSnip::SetMinHeight(float h)
{
  minHeight = h;
  admin->Resized(this, TRUE);
}

float wxMediaSnip::GetMaxWidth(void) { return maxWidth; }
float wxMediaSnip::GetMaxHeight(void) { return maxHeight; }
float wxMediaSnip::GetMinWidth(void) { return minWidth; }
float wxMediaSnip::GetMinHeight(void) { return minHeight; }

Bool wxMediaSnip::Resize(float w, float h)
{
  w -= leftMargin + rightMargin;
  h -= topMargin + bottomMargin;
  if (w < 0)
    w = 0;
  if (h < 0)
    h = 0;
  minWidth = maxWidth = w;
  minHeight = maxHeight = h;

  me->SetMaxWidth(w);
  me->SetMinWidth(w);

  admin->Resized(this, TRUE);

  return TRUE;
}

void wxMediaSnip::ShowBorder(Bool show)
{
  if (withBorder != show) {
    withBorder = show;
    if (admin) {
      wxDC *dc;
      float w, h;

      dc = admin->GetDC();
      if (dc) {
	GetExtent(dc, 0, 0, &w, &h);
	admin->NeedsUpdate(this, leftInset, topInset, 
			   w + rightMargin - rightInset, 
			   h + bottomMargin - bottomInset);
      }
    }
  }
}

Bool wxMediaSnip::BorderVisible()
{
  return withBorder;
}

void wxMediaSnip::SetMargin(int lm, int tm, int rm, int bm)
{
  leftMargin = lm;
  topMargin = tm;
  rightMargin = rm;
  bottomMargin = bm;
  
  if (admin)
    admin->Resized(this, TRUE);
}

void wxMediaSnip::GetMargin(int *lm, int *tm, int *rm, int *bm)
{
  *lm = leftMargin;
  *tm = topMargin;
  *rm = rightMargin;
  *bm = bottomMargin;
}

void wxMediaSnip::SetInset(int lm, int tm, int rm, int bm)
{
  leftMargin = lm;
  topMargin = tm;
  rightMargin = rm;
  bottomMargin = bm;

  if (admin) {
    wxDC *dc;
    float w, h;

    dc = admin->GetDC();
    if (dc) {
      GetExtent(dc, 0, 0, &w, &h);
      admin->NeedsUpdate(this, 0, 0, 
			 w + rightMargin + leftMargin,
			 h + bottomMargin + topMargin);
    }
  }
}

void wxMediaSnip::GetInset(int *li, int *ti, int *ri, int *bi)
{
  *li = leftInset;
  *ti = topInset;
  *ri = rightInset;
  *bi = bottomInset;
}

long wxMediaSnip::GetNumScrollSteps()
{
  if (admin) {
    float h;

    admin->GetViewSize(NULL, &h);

    h -= (topMargin + bottomMargin);

    return me->NumScrollLines(h);
  } else
    return 1;
}

long wxMediaSnip::FindScrollStep(float y)
{
  if (admin) {
    float h;

    admin->GetViewSize(NULL, &h);

    h -= (topMargin + bottomMargin);

    return me->FindScrollLine(y - topMargin, h);
  } else
    return 0;

}

float wxMediaSnip::GetScrollStepOffset(long i)
{
  if (admin) {
    float h;

    admin->GetViewSize(NULL, &h);

    h -= (topMargin + bottomMargin);

    return me->ScrollLineLocation(i, h) + topMargin;
  } else
    return 0;
}

/****************************************************************/

wxSnipMediaAdmin::wxSnipMediaAdmin(wxMediaSnip *s)
{
  snip = s;
  state.drawing = 0;

  WXGC_IGNORE(snip);
  WXGC_IGNORE(state.dc);
}

wxDC *wxSnipMediaAdmin::GetDC(float *xp, float *yp)
{
  if (state.drawing) {
    if (xp)
      *xp = -state.x;
    if (yp)
      *yp = -state.y;
  } else {
    if (xp)
      *xp = 0;
    if (yp)
      *yp = 0;
  }
  
  return state.drawing ? state.dc : snip->admin->GetDC();
}

void wxSnipMediaAdmin::SaveState(DrawState *save, wxDC *dc, float x, float y)
{
  save->drawing = state.drawing;
  save->dc = state.dc;
  save->x = state.x;
  save->y = state.y;

  state.drawing = TRUE;
  state.x = x + snip->leftMargin;
  state.y = y + snip->topMargin;
  state.dc = dc;
}

void wxSnipMediaAdmin::RestoreState(DrawState *saved)
{
  state.drawing = saved->drawing;
  state.dc = saved->dc;
  state.x = saved->x;
  state.y = saved->y;
}

void wxSnipMediaAdmin::GetView(float *x, float *y, float *w, float *h, 
			       Bool full)
{
  if (x)
    *x = state.x;
  if (y)
    *y = state.y;

  snip->admin->GetViewSize(w, h);

  if (w && snip->maxWidth > 0 && *w > snip->maxWidth)
    *w = snip->maxWidth;

  if (h && snip->maxHeight > 0 && *h > snip->maxHeight)
    *h = snip->maxHeight;
}

Bool wxSnipMediaAdmin::ScrollTo(float localx, float localy, float w, float h,
				Bool refresh, int bias)
{
  return snip->admin->ScrollTo(snip, localx + snip->leftMargin,
			       localy + snip->topMargin,
			       w, h, refresh, bias);
}
	
void wxSnipMediaAdmin::GrabCaret(void)
{
  snip->admin->SetCaretOwner(snip);
}

void wxSnipMediaAdmin::Resized(Bool redraw_now)
{
  snip->admin->Resized(snip, redraw_now);
}

void wxSnipMediaAdmin::NeedsUpdate(float localx, float localy, 
				   float w, float h)
{
  snip->admin->NeedsUpdate(snip, 
			   localx + snip->leftMargin,
			   localy + snip->topMargin,
			   w, h);
}

void wxSnipMediaAdmin::SetCursor(wxCursor *cursor)
{
  snip->admin->SetCursor(cursor);
}


/************************************************************************/

#define edf(name, action) \
     static Bool ed_##name(wxMediaEdit *style, wxKeyEvent &event, wxObject *) \
     { style->action; return TRUE; } \

edf(right, MovePosition(WXK_RIGHT))
edf(left, MovePosition(WXK_LEFT))
edf(up, MovePosition(WXK_UP))
edf(down, MovePosition(WXK_DOWN))
edf(pageup, MovePosition(WXK_UP, FALSE, wxMOVE_PAGE))
edf(pagedown, MovePosition(WXK_DOWN, FALSE, wxMOVE_PAGE))
edf(rightword, MovePosition(WXK_RIGHT, FALSE, wxMOVE_WORD))
edf(leftword, MovePosition(WXK_LEFT, FALSE, wxMOVE_WORD))

edf(selectright, MovePosition(WXK_RIGHT, TRUE))
edf(selectleft, MovePosition(WXK_LEFT, TRUE))
edf(selectup, MovePosition(WXK_UP, TRUE))
edf(selectdown, MovePosition(WXK_DOWN, TRUE))
edf(selectpageup, MovePosition(WXK_UP, TRUE, wxMOVE_PAGE))
edf(selectpagedown, MovePosition(WXK_DOWN, TRUE, wxMOVE_PAGE))
edf(selectrightword, MovePosition(WXK_RIGHT, TRUE, wxMOVE_WORD))
edf(selectleftword, MovePosition(WXK_LEFT, TRUE, wxMOVE_WORD))

edf(startofline, MovePosition(WXK_LEFT, FALSE, wxMOVE_LINE))
edf(endofline, MovePosition(WXK_RIGHT, FALSE, wxMOVE_LINE))
edf(home, MovePosition(WXK_HOME))
edf(end,  MovePosition(WXK_END))

edf(selecttostartofline, MovePosition(WXK_LEFT, TRUE, wxMOVE_LINE))
edf(selecttoendofline, MovePosition(WXK_RIGHT, TRUE, wxMOVE_LINE))
edf(selecttohome, MovePosition(WXK_HOME, TRUE))
edf(selecttoend,  MovePosition(WXK_END, TRUE))

edf(clear, Erase())
edf(delete, Delete())

static Bool ed_deletenext(wxMediaEdit *style, wxKeyEvent &, wxObject *)
{
  long s, e;
  style->GetPosition(&s, &e);
  if (s != e)
    style->Delete();
  else
    style->Delete(s, s + 1);
  return TRUE;
}

static Bool ed_deletenextword(wxMediaEdit *style, wxKeyEvent &event, wxObject *)
{
  style->BeginEditSequence();
  ed_selectrightword(style, event, NULL);
  ed_delete(style, event, NULL);
  style->EndEditSequence();
  return TRUE;
}

static Bool ed_deleteprevword(wxMediaEdit *style, wxKeyEvent &event, wxObject *)
{
  style->BeginEditSequence();
  ed_selectleftword(style, event, NULL);
  ed_delete(style, event, NULL);
  style->EndEditSequence();
  return TRUE;
}

static Bool ed_deleteline(wxMediaEdit *style, wxKeyEvent &event, wxObject *)
{
  style->BeginEditSequence();
  ed_startofline(style, event, NULL);
  ed_selecttoendofline(style, event, NULL);
  ed_delete(style, event, NULL);
  style->EndEditSequence();
  return TRUE;
}

void wxMediaEdit::AddEditorFunctions(wxKeymap *tab)
{
#define setf(name, func) \
  tab->AddKeyFunction(name, (wxKeyFunction)ed_##func, NULL)

  setf("forward-character", right);
  setf("backward-character", left);
  setf("previous-line", up);
  setf("next-line", down);
  setf("previous-page", pageup);
  setf("next-page", pagedown);
  setf("forward-word", rightword);
  setf("backward-word", leftword);

  setf("forward-select", selectright);
  setf("backward-select", selectleft);
  setf("select-down", selectdown);
  setf("select-up", selectup);
  setf("select-page-up", selectpageup);
  setf("select-page-down", selectpagedown);
  setf("forward-select-word", selectrightword);
  setf("backward-select-word", selectleftword);

  setf("beginning-of-file", home);
  setf("end-of-file", end);
  setf("beginning-of-line", startofline);
  setf("end-of-line", endofline);

  setf("select-to-beginning-of-file", selecttohome);
  setf("select-to-end-of-file", selecttoend);
  setf("select-to-beginning-of-line", selecttostartofline);
  setf("select-to-end-of-line", selecttoendofline);

  setf("delete-previous-character", delete);
  setf("delete-next-character", deletenext);
  setf("clear-buffer", clear);
  setf("delete-next-word", deletenextword);
  setf("delete-previous-word", deleteprevword);

  setf("delete-line", deleteline);

  AddBufferFunctions(tab);
}

/*************************************************************/

wxMediaWordbreakMap::wxMediaWordbreakMap()
{
  int i;

  usage = 0;
  memset(map, 0, sizeof(map));

  for (i = 0; i < 256; i++) {
    if (isalnum(i))
      map[i] = wxBREAK_FOR_CARET | wxBREAK_FOR_LINE | wxBREAK_FOR_SELECTION;
    else if (!isspace(i))
      map[i] = wxBREAK_FOR_LINE;
  }

  map['-'] -= wxBREAK_FOR_LINE;
}

void wxMediaWordbreakMap::SetMap(int ch, int mask)
{
  map[ch] = mask;
}

int wxMediaWordbreakMap::GetMap(int ch)
{
  return map[ch];
}

void wxMediaWordbreakMap::AdjustUsage(Bool newUser)
{
  if (newUser)
    usage++;
  else
    --usage;
}

Bool wxMediaWordbreakMap::IsUsed(void)
{
  return !!usage;
}
