/*
 * File:        wx_style.cc
 * Purpose:     wxStyle and wxStyleList implementation
 * Author:      Matthew Flatt
 * Created:     1995
 * Copyright:   (c) 1995, Matthew Flatt
 */

#include "wx.h"
#include "wx_style.h"
#include "wx_medio.h"
#include "wx_ptreq.h"

static int defaultSize = 12;

wxStyleList *wxTheStyleList;

static wxColour *whiteColour, *blackColour;

void wxInitStyles(void)
{
  if (wxTheStyleList)
    return;

  whiteColour = wxTheColourDatabase->FindColour("WHITE");
  blackColour = wxTheColourDatabase->FindColour("BLACK");

#if USE_RESOURCES
  wxGetResource(wxTheApp->wx_class, "defaultFontSize", &defaultSize);
#endif

  wxTheStyleList = new wxStyleList;
}

void wxMultColour::Get(float *rf, float *gf, float *bf)
{
  *rf = r;
  *gf = g;
  *bf = b;
}

void wxMultColour::Set(float rf, float gf, float bf)
{
  r = rf;
  g = gf;
  b = bf;
}

void wxAddColour::Get(short *rf, short *gf, short *bf)
{
  *rf = r;
  *gf = g;
  *bf = b;
}

void wxAddColour::Set(short rf, short gf, short bf)
{
  r = rf;
  g = gf;
  b = bf;
}

wxStyleDelta::wxStyleDelta(int changeCommand, int param)
{
  face = NULL;
    
  SetDelta(wxCHANGE_NOTHING);
  SetDelta(changeCommand, param);
}

wxStyleDelta::~wxStyleDelta()
{
  if (face)
    delete[] face;
}

wxStyleDelta *wxStyleDelta::SetDelta(int changeCommand, int param)
{
  switch (changeCommand) {
  case wxCHANGE_NOTHING:
    family = wxBASE;
    if (face) {
      delete[] face;
      face = NULL;
    }
    sizeMult = 1;
    sizeAdd = 0;
    weightOn = wxBASE;
    weightOff = wxBASE;
    styleOn = wxBASE;
    styleOff = wxBASE;
    underlinedOn = underlinedOff = FALSE;
    foregroundMult.Set(1, 1, 1);
    foregroundAdd.Set(0, 0, 0);
    backgroundMult.Set(1, 1, 1);
    backgroundAdd.Set(0, 0, 0);
    alignmentOn = wxBASE;
    alignmentOff = wxBASE;
    break;
  case wxCHANGE_STYLE:
    styleOn = param;
    styleOff = wxBASE;
    break;
  case wxCHANGE_WEIGHT:
    weightOn = param;
    weightOff = wxBASE;
    break;
  case wxCHANGE_UNDERLINE:
    underlinedOn = param;
    underlinedOff = !param;
    break;
  case wxCHANGE_SIZE:
    sizeMult = 0;
    sizeAdd = param;
    break;
  case wxCHANGE_FAMILY:
    family = param;
    if (face) {
      delete[] face;
      face = NULL;
    }
    break;
  case wxCHANGE_ALIGNMENT:
    alignmentOn = param;
    alignmentOff = wxBASE;
    break;
  case wxCHANGE_BOLD:
    weightOn = wxBOLD;
    weightOff = wxBASE;
    break;
  case wxCHANGE_ITALIC:
    styleOn = wxITALIC;
    styleOff = wxBASE;
    break;
  case wxCHANGE_TOGGLE_STYLE:
    styleOn = param;
    styleOff = param;
    break;
  case wxCHANGE_TOGGLE_WEIGHT:
    weightOn = param;
    weightOff = param;
    break;
  case wxCHANGE_TOGGLE_UNDERLINE:
    underlinedOn = TRUE;
    underlinedOff = TRUE;
    break;
  case wxCHANGE_BIGGER:
    sizeMult = 1;
    sizeAdd = param;
    break;
  case wxCHANGE_SMALLER:
    sizeMult = 1;
    sizeAdd = -param;
    break;
  case wxCHANGE_NORMAL:
    family = wxDEFAULT;
    if (face) {
      delete[] face;
      face = NULL;
    }
    sizeMult = 0;
    sizeAdd = defaultSize;
    weightOn = wxNORMAL;
    weightOff = wxBASE;
    styleOn = wxNORMAL;
    styleOff = wxBASE;
    underlinedOn = FALSE;
    underlinedOff = TRUE;
    alignmentOn = wxALIGN_BOTTOM;
    alignmentOff = wxBASE;
    /* fall through ... */
  case wxCHANGE_NORMAL_COLOUR: /* ^^ falls through */
    foregroundMult.Set(0, 0, 0);
    foregroundAdd.Set(0, 0, 0);
    backgroundMult.Set(0, 0, 0);
    backgroundAdd.Set(255, 255, 255);
    break;
  }

  return this;
}

wxStyleDelta *wxStyleDelta::SetDeltaFace(char *name)
{
  if (face)
    delete[] face;
  face = copystring(name);
#ifdef NO_GENERAL_FONTS
  family = wxBASE;
#else
  int id = wxTheFontNameDirectory.FindOrCreateFontId(name, wxDEFAULT);
  family = wxTheFontNameDirectory.GetFamily(id);
#endif
  return this;
}

wxStyleDelta *wxStyleDelta::SetDeltaBackground(char *name)
{
  wxColour *c;

  if (c = wxTheColourDatabase->FindColour(name))
    SetDeltaBackground(*c);
  return this;
}

wxStyleDelta *wxStyleDelta::SetDeltaBackground(wxColour& colour)
{
  unsigned char r, g, b;

  backgroundMult.Set(0, 0, 0);
  colour.Get(&r, &g, &b);
  backgroundAdd.Set(r, g, b);
  return this;
}

wxStyleDelta *wxStyleDelta::SetDeltaForeground(char *name)
{
  wxColour *c;

  if (c = wxTheColourDatabase->FindColour(name))
    SetDeltaForeground(*c);
  return this;
}

wxStyleDelta *wxStyleDelta::SetDeltaForeground(wxColour& colour)
{
  unsigned char r, g, b;

  foregroundMult.Set(0, 0, 0);
  colour.Get(&r, &g, &b);
  foregroundAdd.Set(r, g, b);
  return this;
}

Bool wxStyleDelta::Collapse(wxStyleDelta &deltaIn)
{
  float ambr, ambb, ambg, amfr, amfb, amfg;
  float bmbr, bmbb, bmbg, bmfr, bmfb, bmfg;
  short aabr, aabb, aabg, aafr, aafb, aafg;
  short babr, babb, babg, bafr, bafb, bafg;

  /* If collapsing possible? */
  /* It may not be if add & multiply sequence occurs, */
  /* or certain toggling settings conflict or */
  if (sizeMult != 1 && deltaIn.sizeAdd != 0)
    return FALSE;

  foregroundMult.Get(&amfr, &amfb, &amfg);
  backgroundMult.Get(&ambr, &ambb, &ambg);
  deltaIn.foregroundAdd.Get(&bafr, &bafb, &bafg);
  deltaIn.backgroundAdd.Get(&babr, &babb, &babg);
  if (amfr != 1 && bafr != 0
      || amfg != 1 && bafg != 0
      || amfb != 1 && bafb != 0
      || ambr != 1 && babr != 0
      || ambg != 1 && babg != 0
      || ambb != 1 && babb != 0)
    return FALSE;

  // Cases: simple or double toggle
  //        no further change
  //        formerly no change
  //        style definitely on
  //        style definitely off
  if (!((styleOn == deltaIn.styleOn && styleOff == deltaIn.styleOff)
	|| (styleOn == wxBASE && styleOff == wxBASE)
	|| (deltaIn.styleOn == wxBASE && deltaIn.styleOff == wxBASE)
	|| (styleOn == wxBASE && styleOff != wxBASE)
	|| (styleOff == wxBASE && styleOn != wxBASE)))
    return FALSE;
  if (!((weightOn == deltaIn.weightOn && weightOff == deltaIn.weightOff)
	|| (weightOn == wxBASE && weightOff == wxBASE)
	|| (weightOn == wxBASE && weightOff != wxBASE)
	|| (weightOff == wxBASE && weightOn != wxBASE)))
    return FALSE;
  if (!((alignmentOn == deltaIn.alignmentOn 
	 && alignmentOff == deltaIn.alignmentOff)
	|| (alignmentOn == wxBASE && alignmentOff == wxBASE)
	|| (alignmentOn == wxBASE && alignmentOff != wxBASE)
	|| (alignmentOff == wxBASE && alignmentOn != wxBASE)))
    return FALSE;

  if (!((underlinedOn == deltaIn.underlinedOn
	 && underlinedOff == deltaIn.underlinedOff)
	|| (!underlinedOn && !underlinedOff)
	|| (!deltaIn.underlinedOn && !deltaIn.underlinedOff)
	|| (!underlinedOn && underlinedOff)
	|| (!underlinedOff && underlinedOn)))
    return FALSE;

  /* Collapsing is possible. */

  deltaIn.foregroundMult.Get(&bmfr, &bmfb, &bmfg);
  deltaIn.backgroundMult.Get(&bmbr, &bmbb, &bmbg);
  foregroundAdd.Get(&aafr, &aafb, &aafg);
  backgroundAdd.Get(&aabr, &aabb, &aabg);
  
  sizeMult *= deltaIn.sizeMult;
  sizeAdd += deltaIn.sizeAdd;

  foregroundMult.Set(amfr * bmfr, amfb * bmfb, amfg * amfg);
  backgroundMult.Set(ambr * bmbr, ambb * bmbb, ambg * bmbg);
  foregroundAdd.Set(aafr + bafr, aafb + bafb, aafg + bafg);
  backgroundAdd.Set(aabr + babr, aabb + babb, aabg + babg);

  if (family == wxBASE) {
    family = deltaIn.family;
    if (deltaIn.face)
      face = copystring(deltaIn.face);
  }

  if (styleOn == wxBASE && styleOff == wxBASE) {
    styleOff = deltaIn.styleOff;
    styleOn = deltaIn.styleOn;
  } else if (styleOn != wxBASE && styleOff != wxBASE) {
    if (deltaIn.styleOn != wxBASE || deltaIn.styleOff != wxBASE
	&& styleOn == styleOff)
      styleOn = styleOff = wxBASE; // Double toggle
  }
  if (weightOn == wxBASE && weightOff == wxBASE) {
    weightOff = deltaIn.weightOff;
    weightOn = deltaIn.weightOn;
  } else if (weightOn != wxBASE && weightOff != wxBASE) {
    if (deltaIn.weightOn != wxBASE || deltaIn.weightOff != wxBASE
	&& weightOn == weightOff)
      weightOn = weightOff = wxBASE; // Double toggle
  }
  if (alignmentOn == wxBASE && alignmentOff == wxBASE) {
    alignmentOff = deltaIn.alignmentOff;
    alignmentOn = deltaIn.alignmentOn;
  } else if (alignmentOn != wxBASE && alignmentOff != wxBASE) {
    if (deltaIn.alignmentOn != wxBASE || deltaIn.alignmentOff != wxBASE
	&& alignmentOn == alignmentOff)
      alignmentOn = alignmentOff = wxBASE; // Double toggle
  }
  if (!underlinedOn && !underlinedOff) {
    underlinedOn = deltaIn.underlinedOn;
    underlinedOff = deltaIn.underlinedOff;
  } else if (underlinedOn && underlinedOff) {
    if (deltaIn.underlinedOn && deltaIn.underlinedOff)
      underlinedOn = underlinedOff = FALSE;
  }

  return TRUE;
}

Bool wxStyleDelta::Equal(wxStyleDelta &deltaIn)
{
  float ambr, ambb, ambg, amfr, amfb, amfg;
  float bmbr, bmbb, bmbg, bmfr, bmfb, bmfg;
  short aabr, aabb, aabg, aafr, aafb, aafg;
  short babr, babb, babg, bafr, bafb, bafg;
  
  foregroundMult.Get(&amfr, &amfb, &amfg);
  backgroundMult.Get(&ambr, &ambb, &ambg);
  foregroundAdd.Get(&aafr, &aafb, &aafg);
  backgroundAdd.Get(&aabr, &aabb, &aabg);
  deltaIn.foregroundMult.Get(&bmfr, &bmfb, &bmfg);
  deltaIn.backgroundMult.Get(&bmbr, &bmbb, &bmbg);
  deltaIn.foregroundAdd.Get(&bafr, &bafb, &bafg);
  deltaIn.backgroundAdd.Get(&babr, &babb, &babg);

  return (((!face && !deltaIn.face && family == deltaIn.family)
	   || (face && deltaIn.face && !strcmp(face, deltaIn.face)))
	  && sizeMult == deltaIn.sizeMult
	  && sizeAdd == deltaIn.sizeAdd
	  && weightOn == deltaIn.weightOn
	  && weightOff == deltaIn.weightOff
	  && styleOn == deltaIn.styleOn
	  && styleOff == deltaIn.styleOff
	  && underlinedOn == deltaIn.underlinedOn
	  && underlinedOff == deltaIn.underlinedOff
	  && amfr == bmfr && amfb == bmfb && amfg == bmfg
	  && aafr == bafr && aafb == bafb && aafg == bafg
	  && ambr == bmbr && ambb == bmbb && ambg == bmbg
	  && aabr == babr && aabb == babb && aabg == babg
	  && alignmentOn == deltaIn.alignmentOn
	  && alignmentOff == deltaIn.alignmentOff);
}

/***************************************************************/

static unsigned char ColourNum(float v)
{
  if (v < 0)
    return 0;
  else if (v > 255)
    return 255;
  else
    return (unsigned char)v;
}

wxStyle::wxStyle()
{
  WXGC_IGNORE(styleList);
  joinStyle = FALSE;
}

wxStyle::~wxStyle()
{
  if (!joinStyle)
    delete u.delta;
}

void wxStyle::Update(wxStyle *basic, wxStyle *target, 
		     Bool propogate, Bool topLevel)
{
  int size;
  int fontid;
  int style, weight;
  unsigned char r, g, b;
  float rm, gm, bm;
  short rp, gp, bp; 
  Bool match;
  wxNode *node;
  Bool underlined;
  wxStyle *base;

  base = baseStyle;
  if (basic) {
    if (PTREQ(base, styleList->BasicStyle())) {
      base = basic;
    } else {
      base->Update(basic, target, FALSE, FALSE);
      base = target;
    }
  }

  if (!target)
    target = this;

  if (joinStyle) {
    u.shiftStyle->Update(base, target, FALSE, topLevel);
    return;
  }

  size = (int)(u.delta->sizeMult * base->font->GetPointSize());
  size += u.delta->sizeAdd;
  if (size <= 0)
    size = 1;

  if (!u.delta->face && u.delta->family == wxBASE) {
#ifndef NO_GENERAL_FONTS
    fontid = base->font->GetFontId();
#else
    fontid = base->font->GetFamily();
#endif
  }
#ifndef NO_GENERAL_FONTS
  else if (u.delta->face)
    fontid = wxTheFontNameDirectory.FindOrCreateFontId(u.delta->face, 
						       u.delta->family);
#endif
  else
    fontid = u.delta->family;

  style = base->font->GetStyle();
  match = (style == u.delta->styleOff);
  if (match)
    style = wxNORMAL;
  if (!match || (match && u.delta->styleOn != u.delta->styleOff))
    if (u.delta->styleOn != wxBASE)
      style = u.delta->styleOn;

  weight = base->font->GetWeight();
  match = (weight == u.delta->weightOff);
  if (match)
    weight = wxNORMAL;
  if (!match || (match && u.delta->weightOn != u.delta->weightOff))
    if (u.delta->weightOn != wxBASE)
      weight = u.delta->weightOn;

  target->alignment = base->alignment;
  match = (target->alignment == u.delta->alignmentOff);
  if (match)
    target->alignment = wxALIGN_BOTTOM;
  if (!match || (match && u.delta->alignmentOn != u.delta->alignmentOff))
    if (u.delta->alignmentOn != wxBASE)
      target->alignment = u.delta->alignmentOn;

  if (u.delta->underlinedOff && u.delta->underlinedOn)
    underlined = !base->font->GetUnderlined();
  else if (u.delta->underlinedOff)
    underlined = FALSE;
  else if (u.delta->underlinedOn)
    underlined = TRUE;
  else
    underlined = base->font->GetUnderlined();
  
  target->font = wxTheFontList->FindOrCreateFont(size, fontid,
						 style, weight, underlined);

  base->foreground.Get(&r, &g, &b);
  u.delta->foregroundMult.Get(&rm, &gm, &bm);
  u.delta->foregroundAdd.Get(&rp, &gp, &bp);
  r = ColourNum(r * rm + rp);
  g = ColourNum(g * gm + gp);
  b = ColourNum(b * bm + bp);
  target->foreground.Set(r, g, b);

  base->background.Get(&r, &g, &b);
  u.delta->backgroundMult.Get(&rm, &gm, &bm);
  u.delta->backgroundAdd.Get(&rp, &gp, &bp);
  r = ColourNum(r * rm + rp);
  g = ColourNum(g * gm + gp);
  b = ColourNum(b * bm + bp);
  target->background.Set(r, g, b);

  target->pen = wxThePenList->FindOrCreatePen(&foreground, 1, wxSOLID);
  target->brush = wxTheBrushList->FindOrCreateBrush(&background, wxSOLID);

  if (propogate)
    for (node = children.First(); node; node = node->Next())
      ((wxStyle *)node->Data())->Update(NULL, NULL, TRUE, FALSE);

  styleList->StyleWasChanged(target);
  if (topLevel)
    styleList->StyleWasChanged(NULL);
}

char *wxStyle::GetName()
{
  return name;
}

int wxStyle::GetFamily()
{
  return font->GetFamily();
}

char *wxStyle::GetFace()
{
#ifndef NO_GENERAL_FONTS
  return font->GetFaceString();
#else
  return font->GetFamilyString();
#endif
}

wxFont *wxStyle::GetFont()
{
  return font;
}

int wxStyle::GetSize()
{
  return font->GetPointSize();
}

int wxStyle::GetWeight()
{
  return font->GetWeight();
}

int wxStyle::GetStyle()
{
  return font->GetStyle();
}

Bool wxStyle::GetUnderlined()
{
  return font->GetUnderlined();
}

wxColour &wxStyle::GetForeground()
{
  return foreground;
}

wxColour &wxStyle::GetBackground()
{
  return background;
}

int wxStyle::GetAlignment()
{
  return alignment;
}

Bool wxStyle::IsJoin()
{
  return joinStyle;
}

void wxStyle::GetDelta(wxStyleDelta &d)
{
  if (joinStyle)
    d.SetDelta(wxCHANGE_NOTHING);
  else
    memcpy(&d, u.delta, sizeof(wxStyleDelta));
}

void wxStyle::SetDelta(wxStyleDelta &d)
{
  if (joinStyle || PTREQ(this, styleList->BasicStyle()))
    return;

  memcpy(u.delta, &d, sizeof(wxStyleDelta));
  Update();
}

wxStyle *wxStyle::GetShiftStyle()
{
  if (joinStyle)
    return u.shiftStyle;
  else
    return styleList->BasicStyle();
}

void wxStyle::SetShiftStyle(wxStyle *style)
{
  if (!joinStyle || (styleList->StyleToIndex(style) < 0))
    return;

  if (styleList->CheckForLoop(this, style))
    return;

  if (u.shiftStyle)
    u.shiftStyle->children.DeleteObject(this);
  style->children.Append(this);

  u.shiftStyle = style;
  styleList->StyleHasNewChild(style, this);

  Update();

  u.shiftStyle = style;
  Update();
}

wxStyle *wxStyle::GetBaseStyle(void)
{
  return baseStyle;
}

void wxStyle::SetBaseStyle(wxStyle *style)
{
  if (PTREQ(this, styleList->BasicStyle()))
    return;

  if (!style)
    style = styleList->BasicStyle();
  else
    if (styleList->StyleToIndex(style) < 0)
      return;

  if (styleList->CheckForLoop(this, style))
    return;

  if (baseStyle)
    baseStyle->children.DeleteObject(this);

  baseStyle = style;
  style->children.Append(this);

  styleList->StyleHasNewChild(style, this);

  Update();
}

void wxStyle::SwitchTo(wxDC *dc, wxStyle *oldStyle)
{
  unsigned char afr, afg, afb, bfr, bfg, bfb;
  unsigned char abr, abg, abb, bbr, bbg, bbb;

  if (oldStyle) {
    oldStyle->foreground.Get(&afr, &afg, &afb);
    foreground.Get(&bfr, &bfg, &bfb);
    oldStyle->background.Get(&abr, &abg, &abb);
    background.Get(&bbr, &bbg, &bbb);
  }

  if (!oldStyle || oldStyle->font != font)
    dc->SetFont(font);
  if (!oldStyle || afr != bfr || afb != bfb || afg != bfg)
    dc->SetTextForeground(&foreground);
  if (!oldStyle || abr != bbr || abb != bbb || abg != bbg)
    dc->SetTextBackground(&background);
  if (!oldStyle || oldStyle->pen != pen)
    dc->SetPen(pen);
}

/***************************************************************/

class NotificationRec : public wxObject {
 public:
  wxStyleNotifyFunc f;
  void *data;
  long id;
};

static long nextNotifyId = 0;

wxStyleList::wxStyleList() : wxList()
{
  usage = 0;
  Clear();
}

void wxStyleList::Clear(void)
{
  wxNode *node;
  wxStyle *style;

  while (node = First()) {
    style = (wxStyle *)node->Data();
    delete style;
    DeleteNode(node);
  }

  basic = new wxStyle;

  basic->styleList = this;

  basic->name = "Basic";
  basic->baseStyle = NULL;

  basic->u.delta = new wxStyleDelta;
  basic->u.delta->SetDelta(wxCHANGE_NORMAL);

  basic->font = wxTheFontList->FindOrCreateFont(defaultSize, wxDEFAULT,
						wxNORMAL, wxNORMAL);
  basic->foreground = *blackColour;
  basic->background = *whiteColour;
  basic->pen = wxThePenList->FindOrCreatePen(&basic->foreground, 1, wxSOLID);
  basic->brush = wxTheBrushList->FindOrCreateBrush(&basic->background,wxSOLID);
  basic->alignment = wxALIGN_BOTTOM;

  Append(basic);

  notifications = new wxList();

  listId = 0;
  styleMap = NULL;
}

wxStyleList::~wxStyleList()
{
  DeleteContents(TRUE);
}

wxStyle *wxStyleList::BasicStyle(void)
{
  return basic;
}

wxStyle *wxStyleList::FindOrCreateStyle(wxStyle *baseStyle, 
					wxStyleDelta *deltain)
{
  wxNode *node;
  wxStyle *style;
  wxStyleDelta delta;

  if (!baseStyle || (StyleToIndex(baseStyle) < 0))
    baseStyle = basic;

  /* Collapse the delta: */
  memcpy(&delta, deltain, sizeof(wxStyleDelta));
  while (!baseStyle->name && !baseStyle->joinStyle) {
    if (!delta.Collapse(*baseStyle->u.delta))
      break;
    baseStyle = baseStyle->baseStyle;
  }

  for (node = First(); node; node = node->Next()) {
    style = (wxStyle *)node->Data();
    if (!style->name
	&& !style->joinStyle
	&& PTREQ(style->baseStyle, baseStyle)
	&& delta.Equal(*style->u.delta))
      return style;
  }

  style = new wxStyle;

  style->styleList = this;

  style->name = NULL;

  style->u.delta = new wxStyleDelta;
  memcpy(style->u.delta, &delta, sizeof(wxStyleDelta));

  style->baseStyle = baseStyle;
  baseStyle->children.Append(style);

  style->Update();

  Append(style);

  return style;
}

wxStyle *wxStyleList::FindOrCreateJoinStyle(wxStyle *baseStyle, 
					    wxStyle *shiftStyle)
{
  wxNode *node;
  wxStyle *style;

  if (!baseStyle || (StyleToIndex(baseStyle) < 0))
    baseStyle = basic;
  if (!shiftStyle || (StyleToIndex(shiftStyle) < 0))
    return baseStyle;

  for (node = First(); node; node = node->Next()) {
    style = (wxStyle *)node->Data();
    if (!style->name
	&& style->joinStyle
	&& PTREQ(style->baseStyle, baseStyle)
	&& PTREQ(style->u.shiftStyle, shiftStyle))
      return style;
  }

  style = new wxStyle;

  style->styleList = this;

  style->name = NULL;

  style->joinStyle = TRUE;

  style->u.shiftStyle = shiftStyle;
  shiftStyle->children.Append(style);

  style->baseStyle = baseStyle;
  baseStyle->children.Append(style);

  style->Update();

  Append(style);

  return style;
}

wxStyle *wxStyleList::FindNamedStyle(char *name)
{
  wxNode *node;
  wxStyle *style;

  for (node = First(); node; node = node->Next()) {
    style = (wxStyle *)node->Data();
    if (style->name && !strcmp(name, style->name))
      return style;
  }

  return NULL;
}

wxStyle *wxStyleList::DoNamedStyle(char *name, wxStyle *plainStyle, Bool replac)
{
  wxNode *node;
  wxStyle *style;

  if (!plainStyle || (StyleToIndex(plainStyle) < 0))
    plainStyle = basic;
  
  for (node = First(); node; node = node->Next()) {
    style = (wxStyle *)node->Data();
    if (style->name && !strcmp(name, style->name)) {
      if (!replac)
	return style;
      break;
    }
  }

  if (!node) {
    style = new wxStyle;
    style->name = copystring(name);
    style->styleList = this;
  } else {
    style->baseStyle->children.DeleteObject(style);
    if (style->joinStyle)
      style->u.shiftStyle->children.DeleteObject(style);
  }

  if (plainStyle->joinStyle) {
    style->joinStyle = TRUE;
    style->u.shiftStyle = plainStyle->u.shiftStyle;
    style->u.shiftStyle->children.Append(style);
  } else {
    style->u.delta = new wxStyleDelta;
    if (PTRNE(plainStyle, basic))
      memcpy(style->u.delta, plainStyle->u.delta, sizeof(wxStyleDelta));
  }

  if (PTREQ(plainStyle, basic))
    style->baseStyle = basic;
  else
    style->baseStyle = plainStyle->baseStyle;
  style->baseStyle->children.Append(style);

  style->Update();

  Append(style);

  return style;
}

wxStyle *wxStyleList::NewNamedStyle(char *name, wxStyle *plainStyle)
{
  return DoNamedStyle(name, plainStyle, FALSE);
}

wxStyle *wxStyleList::ReplaceNamedStyle(char *name, wxStyle *plainStyle)
{
  return DoNamedStyle(name, plainStyle, TRUE);
}

wxStyle *wxStyleList::Convert(wxStyle *style)
{
  wxStyle *base, *newstyle;

  if (StyleToIndex(style) >= 0)
    return style;

  if (style->name) {
    newstyle = FindNamedStyle(style->name);
    if (newstyle)
      return newstyle;
  }

  if (!style->baseStyle)
    base = BasicStyle();
  else
    base = Convert(style->baseStyle);

  if (style->joinStyle) {
    wxStyle *shift;

    shift = Convert(style->u.shiftStyle);
    newstyle = FindOrCreateJoinStyle(base, shift);
  } else
    newstyle = FindOrCreateStyle(base, style->u.delta);

  if (style->name)
    return NewNamedStyle(style->name, newstyle);
  else
    return newstyle;
}

void wxStyleList::StyleWasChanged(wxStyle *which)
{
  NotificationRec *rec;
  wxNode *node;

  for (node = notifications->First(); node; node = node->Next()) {
    rec = (NotificationRec *)node->Data();
    rec->f(which, rec->data);
  }
}

long wxStyleList::NotifyOnChange(wxStyleNotifyFunc f, void *data)
{
  NotificationRec *rec = new NotificationRec;

  rec->f = f;
  rec->data = data;
  WXGC_IGNORE(rec->data);
  rec->id = nextNotifyId++;
  notifications->Append(rec);

  return rec->id;
}

void wxStyleList::ForgetNotification(long id)
{
  NotificationRec *rec;
  wxNode *node;
  
  for (node = notifications->First(); node; node = node->Next()) {
    rec = (NotificationRec *)node->Data();
    if (rec->id == id) {
      notifications->DeleteNode(node);
      delete rec;
      return;
    }
  }
}

Bool wxStyleList::CheckForLoop(wxStyle *s, wxStyle *p)
{
  if (PTREQ(p, s))
    return TRUE;

  if (!p->baseStyle)
    return FALSE;

  if (p->joinStyle) {
    if (CheckForLoop(s, p->baseStyle))
      return TRUE;
    return CheckForLoop(s, p->u.shiftStyle);
  } else
    return CheckForLoop(s, p->baseStyle);
}

void wxStyleList::StyleHasNewChild(wxStyle *s, wxStyle *c)
{
  wxNode *cnode, *snode, *node;

  /* Need to maintain the invariant that parents are in the list
     before children... */
  cnode = Member(c);
  snode = Member(s);
  
  for (node = cnode; node; node = node->Next()) {
    if (PTREQ(node, snode)) {
      /* Move base style to before this style */
      DeleteNode(snode);
      Insert(cnode, s);
      break;
    }
  }
}

int wxStyleList::Number(void)
{
  return wxList::Number();
}

wxStyle *wxStyleList::IndexToStyle(int i)
{
  wxNode *node;

  for (node = First(); i && node; i--, node = node->Next());

  if (node)
    return (wxStyle *)node->Data();
  else
    return NULL;
}
 
int wxStyleList::StyleToIndex(wxStyle *s)
{
  wxNode *node;
  int i = 0;

  for (node = First(); 
       node && PTRNE((wxStyle *)node->Data(), s); 
       i++, node = node->Next());

  if (node)
    return i;
  else
    return -1;
}

void wxStyleList::AdjustUsage(Bool newUser)
{
  if (newUser)
    usage++;
  else
    --usage;
}

Bool wxStyleList::IsUsed(void)
{
  return !!usage;
}

Bool wxStyleList::WriteToFile(class wxMediaStreamOut &f)
{
  return wxmbWriteStylesToFile(this, f);
}

wxStyle *wxStyleList::MapIndexToStyle(int i)
{
  if (styleMap && i < numMappedStyles)
    return styleMap[i];
  else
    return basic;
}

wxStyleList *wxReadStyleList(class wxMediaStreamIn &f)
{
  wxStyleList *l = new wxStyleList;

  return wxmbReadStylesFromFile(l, f);
}

static wxList *readStyles = NULL;

void wxmbSetupStyleReadsWrites(void)
{
  readStyles = new wxList(wxKEY_INTEGER);
}

void wxmbDoneStyleReadsWrites(void)
{
  wxNode *node;
  wxStyleList *l;

  for (node = readStyles->First(); node; node = node->Next()) { 
    l = (wxStyleList *)node->Data();
    if (l->styleMap)
      delete[] l->styleMap;
    l->listId = 0;
  }

  delete readStyles;
  readStyles = NULL;
}

wxStyleList *wxmbReadStylesFromFile(wxStyleList *styleList, 
				    wxMediaStreamIn& f)
{
#define MAX_STYLE_NAME 256
  int baseIndex, shiftIndex;
  long nameSize;
  char name[MAX_STYLE_NAME];
  char face[MAX_STYLE_NAME];
  short r, g, b;
  int i, isJoin, listId;
  wxStyleDelta delta;
  wxNode *node;

  f >> listId;
  
  if (node = readStyles->Find(listId))
    return (wxStyleList *)node->Data();

  f >> styleList->numMappedStyles;
  styleList->styleMap = new wxStyle*[styleList->numMappedStyles];

  styleList->styleMap[0] = styleList->BasicStyle();
  for (i = 1; i < styleList->numMappedStyles; i++) {
    f >> baseIndex;

    if (baseIndex >= i) {
      wxMessageBox("Bad style index.", "Error");
      return FALSE;
    }

    nameSize = MAX_STYLE_NAME;
    f.Get((long *)&nameSize, (char *)name);

    f >> isJoin;

    if (isJoin) {
      f >> shiftIndex;

      styleList->styleMap[i] = 
	styleList->FindOrCreateJoinStyle(styleList->styleMap[baseIndex], 
					 styleList->styleMap[shiftIndex]);
    } else {
      f >> delta.family;
      nameSize = MAX_STYLE_NAME;
      f.Get((long *)&nameSize, (char *)face);
      
      if (*face)
	delta.face = copystring(face);
      else
	delta.face = NULL;
      
      f >> delta.sizeMult;
      f >> delta.sizeAdd;
      f >> delta.weightOn; 
      f >> delta.weightOff;
      f >> delta.styleOn;
      f >> delta.styleOff;
      f >> delta.underlinedOn;
      f >> delta.underlinedOff;
      
      f >> delta.foregroundMult.r;
      f >> delta.foregroundMult.g;
      f >> delta.foregroundMult.b;
      f >> delta.backgroundMult.r;
      f >> delta.backgroundMult.g;
      f >> delta.backgroundMult.b;
      f >> r;
      f >> g;
      f >> b;
      delta.foregroundAdd.Set(r, g, b);
      f >> r;
      f >> g;
      f >> b;
      delta.backgroundAdd.Set(r, g, b);
      
      f >> delta.alignmentOn;
      f >> delta.alignmentOff;

      styleList->styleMap[i] = 
	styleList->FindOrCreateStyle(styleList->styleMap[baseIndex], &delta);
    }

    if (*name)
      styleList->styleMap[i] = 
	styleList->NewNamedStyle(name, styleList->styleMap[i]);
  }

  readStyles->Append(listId, styleList);

  return styleList;
}

Bool wxmbWriteStylesToFile(wxStyleList *styleList, wxMediaStreamOut &f)
{
  int i, count;
  wxStyle *style;
  short r, g, b;
  char *name;
  wxStyleDelta delta;

  if (styleList->listId) {
    f << styleList->listId;
    return TRUE;
  }

  styleList->listId = readStyles->Number() + 1;

  f << styleList->listId;

  count = styleList->Number();

  f << count;

  // Basic style is implied

  for (i = 1; i < count; i++) {
    style = styleList->IndexToStyle(i);

    f << styleList->StyleToIndex(style->GetBaseStyle());
    if (name = style->GetName())
      f << name;
    else
      f << "";

    if (style->IsJoin()) {
      f << 1;

      f << styleList->StyleToIndex(style->GetShiftStyle());
    } else {
      style->GetDelta(delta);
      
      f << 0;

      f << delta.family;
      if (delta.face)
	f << delta.face;
      else
	f << "";

      f << delta.sizeMult;
      f << delta.sizeAdd;
      f << delta.weightOn; 
      f << delta.weightOff;
      f << delta.styleOn;
      f << delta.styleOff;
      f << delta.underlinedOn;
      f << delta.underlinedOff;

      f << delta.foregroundMult.r;
      f << delta.foregroundMult.g;
      f << delta.foregroundMult.b;
      f << delta.backgroundMult.r;
      f << delta.backgroundMult.g;
      f << delta.backgroundMult.b;
      delta.foregroundAdd.Get(&r, &g, &b);
      f << r;
      f << g;
      f << b;
      delta.backgroundAdd.Get(&r, &g, &b);
      f << r;
      f << g;
      f << b;

      f << delta.alignmentOn;
      f << delta.alignmentOff;
    }
  }

  readStyles->Append(styleList->listId, styleList);

  return TRUE;
}
